/*==========================================================================
**
**	hal.c
**
**	Hardware abstraction layer for Intel 557, 558, 559, 559ER, 550
**
**==========================================================================
**
**	(c) Copyright 2001, 2002 ZNYX Networks
**	All Rights Reserved.
**
**==========================================================================
**
**	Modification History:
**	jr	09/12/01	New
**	jr	01/28/02	vxworks
**	jr	02/18/02	added promiscuous mode
**
**==========================================================================
**
**	Needed: promiscuous mode, phy interrupts
**
**	SYS_DESC_SYNC - to enable calls to pdk_mem_desc_sync and descriptor_sync
**	PDK_BUF_SYNC - to enable calls to pdk_buf_sync	
**
**==========================================================================
*/

#include <pdk_port.h>
#include <pdk_global.h>
#include <pdk_types.h>
#include <pdk_hton.h>

#include <pdk_lib.h>
#include <pdk_buf.h>
#include <pdk_print.h>
#include <pdk_queue.h>
#include <pdk_lock.h>
#include <pdk_mem.h>
#include <pdk_time.h>
#include <pdk_errno.h>
#include <pdk_reg.h>
#include <pdk_hal.h>

#include <pdk_layer.h>
#include <pdk_constant.h>

#include <pdk_br554.h>
#include <pdk_psdb.h>

#include "deviceid.h"
#include "i82558.h"
#include "i82555.h"
#include "dp83840.h"



#define	ZIN_SMALL_RINGS		1
#define	ZIN_SYNC_MEDIA_CHECK	1
#define	ZIN_TIMER_CHECK		1
#define	ZIN_BUFFER_MARK		1
#define	ZIN_LOCK_CHECK		0
#define	ZIN_DUMP_BUFFERS	0
#define	ZIN_DUMP_PACKETS	0

#define	ZIN_LOCAL /*static*/

#ifndef	SYS_INSTR_SYNC
#define	SYS_INSTR_SYNC()
#endif


#define H_READ  1
#define H_WRITE 2

#if 0
#define	H_BROADCAST		4
#endif



#ifdef SYS_DESC_SYNC
#undef SYS_DESC_SYNC
#define	SYS_DESC_SYNC(handlep, deqaddr, enqaddr, type) (pdk_mem_desc_sync(handlep, (char *)deqaddr, (char *) enqaddr, type));
#define	DESCRIPTOR_SYNC(lp, flag) descriptor_sync(lp, flag)
#else
#define	SYS_DESC_SYNC(handlep, deqaddr, enqaddr, type)
#define	DESCRIPTOR_SYNC(lp, flag)
#endif

#ifdef PDK_BUF_SYNC
#undef PDK_BUF_SYNC
#define PDK_BUF_SYNC(bp, len, flags) (pdk_buf_sync(bp, len, flags))
#else
#define PDK_BUF_SYNC(bp, len, flags)
#endif



int32_t pdk_hal_debug = 2;

#define DEBUG0 pdk_print
#define DEBUG1 if(pdk_hal_debug>0) pdk_print
#define DEBUG2 if(pdk_hal_debug>1) pdk_print
#define DEBUG3 if(pdk_hal_debug>2) pdk_print
#define DEBUG4 if(pdk_hal_debug>3) pdk_print
#define DEBUG5 if(pdk_hal_debug>4) pdk_print




/* Hardware abstraction layer interface */

char *pdk_hal_version="ZNYX 8255x HAL Version M11 04";

#ifndef RXD_SIZE
#define RXD_SIZE	64
#endif
#ifndef TXD_SIZE
#define TXD_SIZE	128
#endif

#if ZIN_SMALL_RINGS
int32_t zin_num_rxd = 16;
int32_t zin_num_txd = 16;
#else
int32_t zin_num_rxd = RXD_SIZE;
int32_t zin_num_txd = TXD_SIZE;
#endif
int32_t zxe_loopback_delay = 1000;



/* tx linked descriptor structure */

#define	TXLD_CB_OFFSET	(sizeof(txld_t*)+sizeof(pdk_buf_t*)+sizeof(uint32_t)*2)
#define TXLD_NUM_TBD	((256 - TXLD_CB_OFFSET - sizeof(i82558_tx_t))/sizeof(i82558_tbd_t))

typedef struct txld txld_t;
struct txld {
	txld_t		*next;
	pdk_buf_t	*bp;
	uint32_t	length;
	uint32_t	pad;
	i82558_tx_t	tcb;
	i82558_tbd_t	tbd[TXLD_NUM_TBD];
};



/* offset of rfa from head of pdk_buf_t rptr */

#define	RFA_OFFSET	2



/* local adapter states */

#define LOCAL_ADAPTER_ALLOCATED 0
#define LOCAL_ADAPTER_STOPPED 1
#define LOCAL_ADAPTER_RUNNING 2
#define LOCAL_ADAPTER_RELEASED 3



/* phy bits and pieces */

ZIN_LOCAL char *zin_phy_type[] = {
	"none", "i82553A/B", "i82553C", "i82503", "83840",
	"80c240", "80c24", "i82555", "type 8", "type 9",
	"83840A", "i82555B" };

ZIN_LOCAL int zin_phy_types = sizeof(zin_phy_type)/sizeof(zin_phy_type[0]);

ZIN_LOCAL char *zin_connectors[] = { "RJ45", "BNC", "AUI", "MII", NULL };

ZIN_LOCAL char *zin_media_names[] = {
	"Link Down", "10Mbps Half-Duplex", "10Mbps BNC", "10Mbps AUI",
	"10Mbps Full-Duplex", "100Mbps Half-Duplex", "100Mbps Full-Duplex",
	"100Mbps T4", "Link Auto-Negotiating" };



typedef void zin_mdi_async_cb_t(int32_t ppa, int phy, int reg, uint16_t value);

typedef struct _local_adapter {
	int			hardware_initialized;
	uint32_t		state;

	/* added begin */

	int			la_tbd_array_start;

	pdk_desc_handle_t	*cb_buffer_handlep;
	i82558_union_t		*cb_buffer;
	uint32_t		cb_buffer_paddr;

	pdk_desc_handle_t	*stats_buffer_handlep;
	i82558_stats_t		*stats_buffer;
	uint32_t		stats_buffer_paddr;
	uint32_t		stats_dump_check;

	pdk_desc_handle_t	*txld_space_handlep;
	txld_t			*txld_space;
	uint32_t		txld_space_paddr;
	txld_t			*txld_first;
	txld_t			*txld_last;
	int			tx_queued;

	uint32_t		rfa_head_paddr;
	pdk_buf_t		*rfa_head_bp;
	pdk_buf_t		*rfa_tail_bp;

	i82558_mcs_t		mcs_template;
	i82558_config_t		config_template;
	i82558_ias_t		ias_template;

	int			mdi_async_started;
	int			mdi_async_phy;
	int			mdi_async_reg;
	zin_mdi_async_cb_t	*mdi_async_cb;

	int			phy_addr;
	int			phy_type;
	int			phy_rev;
	int			phy_serial;
	int16_t			phy_control, dummy;
	int16_t			phy_cur_anlpar, phy_cur_par, phy_cur_status;
	int16_t			phy_last_anlpar, phy_last_par, phy_last_status;

	int			led_test;

	int			controller;
	int			flags;
	int			promiscuous_multicast;
	int			promiscuous_physical;
	int			recv_giants;
	int			cu_resume_bug;
	int			to_one_sec;
	int			to_ten_sec;

	i82558_eeprom_t		eeprom;
	uint16_t		*eeprom_w;
	int			eeprom_size;

	void			*timer_id;

	/* added end */

	int32_t			rx_num_bufs;
	int32_t			rx_lwm;
	int32_t			rx_hwm;
	uint32_t		rx_depleted;
	uint32_t		rx_icount;

	int32_t			free_tx_descriptor_count;
	int32_t			tx_lwm;
	int32_t			tx_depleted;
	uint32_t		tx_icount;

	/* fixme: makes no sense */
	uint32_t		rx_descriptor_count;
	uint32_t		tx_descriptor_count;

	int32_t			tx_pending_count; /* for periodic interrupts */
	uint32_t		tx_pending_threshold; /* interrupt when this number is hit */
	uint32_t		other_icount;

	pdk_lock_t		*rx_descriptor_lock;
	pdk_lock_t		*tx_descriptor_lock;

#if 0
	uint32_t		pdk_xmit_complete_check;
#endif
	uint32_t		fault_notification;

	uchar			factory_mac_address[PDK_ETHERADDRL];
	uchar			current_mac_address[PDK_ETHERADDRL];

	uint32_t		configured_media_type;
	uint32_t		current_media_type;
	int			link_status_up;


#if 0
	uint32_t		PacketFilter;
	uint32_t		PollTimeout;
	uint32_t		RxTimeoutCounter;
	uint32_t		TxBufferThreshold;
	uint32_t		TxTimeoutCounter;
#endif

	int32_t			multicast_address_count;
	uchar			multicast_addresses[I82558_MAX_MCA][PDK_ETHERADDRL];

	/* adapter statistics */

	pdk_hal_stat_t		Statistics;
	int32_t			ppa;
	pdk_layer_routines_t	layer_routines;

#ifdef PDK_HAL_SPEEDO
	void			*hal_speedometer_timeout_id;
	pdk_hal_stat_t		prev_Statistics;/* previous stats */
	hal_speedometer_t	hsp;
	hal_speedometer_t	set_params_gte_hsp;	/* greater than */
	hal_speedometer_t	set_params_lt_hsp;	/* less than or equal */
	uint32_t		prev_rx_icount;
	uint32_t		prev_tx_icount;
	uint32_t		tx_bytes;
	uint32_t		rx_bytes;
	uint32_t		speedometer_poll_rate;
#endif

} local_adapter_t;

local_adapter_t *local_adapters[NUM_PORTS];



#define DRV_FLAG_MWI_ENABLE	0x0001	/* MWI enable */
#define DRV_FLAG_READ_ALIGN	0x0002	/* align read access with cacheline */
#define DRV_FLAG_WRITE_ALIGN	0x0004	/* end write on cacheline */
#define DRV_FLAG_EXT_TXCB	0x0008	/* enable use of extended TXCB */
#define DRV_FLAG_LONG_PKT_EN	0x0020	/* enable long packet reception */
#define DRV_FLAG_ALL_MCAST	0x0040	/* accept all multicast frames */
#define DRV_FLAG_CU_RESUME_BUG	0x0080	/* requires workaround for CU_RESUME */





/* timer intervals */

#define	TIMEOUT_ONE_SECOND	1000000
#define	TIMEOUT_DELAY		1000000
#define	TIMEOUT_INTERVAL	1000000
ZIN_LOCAL void zin_timer(int32_t ppa);



uint32_t pdk_hal_iac(int32_t ppa);
uint32_t pdk_hal_intr(int32_t ppa, uint32_t iac);
uint32_t pdk_hal_xmit(int32_t ppa, pdk_buf_t *bp);
uint32_t pdk_hal_recv(int32_t ppa, pdk_buf_t *bp);

void hal_psdb_return_state(local_adapter_t *lp, pdk_psdb_hal_state_t *phsp);

#ifdef PDK_HAL_SPEEDO
/* convert ms to usec */
#define HAL_SPEEDOMETER_TICKS (lp->speedometer_poll_rate*1000)
#define HAL_SPEEDOMETER_RATE(val) \
	((((uint32_t) (val)) < (0xffffffff/1000)) ? \
		 ((((uint32_t) (val))*1000)/lp->speedometer_poll_rate) : \
		 ((((uint32_t) (val))/lp->speedometer_poll_rate)*1000))
void hal_psdb_return_speed(local_adapter_t *lp, pdk_psdb_hal_speedometer_get_t *phsp);
void hal_speedometer_timeout(void *arg);
#endif

void hal_psdb_return(local_adapter_t *lp, uint32_t type, uint32_t return_val);
uint32_t hal_psdb(pdk_psdb_rhdr_t *record);
ZIN_LOCAL uint32_t hal_get_state(int32_t ppa, hal_state_t *hsp);
ZIN_LOCAL uint32_t hal_get_stats(int32_t ppa, pdk_hal_stat_t *stat);

#ifdef PDK_HAL_SPEEDO
uint32_t hal_get_speed(int32_t ppa, hal_speedometer_t *hsp);
uint32_t hal_set_speedometer_params(int32_t ppa, hal_speedometer_t *hsp);
#endif



ZIN_LOCAL pdk_layer_routines_t zin_layer_routines = {
	pdk_hal_iac, /* pdk_iac */
	pdk_hal_intr, /* pdk_intr */
	pdk_hal_xmit, /* pdk_xmit */
	pdk_hal_recv, /* pdk_recv */
	pdk_hal_ctl, /* pdk_ctl */
	0, /* pdk_xmit_complete */
	0, /* pdk_recv_complete */
	0  /* pdk_status */
};



/* Local routines */

ZIN_LOCAL void zin_process_recv_complete(int32_t ppa, local_adapter_t *lp, uint32_t isr_value);
ZIN_LOCAL void zin_process_xmit_complete(local_adapter_t *lp);

ZIN_LOCAL uint32_t pdk_hal_stop(int32_t ppa);
ZIN_LOCAL int32_t zin_add_multicast_address(int32_t ppa, uchar *addr);
ZIN_LOCAL int32_t zin_del_multicast_address(int32_t ppa, uchar *addr);




/* move this to a different module */

void	pdk_print_ether_packet(char *id, uint8_t *p, int length)
{
	if(id == NULL) id = "ETH:";

	if(p == NULL || length < 14) {
		pdk_print("pdk_print_ether_packet(%s, 0x%08x, %d)\n", id, p, length);
		return;
	}

	pdk_print("%s %02x%02x%02x%02x%02x%02x %02x%02x%02x%02x%02x%02x %02x%02x %02x%02x%02x%02x len=%u\n",
		id,
		p[0], p[1], p[2], p[3], p[4], p[5],
		p[6], p[7], p[8], p[9], p[10], p[11],
		p[12], p[13],
		p[14], p[15], p[16], p[17],
		length);

	if(p[12] == 0x08 && p[13] == 0x06 && length >= 14+28) {
		uint8_t *arp = p+14, *a = arp+8;
		pdk_print("  arp: %s  sender=%02x%02x%02x%02x%02x%02x/%u.%u.%u.%u  target=%02x%02x%02x%02x%02x%02x/%u.%u.%u.%u\n",
		(arp[7] == 1) ? "reqst" : (arp[7] == 2) ? "reply" : "op???",
		a[0], a[1], a[2], a[3], a[4], a[5],
		a[6], a[7], a[8], a[9],
		a[10], a[11], a[12], a[13], a[14], a[15],
		a[16], a[17], a[18], a[19]);
	}

	if(p[12] == 0x08 && p[13] == 0x00 && length >= 14+20) {
		uint8_t *ip = p+14, *dp = (ip+((ip[0]&0x0f)<<2));
		pdk_print("  ip: %u.%u.%u.%u -> %u.%u.%u.%u",
		ip[12], ip[13], ip[14], ip[15],
		ip[16], ip[17], ip[18], ip[19]);

		if(ip[9] == 1) {
			/* icmp */
			static char *icmps[] = {
				"echo reply", "1", "2", "unreach", "quench",
				"redirect", "6", "7", "echo request", "9",
				"10", "time exc", "param pr", "timest req",
				"timest rep", "info req", "info rep",
				"amask req", "amask rep" };

			pdk_print(" icmp %s %u/%u id-seq/data=%02x.%02x.%02x.%02x\n",
				(dp[0] < 19) ? icmps[dp[0]] : "???",
				dp[0], dp[1], dp[4], dp[5], dp[6], dp[7]);
		} else if(ip[9] == 6 || ip[9] == 17) {
			/* tcp, udp */
			uint16_t *ports = (uint16_t*)dp;
			pdk_print(" %s %u -> %u\n",
				(ip[9] == 6) ? "tcp" : "udp",
				pdk_htons(ports[0]), pdk_htons(ports[1]));
		} else {
			pdk_print(" prot=%u ??\n", ip[9]);
		}
	}
}



#if ZIN_LOCK_CHECK

#define	zin_acquire_tx_lock(_la) {					\
	if(pdk_lock_held((_la)->tx_descriptor_lock)) {			\
		DEBUG1("eehgawd: tx locked (line %d)\n", __LINE__);	\
	}								\
	else {								\
		pdk_lock_acquire((_la)->tx_descriptor_lock);		\
	}								\
}

#define	zin_release_tx_lock(_la) {					\
	if(!pdk_lock_held((_la)->tx_descriptor_lock)) {			\
		DEBUG1("eehgawd: tx not locked (line %d)\n", __LINE__);	\
	}								\
	else {								\
		pdk_lock_release((_la)->tx_descriptor_lock);		\
	}								\
}

#define	zin_acquire_rx_lock(_la) {					\
	if(pdk_lock_held((_la)->rx_descriptor_lock)) {			\
		DEBUG1("eehgawd: rx locked (line %d)\n", __LINE__);	\
	}								\
	else {								\
		pdk_lock_acquire((_la)->rx_descriptor_lock);		\
	}								\
}

#define	zin_release_rx_lock(_la) {					\
	if(!pdk_lock_held((_la)->rx_descriptor_lock)) {			\
		DEBUG1("eehgawd: rx not locked (line %d)\n", __LINE__);	\
	}								\
	else {								\
		pdk_lock_release((_la)->rx_descriptor_lock);		\
	}								\
}

#else /* ZIN_LOCK_CHECK */

#define	zin_acquire_tx_lock(_la) pdk_lock_acquire((_la)->tx_descriptor_lock)

#define	zin_release_tx_lock(_la) pdk_lock_release((_la)->tx_descriptor_lock)

#define	zin_acquire_rx_lock(_la) pdk_lock_acquire((_la)->rx_descriptor_lock)

#define	zin_release_rx_lock(_la) pdk_lock_release((_la)->rx_descriptor_lock)

#endif /* ZIN_LOCK_CHECK */




ZIN_LOCAL void zin_reset_chip(int32_t ppa)
{
	/* fixme: does this disable interrupts? */
	pdk_reg_put32(ppa, I82558_PI_REG, I82558_PI_SEL_RESET);
	pdk_time_delay(50);
}

ZIN_LOCAL void zin_enable_interrupts(int32_t ppa)
{
	DEBUG5("zin_enable_interrupts(%d)\n", (int)ppa);
	pdk_reg_put8(ppa, I82558_IC_REG, 0);
}

ZIN_LOCAL void zin_disable_interrupts(int32_t ppa)
{
	pdk_reg_put8(ppa, I82558_IC_REG, I82558_IC_MASK);
	DEBUG5("zin_disable_interrupts(%d)\n", (int)ppa);
}




ZIN_LOCAL uchar zin_ether_is_zero(uint8_t *ether)
{
	return(!(ether[0] || ether[1] || ether[2] || ether[3] || ether[4] || ether[5]));
}




ZIN_LOCAL void zin_eeprom_write(uint32_t ppa, uint16_t value)
{
	pdk_reg_put16(ppa, I82558_EC_REG, value);
	pdk_time_delay(1);
}

ZIN_LOCAL void zin_eeprom_write_bits(uint32_t ppa, int data, int length)
{
	uint16_t reg;

	while(length--) {
		if(data & (1<<length))
			reg = I82558_EC_EECS | I82558_EC_EEDI;
		else
			reg = I82558_EC_EECS;
		zin_eeprom_write(ppa, reg);
		zin_eeprom_write(ppa, reg | I82558_EC_EESK);
		zin_eeprom_write(ppa, reg);
	}
}

ZIN_LOCAL uint16_t zin_eeprom_getword(uint32_t ppa, int offset)
{
	local_adapter_t *lp = local_adapters[ppa];
	int bits;
	uint16_t reg;

	/* chip select */
	zin_eeprom_write(ppa, I82558_EC_EECS);

	/* read op */
	zin_eeprom_write_bits(ppa, I82558_EOP_READ, 3);

	/* address */
	bits = lp->eeprom_size ? lp->eeprom_size : 12;
	while(bits--) {
		if(offset & (1<<bits))
			reg = I82558_EC_EECS | I82558_EC_EEDI;
		else
			reg = I82558_EC_EECS;
		zin_eeprom_write(ppa, reg);
		zin_eeprom_write(ppa, reg | I82558_EC_EESK);
		zin_eeprom_write(ppa, reg);
		if(lp->eeprom_size == 0 && (pdk_reg_get16(ppa, I82558_EC_REG) & I82558_EC_EEDO) == 0) {
			lp->eeprom_size = 12 - bits;
			break;
		}
	}

	/* data */
	bits = 16;
	reg = 0;
	while(bits--) {
		zin_eeprom_write(ppa, I82558_EC_EECS | I82558_EC_EESK);
		if(pdk_reg_get16(ppa, I82558_EC_REG) & I82558_EC_EEDO)
			reg |= (1<<bits);
		zin_eeprom_write(ppa, I82558_EC_EECS);
	}

	/* drop chip select */
	zin_eeprom_write(ppa, 0);

	return(reg);
}




ZIN_LOCAL int zin_mdi_ready(int32_t ppa)
{
	int count = 10000;
	while(count-- && (pdk_reg_get32(ppa, I82558_MC_REG) & I82558_MC_READY) == 0)
		pdk_time_delay(10);
	return(count);
}

/* start an asynchronous mdi read operation */

ZIN_LOCAL int32_t
zin_mdi_async_read_reg(int32_t ppa, int phy, int reg, zin_mdi_async_cb_t *cb)
{
	local_adapter_t	*lp;

	if((lp = local_adapters[ppa]) == NULL) {
		return(PDK_OUT_STATE);
	}

	if(!lp->tx_descriptor_lock) {
		return(PDK_OUT_STATE);
	}

	zin_acquire_tx_lock(lp);

	if(!zin_mdi_ready(ppa)) {
		zin_release_tx_lock(lp);
		return(PDK_HW_ERR);
	}

	if(lp->mdi_async_started) {
		zin_release_tx_lock(lp);
		return(PDK_OUT_STATE);
	}

	lp->mdi_async_phy = phy;
	lp->mdi_async_reg = reg;
	lp->mdi_async_cb = cb;

	pdk_reg_put32(ppa, I82558_MC_REG, I82558_MC_IE | I82558_MC_READ | (reg << 16) | (phy << 21));
	lp->mdi_async_started = 1;

	zin_release_tx_lock(lp);

	return(0);
}

/* finish an asynchronous mdi read operation, called from isr */

ZIN_LOCAL void
zin_mdi_async_read_complete(int32_t ppa, local_adapter_t *lp)
{
	uint16_t		value;
	int			phy, reg;
	zin_mdi_async_cb_t	*cb;

	zin_acquire_tx_lock(lp);

	phy = lp->mdi_async_phy;
	reg = lp->mdi_async_reg;
	cb = lp->mdi_async_cb;

	value = pdk_reg_get32(ppa, I82558_MC_REG);

	lp->mdi_async_cb = NULL;
	lp->mdi_async_started = 0;

	zin_release_tx_lock(lp);

	if(cb) cb(ppa, phy, reg, value);
}

ZIN_LOCAL volatile uint16_t
zin_mdi_read_reg(int32_t ppa, int phy, int reg)
{
	local_adapter_t	*lp;
	uint32_t	value = 0;

	if((lp = local_adapters[ppa]) != NULL) lp->mdi_async_started = 0;

	zin_mdi_ready(ppa);

	pdk_reg_put32(ppa, I82558_MC_REG, I82558_MC_READ | (reg << 16) | (phy << 21));

	if(zin_mdi_ready(ppa)) {
		value = pdk_reg_get32(ppa, I82558_MC_REG);
	}
	else {
		DEBUG2("zin_mdi_read_reg(%d): phy=%d reg=%d timed out\n", (int)ppa, phy, reg);
	}
	return((uint16_t)value & 0xffff);
}

ZIN_LOCAL void
zin_mdi_write_reg(int32_t ppa, int phy, int reg, uint16_t value)
{
	local_adapter_t	*lp;

	if((lp = local_adapters[ppa]) != NULL) lp->mdi_async_started = 0;

	zin_mdi_ready(ppa);

	pdk_reg_put32(ppa, I82558_MC_REG, I82558_MC_WRITE | (reg << 16) | (phy << 21) | (value & 0xffff));

#if 0
	count = 10000;
	while(count-- && (pdk_reg_get32(ppa, I82558_MC_REG) & I82558_MC_READY) == 0)
		pdk_time_delay(10);

	if (count <= 0) {
#endif
	if(!zin_mdi_ready(ppa)) {
		DEBUG1("zin_mdi_write_reg(%d): phy=%d reg=%d timed out\n", (int)ppa, phy, reg);
	}
}



ZIN_LOCAL int zin_detect_phy(int32_t ppa)
{
	local_adapter_t	*lp;
	int phy;
	uint16_t pid;

	if((lp = local_adapters[ppa]) == NULL) return(-1);

	for(phy = 0; phy < 32; phy++) {
		pid = zin_mdi_read_reg(ppa, phy, 2);
	DEBUG1("phy%d pid1 = 0x%04x\n", phy, pid);
		if(pid) {
			lp->phy_addr = phy;
			return(0);
		}
	}

	return(-1);
}

ZIN_LOCAL volatile uint16_t zin_read_phy_reg(int32_t ppa, int reg)
{
	local_adapter_t	*lp;

	if((lp = local_adapters[ppa]) == NULL) return(-1);

	return(zin_mdi_read_reg(ppa, lp->phy_addr, reg));
}

ZIN_LOCAL void zin_write_phy_reg(int32_t ppa, int reg, uint16_t value)
{
	local_adapter_t	*lp;

	if((lp = local_adapters[ppa]) == NULL) return;

	zin_mdi_write_reg(ppa, lp->phy_addr, reg, value);
}



/* initialize media-dependent hardware */

ZIN_LOCAL int32_t zin_media_init(int32_t ppa)
{
	local_adapter_t *lp = local_adapters[ppa];
	uint16_t	pid1, pid2, pid2m;

	/* read phy id words, mask stepping */

	pid1 = zin_read_phy_reg(ppa, 2);
	pid2 = zin_read_phy_reg(ppa, 3);
	pid2m = pid2 & 0xfff0;

	lp->phy_rev = pid2 & 0xf;

	if(pid1 == I82555_PHYIDR1 && pid2m == I82555_PHYIDR2) {
		DEBUG2("zin_media_init(%d): Intel 82555\n", (int)ppa);
		lp->phy_type = DEVICE_82555;
		zin_write_phy_reg(ppa, I82555_CR_REG, I82555_CR_PWRDN);
	}
	else if(pid1 == DP83840_PHYIDR1 && pid2m == DP83840_PHYIDR2) {
		DEBUG2("zin_media_init(%d): National DP83840 rev %d\n", (int)ppa, lp->phy_rev);
		lp->phy_type = DEVICE_DP83840;
		zin_write_phy_reg(ppa, DP83840_BMCR_REG, DP83840_BMCR_PWRDN);
	}
	else {
		lp->phy_type = -1;
		DEBUG1("zin_media_init(%d): unknown PHY type %04x/%04x\n", (int)ppa, pid1, pid2);
	}

	return(0);
}



/* called when interface goes up or configured media changes */

ZIN_LOCAL int32_t zin_media_config(int32_t ppa)
{
	local_adapter_t *lp = local_adapters[ppa];

	if(lp->phy_type == DEVICE_82555) {
		switch(lp->configured_media_type) {
		case ZXE_MEDIA_AUTO:
			lp->phy_control = I82555_CR_ANENA;
			zin_write_phy_reg(ppa, I82555_CR_REG, 0);
			break;
		case ZXE_MEDIA_TP:
			lp->phy_control = 0;
			break;
		case ZXE_MEDIA_TPFD:
			lp->phy_control = I82555_CR_FD;
			break;
		case ZXE_MEDIA_TX:
			lp->phy_control = I82555_CR_100;
			break;
		case ZXE_MEDIA_TXFD:
			lp->phy_control = I82555_CR_100 | I82555_CR_FD;
			break;
#if 0
		case ZXE_LOOPBACK_TPINT:
			break;
		case ZXE_LOOPBACK_TXINT:
			break;
		case ZXE_LOOPBACK_TXEXT:
			break;
#endif
		}
		zin_write_phy_reg(ppa, I82555_CR_REG, lp->phy_control);
	}
	else if(lp->phy_type == DEVICE_DP83840) {
		uint16_t temp;
		temp = zin_read_phy_reg(ppa, DP83840_PCR_REG);
		temp |= DP83840_PCR_TRS|DP83840_PCR_FC;
		DEBUG1("zin_media_config(%d): setting 83840 pcr to 0x%04x\n", (int)ppa, temp);
		zin_write_phy_reg(ppa, DP83840_PCR_REG, temp);
		switch(lp->configured_media_type) {
		case ZXE_MEDIA_AUTO:
			lp->phy_control = DP83840_BMCR_ANENA;
			zin_write_phy_reg(ppa, DP83840_BMCR_REG, 0);
			break;
		case ZXE_MEDIA_TP:
			lp->phy_control = 0;
			break;
		case ZXE_MEDIA_TPFD:
			lp->phy_control = DP83840_BMCR_FD;
			break;
		case ZXE_MEDIA_TX:
			lp->phy_control = DP83840_BMCR_100;
			break;
		case ZXE_MEDIA_TXFD:
			lp->phy_control = DP83840_BMCR_100 | DP83840_BMCR_FD;
			break;
#if 0
		case ZXE_LOOPBACK_TPINT:
			break;
		case ZXE_LOOPBACK_TXINT:
			break;
		case ZXE_LOOPBACK_TXEXT:
			break;
#endif
		}
		zin_write_phy_reg(ppa, DP83840_BMCR_REG, lp->phy_control);
	}
	else {
		DEBUG1("zin_media_config(%d): no handler for phy_type %d\n", (int)ppa, lp->phy_type);
	}

	return(0);
}




/* called to observe current media, check on negotiation, etc */

ZIN_LOCAL int32_t zin_media_check_finish(int32_t ppa)
{
	local_adapter_t *lp = local_adapters[ppa];
	int16_t		status, anlpar, par;
	int		media_changed, link_status_up;
	uint32_t	media_type;

	status = lp->phy_cur_status;
	anlpar = lp->phy_cur_anlpar;
	par = lp->phy_cur_par;
	media_changed = 0;
	media_type = lp->current_media_type;
	link_status_up = lp->link_status_up;

	if(lp->phy_type == DEVICE_DP83840) {
		if(status & DP83840_BMSR_RFLT) {
			media_type = ZXE_MEDIA_AUTO;
		}
		else if(!(status & DP83840_BMSR_LNKST)) {
			media_type = ZXE_MEDIA_AUTO;
		}
		else if(status & DP83840_BMSR_ANCPL) {
			if(anlpar & DP83840_ANLPAR_TXFD)
				media_type = ZXE_MEDIA_TXFD;
			else if(anlpar & DP83840_ANLPAR_TX)
				media_type = ZXE_MEDIA_TX;
			else if(anlpar & DP83840_ANLPAR_TPFD)
				media_type = ZXE_MEDIA_TPFD;
			else if(anlpar & DP83840_ANLPAR_TP)
				media_type = ZXE_MEDIA_TP;
			else {
				DEBUG1("zin_media_check_finish(%d): ancmpl, no capabilities\n");
				if(par & DP83840_PAR_SPEED10) {
					media_type = ZXE_MEDIA_TP;
				}
				else {
					media_type = ZXE_MEDIA_TX;
				}
			}
		}

		if(media_type == ZXE_MEDIA_AUTO) {
			link_status_up = 0;
		}
		else {
			link_status_up = 1;
		}
	}
	else if(lp->phy_type == DEVICE_82555) {
		if(status & I82555_SR_RFLT) {
			media_type = ZXE_MEDIA_AUTO;
		}
		else if(!(status & I82555_SR_LNKST)) {
			media_type = ZXE_MEDIA_AUTO;
		}
		else if(status & I82555_SR_ANCPL) {
			if(anlpar & I82555_ANLPAR_TXFD)
				media_type = ZXE_MEDIA_TXFD;
			else if(anlpar & I82555_ANLPAR_TX)
				media_type = ZXE_MEDIA_TX;
			else if(anlpar & I82555_ANLPAR_TPFD)
				media_type = ZXE_MEDIA_TPFD;
			else if(anlpar & I82555_ANLPAR_TP)
				media_type = ZXE_MEDIA_TP;
			else {
				media_type = ZXE_MEDIA_AUTO;
				DEBUG1("zin_media_check_finish(%d): ancmpl, no capabilities?\n");
			}
		}

		if(media_type == ZXE_MEDIA_AUTO) {
			link_status_up = 0;
		}
		else {
			link_status_up = 1;
		}
	}
	else {
		DEBUG1("zin_media_check_finish(%d): no handler for phy_type %d\n", (int)ppa, lp->phy_type);
	}

	if(lp->phy_last_status != status || lp->phy_last_anlpar != anlpar || lp->phy_last_par != par) {
		DEBUG2("zin_media_check_finish(%d): st=0x%04x anlpar=0x%04x par=0x%04x\n", (int)ppa, status, anlpar, par);
		lp->phy_last_status = status;
		lp->phy_last_anlpar = anlpar;
		lp->phy_last_par = par;
	}

	if(media_type != lp->current_media_type) {
		DEBUG2("zin_media_check_finish(%d): changed from %s to %s\n", (int)ppa, zin_media_names[lp->current_media_type], zin_media_names[media_type]);
		lp->current_media_type = media_type;
		media_changed = 1;
	}

	if(link_status_up != lp->link_status_up) {
		lp->link_status_up = link_status_up;
		media_changed = 1;
		if(link_status_up)
			(*lp->layer_routines.pdk_status)(ppa, PDK_STATUS_LINK_UP, 0);
		else
			(*lp->layer_routines.pdk_status)(ppa, PDK_STATUS_LINK_DOWN, 0);
	}

	return(media_changed);
}

ZIN_LOCAL void zin_media_check_have_par(int32_t ppa, int phy, int reg, uint16_t value)
{
	local_adapter_t *lp = local_adapters[ppa];

	lp->phy_cur_par = value;

	DEBUG4("zin_media_check_have_par(%d): 0x%04x\n", (int)ppa, value);
	zin_media_check_finish(ppa);
}

ZIN_LOCAL void zin_media_check_have_anlpar(int32_t ppa, int phy, int reg, uint16_t value)
{
	local_adapter_t *lp = local_adapters[ppa];

	lp->phy_cur_anlpar = value;

	DEBUG4("zin_media_check_have_anlpar(%d): 0x%04x\n", (int)ppa, value);
	if(lp->phy_type == DEVICE_82555) {
		zin_media_check_finish(ppa);
	}
	else {
		zin_mdi_async_read_reg(ppa, lp->phy_addr, DP83840_PAR_REG, zin_media_check_have_par);
	}
}

ZIN_LOCAL void zin_media_check_have_status(int32_t ppa, int phy, int reg, uint16_t value)
{
	local_adapter_t *lp = local_adapters[ppa];

	lp->phy_cur_status = value;

	DEBUG4("zin_media_check_have_status(%d): 0x%04x\n", (int)ppa, value);
	zin_mdi_async_read_reg(ppa, lp->phy_addr, I82555_ANLPAR_REG, zin_media_check_have_anlpar);
}

ZIN_LOCAL void zin_media_check_start(int32_t ppa)
{
	local_adapter_t *lp = local_adapters[ppa];

	DEBUG4("zin_media_check_start(%d)\n", (int)ppa);
	zin_mdi_async_read_reg(ppa, lp->phy_addr, I82555_SR_REG, zin_media_check_have_status);
}

/* synchronous version of zin_media_check_start() */

ZIN_LOCAL int32_t zin_media_check(int32_t ppa)
{
	local_adapter_t *lp = local_adapters[ppa];

	lp->phy_cur_status = zin_read_phy_reg(ppa, DP83840_BMSR_REG);
	lp->phy_cur_anlpar = zin_read_phy_reg(ppa, DP83840_ANLPAR_REG);
	lp->phy_cur_par = zin_read_phy_reg(ppa, DP83840_PAR_REG);

	return(zin_media_check_finish(ppa));
}




ZIN_LOCAL uint32_t
zin_tcb_paddr(local_adapter_t *lp, void *vaddr)
{
	return(lp->txld_space_paddr + ((char*)vaddr) - ((char*)lp->txld_space));
}



/* Wait for the previous command to be accepted (but not
   necessarily completed). */

ZIN_LOCAL void
zin_scb_wait(int32_t ppa)
{
	int i = 10000;

	while (pdk_reg_get8(ppa, I82558_CMD_REG) && --i)
		pdk_time_delay(2);
	if (i == 0) {
		DEBUG1("zin_scb_wait(%d): SCB timeout: 0x%x 0x%x 0x%x 0x%x\n", (int)ppa,
		    pdk_reg_get8(ppa, I82558_CMD_REG),
		    pdk_reg_get8(ppa, I82558_IS_REG),
		    pdk_reg_get8(ppa, I82558_ST_REG),
		    pdk_reg_get16(ppa, I82558_FCT_REG));
	}
}

ZIN_LOCAL void
zin_scb_cmd(int32_t ppa, int cmd)
{
	local_adapter_t *lp = local_adapters[ppa];
	SYS_INSTR_SYNC();
	if (cmd == I82558_CMD_CU_RESUME && lp->cu_resume_bug) {
		pdk_reg_put8(ppa, I82558_CMD_REG, I82558_CMD_NOP);
		SYS_INSTR_SYNC();
		zin_scb_wait(ppa);
	}
	pdk_reg_put8(ppa, I82558_CMD_REG, cmd);
	SYS_INSTR_SYNC();
}

ZIN_LOCAL void
zin_dma_wait(int32_t ppa, volatile uint16_t *status)
{
	int i = 10000;

	while (!(*status & pdk_htobs(I82558_CB_STATUS_C)) && --i)
		pdk_time_delay(2);
	if (i == 0)
		DEBUG1("zin_dma_wait(%d): DMA timeout\n", (int)ppa);
}





/* Must hold transmit lock when calling */

ZIN_LOCAL uint32_t
zin_sync_command(ppa, data_buf, data_size)
int32_t ppa;
char *data_buf;
uint32_t data_size;
{
	local_adapter_t	*lp;
	uchar		*rptr;


	if ((lp = local_adapters[ppa]) == 0) {
		DEBUG1("zin_sync_command(%d):  out of state\n", (int)ppa);
		return(PDK_OUT_STATE);
	}
	if (!lp->cb_buffer) {
		DEBUG1("zin_sync_command(%d):  no cb buffer\n", (int)ppa);
		return(PDK_OUT_STATE);
	}

	DEBUG2("zin_sync_command(%d): command=0x%x size=%d\n", (int)ppa, pdk_htobs(((i82558_nop_t*)data_buf)->nop_command), data_size);

#if 0
/* fixme: check out */
	zin_process_xmit_complete(lp);
#endif


	rptr = (char*)lp->cb_buffer;

	pdk_bcopy(data_buf, rptr, data_size);

	DEBUG2("zin_sync_command(%d): pa=0x%x\n", (int)ppa, lp->cb_buffer_paddr);
	DEBUG2("zin_sync_command(%d): command=0x%x link=0x%x\n", (int)ppa, pdk_htobs(lp->cb_buffer->nop.nop_command), pdk_htobl(lp->cb_buffer->nop.nop_link_addr));

	zin_scb_wait(ppa);
	pdk_reg_put32(ppa, I82558_GP_REG, lp->cb_buffer_paddr);
	zin_scb_cmd(ppa, I82558_CMD_CU_START);
	zin_dma_wait(ppa, &lp->cb_buffer->nop.nop_status);

	DEBUG2("zin_sync_command(%d): status=0x%x\n", (int)ppa, pdk_htobs(lp->cb_buffer->nop.nop_status));

	return(0);
}




ZIN_LOCAL void zin_build_config(local_adapter_t *lp, i82558_config_t *ccb)
{
	int	i;

	/* preset config block */

	for(i = 0; i < 22; i++)
		ccb->cf_bytes[i] = 0;
#if 1
	ccb->cf_bytes[6]  = 0x32;
	ccb->cf_bytes[10] = 0x06;
	ccb->cf_bytes[14] = 0xf2;
	ccb->cf_bytes[15] = 0x48;
	ccb->cf_bytes[17] = 0x40;
	ccb->cf_bytes[18] = 0xf0;
	ccb->cf_bytes[20] = 0x3f;
	ccb->cf_bytes[21] = 0x05;
#endif

	ccb->cf_status = 0;
	ccb->cf_command = pdk_htobs(I82558_CB_COMMAND_CONFIGURE | I82558_CB_COMMAND_EL);
	ccb->cf_link_addr = I82558_CB_NULL;

	ccb->cf_bytes[0] = 22;		/* byte count */
	ccb->cf_bytes[1] = 0x08;	/* fifo limits */
	ccb->cf_bytes[2] = 0;		/* ifs */
	ccb->cf_bytes[3] = 0;		/* mwi enable on 8,9 */
	ccb->cf_bytes[4] = 0;		/* rx dma */
	ccb->cf_bytes[5] = 0;		/* tx dma */
	ccb->cf_bytes[6] = 0x2a;	/* ci int */
	if(lp->la_tbd_array_start == 0) ccb->cf_bytes[6] |= 0x10;
	if(lp->controller == 557 || lp->promiscuous_physical) ccb->cf_bytes[6] |= 0x80;
	ccb->cf_bytes[7] = 0x02;	/* 1 underrun retry */
	if(!lp->promiscuous_physical) ccb->cf_bytes[7] |= 0x01;
	ccb->cf_bytes[8] = 0x01;	/* MII */
	ccb->cf_bytes[9] = 0x00;
	ccb->cf_bytes[10] = 0x2e;	/* preamble, source addr ins */
	ccb->cf_bytes[11] = 0x00;
	ccb->cf_bytes[12] = 0x60;
	ccb->cf_bytes[15] = 0x48;
	if(lp->promiscuous_physical) ccb->cf_bytes[15] |= 0x01;
	ccb->cf_bytes[18] = 0xf2;
	if(lp->recv_giants) ccb->cf_bytes[18] |= 0x08;
	ccb->cf_bytes[19] = 0x80;
	ccb->cf_bytes[21] = 0x05;
	if(lp->promiscuous_multicast) ccb->cf_bytes[21] |= 0x08;

#if 0
	ccb->cf_bytes[
	ccb->cf_bytes[
	ccb->cf_bytes[
	ccb->cf_bytes[
#endif

	if(lp->controller != 557) {
		ccb->cf_bytes[16] = 0x1f;
		ccb->cf_bytes[17] = 0x01;
		ccb->cf_bytes[18] = 0xba;
		ccb->cf_bytes[19] = 0xb8;
	}
}




ZIN_LOCAL int32_t zin_config_update(int32_t ppa)
{
	local_adapter_t *lp;
	int32_t		status;
	i82558_config_t	*ccb;
	i82558_ias_t	*cb_ias;
	i82558_mcs_t	*mcs;

	if ((lp = local_adapters[ppa]) == 0) {
		return(PDK_OUT_STATE);
	}

	status = 0;


	/* wait for tx to complete if active */

	if(lp->tx_queued && lp->txld_last) {
		txld_t *txp = lp->txld_last;
		if(!(txp->tcb.tcb_status & pdk_htobs(I82558_CB_STATUS_C))) {
			/* queued tx not complete yet */
			int i, complete;
			/* full frame takes up to 1.2ms, allow for congestion */
			for(i = complete = 0; i < lp->tx_queued*10*4; i++) {
				pdk_time_delay(120);
				if(txp->tcb.tcb_status & pdk_htobs(I82558_CB_STATUS_C)) {
					complete = 1;
					break;
				}
			}
			if(!complete) {
				DEBUG1("zin_multicast_update(%d): %d tx cbs not completed!\n", (int)ppa, lp->tx_queued);
				return(PDK_HW_ERR);
			}
		}
	}


	/* config control block */

	ccb = &lp->config_template;
	zin_build_config(lp, &lp->config_template);

	if(zin_sync_command(ppa, (char*)ccb, sizeof(i82558_config_t))) {
		status = PDK_HW_ERR;
	}


	/* station address setup */

	cb_ias = &lp->ias_template;
	cb_ias->ias_status = 0;
	cb_ias->ias_command = pdk_htobs(I82558_CB_COMMAND_IAS | I82558_CB_COMMAND_EL);
	cb_ias->ias_link_addr = I82558_CB_NULL;
	pdk_bcopy(lp->current_mac_address, (void*)cb_ias->ias_ia, PDK_ETHERADDRL);

	if(zin_sync_command(ppa, (char*)cb_ias, sizeof(i82558_ias_t))) {
		status = PDK_HW_ERR;
	}


	/* multicast filter setup */

	mcs = &lp->mcs_template;
	mcs->mcs_status = 0;
	mcs->mcs_command = pdk_htobs(I82558_CB_COMMAND_MCS | I82558_CB_COMMAND_EL);
	mcs->mcs_link_addr = I82558_CB_NULL;
	mcs->mcs_count = pdk_htobs(lp->multicast_address_count*6);
	/* assume that PDK_ETHERADDRL == 6 */
	pdk_bcopy(lp->multicast_addresses, mcs->mcs_addr, lp->multicast_address_count*6);

	if(zin_sync_command(ppa, (char*)mcs, sizeof(i82558_mcs_t))) {
		status = PDK_HW_ERR;
	}


	/* restart command unit at proper dma address in tcb ring */

	if(lp->txld_last && lp->txld_space_paddr) {
		txld_t *txp = lp->txld_last;

		/* The 55x tcb status field does not reflect the final
		   transmission status as the controller closes out the
		   descriptor before the transmission completes.
		   So if a completed tcb is pending, we replace that
		   tcb with a nop to restart the 55x CU. Otherwise
		   we queue a tcb to do the same. */

		if(!lp->tx_queued) {
			txp = lp->txld_last = lp->txld_first;
			/* txp->bp = NULL; we null them out in xmit_complete() as well */
			txp->length = 0;
			lp->tx_queued = 1;
			}
		txp->tcb.tcb_command = pdk_htobs(I82558_CB_COMMAND_NOP | I82558_CB_COMMAND_S);

		zin_scb_wait(ppa);
		pdk_reg_put32(ppa, I82558_GP_REG, zin_tcb_paddr(lp, &txp->tcb));
		zin_scb_cmd(ppa, I82558_CMD_CU_START);

	DEBUG1("resume at 0x%08x\n", zin_tcb_paddr(lp, &txp->tcb));
	}

	return(status);
}




ZIN_LOCAL int32_t zin_station_address_update(int32_t ppa)
{
	local_adapter_t *lp;

	if ((lp = local_adapters[ppa]) == 0) {
		return(PDK_OUT_STATE);
	}

	DEBUG2("zin_station_address_update(%d)\n", (int)ppa);

	return(zin_config_update(ppa));
}




ZIN_LOCAL int32_t zin_multicast_update(int32_t ppa)
{
	local_adapter_t *lp;

	if ((lp = local_adapters[ppa]) == 0) {
		return(PDK_OUT_STATE);
	}

	DEBUG2("zin_multicast_update(%d)\n", (int)ppa);

	return(zin_config_update(ppa));
}




ZIN_LOCAL int32_t zin_recv_mode_update(int32_t ppa)
{
	local_adapter_t *lp;

	if ((lp = local_adapters[ppa]) == 0) {
		return(PDK_OUT_STATE);
	}

	DEBUG1("zin_recv_mode_update(%d)\n", (int)ppa);

	return(zin_config_update(ppa));
}




/* start the statistical counters dump process */

ZIN_LOCAL void zin_dump_counters_start(int32_t ppa)
{
	local_adapter_t *lp = local_adapters[ppa];

	if(lp->state != LOCAL_ADAPTER_RUNNING) return;

	if(lp->stats_dump_check) return;

	lp->stats_buffer->stats_completion_status = 0;
	lp->stats_dump_check = pdk_htobl(I82558_STATS_COMPL_DRSC);
	zin_scb_wait(ppa);
	zin_scb_cmd(ppa, I82558_CMD_CU_DRSC);
}

/* check for statistical counters dump process completion */

ZIN_LOCAL int zin_dump_counters_complete(int32_t ppa)
{
	int		status = 0;
	local_adapter_t	*lp = local_adapters[ppa];

	if(lp->stats_dump_check && lp->stats_buffer->stats_completion_status == lp->stats_dump_check) {
		i82558_stats_t *st = lp->stats_buffer;

		lp->Statistics.hs_tx_err_packets += pdk_htobl(st->stats_transmit_maxcols)
				+ pdk_htobl(st->stats_transmit_underruns)
				+ pdk_htobl(st->stats_transmit_latecols)
				+ pdk_htobl(st->stats_transmit_lost_crs);
		lp->Statistics.hs_tx_max_collisions += pdk_htobl(st->stats_transmit_maxcols);
		lp->Statistics.hs_tx_late_collisions += pdk_htobl(st->stats_transmit_latecols);
		lp->Statistics.hs_tx_underrun += pdk_htobl(st->stats_transmit_underruns);
		lp->Statistics.hs_tx_carrier_loss += pdk_htobl(st->stats_transmit_lost_crs);
		lp->Statistics.hs_tx_deferred += pdk_htobl(st->stats_transmit_deferred);
		lp->Statistics.hs_tx_one_collision += pdk_htobl(st->stats_transmit_single_collisions);
		lp->Statistics.hs_tx_more_collisions += pdk_htobl(st->stats_transmit_multiple_collisions);

		lp->Statistics.hs_rx_err_packets += pdk_htobl(st->stats_receive_crc_errors)
				+ pdk_htobl(st->stats_receive_alignment_errors)
				+ pdk_htobl(st->stats_receive_rnr_errors)
				+ pdk_htobl(st->stats_receive_overrun_errors)
				+ pdk_htobl(st->stats_receive_cdt_errors)
				+ pdk_htobl(st->stats_receive_short_frames);
		lp->Statistics.hs_rx_crc += pdk_htobl(st->stats_receive_crc_errors);
		lp->Statistics.hs_rx_alignment_error += pdk_htobl(st->stats_receive_alignment_errors);
		lp->Statistics.hs_rx_overrun += pdk_htobl(st->stats_receive_rnr_errors);
		lp->Statistics.hs_rx_overrun += pdk_htobl(st->stats_receive_overrun_errors);
		lp->Statistics.hs_rx_collisions += pdk_htobl(st->stats_receive_cdt_errors);
		lp->Statistics.hs_rx_runt_frame += pdk_htobl(st->stats_receive_short_frames);

		lp->stats_dump_check = 0;
		status = 1;
	}

	return(status);
}

ZIN_LOCAL void zin_dump_counters_wait(int32_t ppa)
{
	int		count = 1000;
	/* local_adapter_t *lp = local_adapters[ppa]; */

	count = 1000;
	while(--count && !zin_dump_counters_complete(ppa)) {
		pdk_time_delay(1);
	}
	if(!count) {
		DEBUG1("zin_dump_counters_wait(%d): timeout\n", (int)ppa);
	}
}

ZIN_LOCAL void zin_dump_counters(int32_t ppa)
{
	local_adapter_t *lp = local_adapters[ppa];

	if(lp->state != LOCAL_ADAPTER_RUNNING) return;

	if(lp->stats_dump_check) {
		zin_dump_counters_wait(ppa);
	}

	zin_dump_counters_start(ppa);
	zin_dump_counters_wait(ppa);
}




#if ZIN_TIMER_CHECK
ZIN_LOCAL void zin_debug_check(int32_t ppa)
{
	local_adapter_t	*lp;
	uint8_t		isr_value, irq_ctrl;
	uint16_t	tx_status, rx_status, rx_length;
	int		rx_link, rx_dump;
	pdk_buf_t	*bp;

	if ((lp = local_adapters[ppa]) == 0) {
		return;
	}

	isr_value = pdk_reg_get8(ppa, I82558_IS_REG);
	irq_ctrl = pdk_reg_get8(ppa, I82558_IC_REG);

	zin_acquire_rx_lock(lp);

	tx_status = rx_status = -1;
	rx_length = 0;
	rx_link = rx_dump = 0;
	if(lp->txld_first) tx_status = lp->txld_first->tcb.tcb_status;

	if((bp = lp->rfa_head_bp) != NULL) {
		uint8_t *rptr;
		i82558_rfa_t *rfa;
		if(bp->b_link) rx_link = 1;
		rptr = pdk_buf_rptr(bp);
		rfa = (i82558_rfa_t*)(rptr + RFA_OFFSET);
		rx_status = rfa->rfa_status;
		rx_length = pdk_htobs(rfa->rfa_actual_count) & I82558_RFA_AC_MASK;
#if ZIN_BUFFER_MARK
		if(*((uint32_t*)(rptr + sizeof(i82558_rfa_t) + RFA_OFFSET)) != 0) {
			rx_dump = 1;
		}
#endif
	}

	if(lp->tx_queued || rx_status || rx_dump || (lp->to_ten_sec == 0 && lp->to_one_sec == 0)) {
		DEBUG2("timer(%d): ic=0x%02x is=0x%02x txq=%-2d txst=0x%04x rxlk=%d rxst=0x%04x\n",
			(int)ppa, irq_ctrl, isr_value, lp->tx_queued, tx_status, rx_link, rx_status);
	}

#if ZIN_DUMP_PACKETS
	if(rx_dump) {
		uint8_t *p = pdk_buf_rptr(bp) + sizeof(i82558_rfa_t) + RFA_OFFSET;
		if(pdk_hal_debug > 1) {
			pdk_print_ether_packet("Timer", p, rx_length >= 14 ? rx_length : 256);
		}
	}
#endif

	zin_release_rx_lock(lp);
}
#endif




ZIN_LOCAL void zin_timer(int32_t ppa)
{
	local_adapter_t	*lp;

	if ((lp = local_adapters[ppa]) == 0) {
		return;
	}

	lp->to_one_sec += TIMEOUT_INTERVAL;
	if(lp->to_one_sec >= TIMEOUT_ONE_SECOND) {
		lp->to_one_sec = 0;
		if(++lp->to_ten_sec >= 10) lp->to_ten_sec = 0;
#if ZIN_SYNC_MEDIA_CHECK
		zin_media_check(ppa);
#endif
	}

	if(lp->to_ten_sec == 0 && lp->to_one_sec == 0) {
#if 0
		if(lp->led_test < 4 || lp->led_test >= 7)
			lp->led_test = 4;
		else
			lp->led_test++;
		DEBUG1("led(%d) = %d\n", (int)ppa, lp->led_test);
		zin_write_phy_reg(ppa, I82555_LC_REG, lp->led_test);
#endif
	}


	lp->timer_id = NULL;

#if ZIN_TIMER_CHECK
	zin_debug_check(ppa);
#endif


#if 0
	zin_acquire_tx_lock(lp);
	zin_config_update(ppa);
	zin_release_tx_lock(lp);

	zin_media_check(ppa);
#endif


#if !ZIN_SYNC_MEDIA_CHECK
	zin_media_check_start(ppa);
#endif


	lp->timer_id = pdk_timeout(zin_timer, (void*)ppa, TIMEOUT_INTERVAL);
}




pdk_layer_routines_t *
pdk_hal_register(ppa, up, down)
int32_t ppa;
pdk_layer_routines_t *up;
pdk_layer_routines_t *down;
{
	local_adapter_t *lp;

	DEBUG2("pdk_hal_register(%d)\n", (int)ppa);

	if ((lp = local_adapters[ppa]) == 0) {
		DEBUG1("pdk_hal_register(%d):  Local adapter not initialized\n", (int)ppa);
		return(0);
	}

	if (down) {
		DEBUG1("pdk_hal_register(%d):  No layers should be below the HAL\n", (int)ppa);
		return(0);
	}

	if (up) {
		lp->layer_routines.pdk_xmit_complete = up->pdk_xmit_complete;
		lp->layer_routines.pdk_recv_complete = up->pdk_recv_complete;
		lp->layer_routines.pdk_status = up->pdk_status;
	}

	return(&zin_layer_routines);
}

uint32_t
pdk_hal_global_init()
{
	DEBUG3("pdk_hal_global_init\n");

	DEBUG1("sizeof(i82558_eeprom_t)=%d\n", sizeof(i82558_eeprom_t));
	DEBUG1("sizeof(i82558_config_t)=%d\n", sizeof(i82558_config_t));
	DEBUG1("\n");

#if 0
	{
	pdk_lock_t *lock;

	DEBUG1("checking lock stuff:\n");
	if((lock = pdk_lock_alloc(0, PDK_LOCK_INTR)) != 0) {
		pdk_lock_print(lock);
		pdk_lock_acquire(lock);
		pdk_lock_print(lock);
		pdk_lock_release(lock);
		pdk_lock_print(lock);
		pdk_lock_free(lock);
	}
	}
#endif

	return(0);
}

uint32_t
pdk_hal_global_deinit()
{
	DEBUG3("pdk_hal_global_deinit\n");
	return(0);
}




ZIN_LOCAL uint32_t
hal_hdwr_init(uint32_t ppa)
{
	local_adapter_t	*lp;
	int32_t		i;
	uint8_t		*ep;
	int		valid_eeprom;

	lp = local_adapters[ppa];

	DEBUG1("hal_hdwr_init(%d) lp=%x\n", (int)ppa, lp);

#if 0
	/* fixme: Take the board out of snooze and sleep modes */
	pdk_reg_config_put32(ppa, CFDA, 0);
#endif

	/* reset chip */

	zin_reset_chip(ppa);

	DEBUG1("hal_hdwr_init(%d) st=0x%04x cmd=0x%02x ic=0x%02x\n", (int)ppa,
		pdk_reg_get16(ppa, I82558_ST_REG),
		pdk_reg_get8(ppa, I82558_CMD_REG), pdk_reg_get8(ppa, I82558_IC_REG));
	DEBUG1("hal_hdwr_init(%d) ec=0x%04x mc=0x%04x\n", (int)ppa,
		pdk_reg_get16(ppa, I82558_EC_REG),
		pdk_reg_get16(ppa, I82558_MC_REG));

	/* size and read eeprom */

	lp->eeprom_size = 0;
	zin_eeprom_getword(ppa, 0);
	DEBUG1("hal_hdwr_init(%d): eeprom size=%d\n", (int)ppa, lp->eeprom_size);

	lp->eeprom_w = (uint16_t*)&lp->eeprom;
	for(i = 0, ep = (uint8_t*)&lp->eeprom; i < sizeof(lp->eeprom)/2; i++) {
		uint16_t data = zin_eeprom_getword(ppa, i);
		*ep++ = data & 0xff;
		*ep++ = data >> 8;
		if(i < 16 || (data && data != 0xffff)) {
			DEBUG2("eeprom[%2d] = 0x%04x\n", i, lp->eeprom_w[i]);
		}
	}

	/* copy station address */
	pdk_bcopy(lp->eeprom.eeprom_ia, lp->factory_mac_address, PDK_ETHERADDRL);
	if(zin_ether_is_zero(lp->current_mac_address)) {
		pdk_bcopy(lp->factory_mac_address, lp->current_mac_address, PDK_ETHERADDRL);
	}
	DEBUG2("factory mac address (%d) %02X%02X%02X%02X%02X%02X\n", (int)ppa,
		lp->factory_mac_address[0], lp->factory_mac_address[1],
		lp->factory_mac_address[2], lp->factory_mac_address[3],
		lp->factory_mac_address[4], lp->factory_mac_address[5]);


	/* check eeprom validity */
	valid_eeprom = 1;
	if(lp->eeprom.eeprom_controller > I82558_ED_CT_558) {
		DEBUG1("eeprom: controller = %d\n", lp->eeprom.eeprom_controller);
		valid_eeprom = 0;
	}
	if(lp->eeprom.eeprom_pri_phy_address > 31) {
		DEBUG1("eeprom: phy address = %d\n", lp->eeprom.eeprom_pri_phy_address);
		valid_eeprom = 0;
	}

	lp->controller = 557;

	if(valid_eeprom) {
		if(lp->eeprom.eeprom_controller == I82558_ED_CT_558)
			lp->controller = 558;
		/* else if(lp->eeprom.eeprom_controller > I82558_ED_CT_558)
			lp->controller = 599; */

		i = lp->eeprom.eeprom_pri_phy_type & I82558_ED_PHY_DC_MASK;
		DEBUG1("phy %s at address %d\n", i < zin_phy_types ? zin_phy_type[i] : "???", lp->eeprom.eeprom_pri_phy_address);

		lp->phy_addr = lp->eeprom.eeprom_pri_phy_address;

		lp->phy_serial = lp->eeprom.eeprom_pri_phy_type & I82558_ED_PHY_SERIAL ? 1 : 0;
	}
	else {
		uint16_t device;
		uint8_t rev;

		device = pdk_reg_config_get16(ppa, 0x02);
		rev = pdk_reg_config_get8(ppa, 0x08);

		if(device == INTEL_82557_DEVICE_ID && rev >= 4)
			lp->controller = 558;
		if(device == INTEL_82559ER_DEVICE_ID)
			lp->controller = 558;

		DEBUG1("incomplete eeprom configuration, device=0x%04x, rev=0x%02x\n", device, rev);

		lp->phy_serial = 0;
		if(zin_detect_phy(ppa)) {
			lp->phy_addr = 1;
		}
	}

	if(lp->controller != 557) {
		/*
		 * If MWI is enabled in the PCI configuration, and there
		 * is a valid cacheline size (8 or 16 dwords), then tell
		 * the board to turn on MWI.
		 */
#if 0
		if (val & PCIM_CMD_MWRICEN &&
		    pci_read_config(dev, PCIR_CACHELNSZ, 1) != 0)
			lp->flags |= DRV_FLAG_MWI_ENABLE;
#endif

		/* we use extended TCBs */
		lp->la_tbd_array_start = 2;

		/* enable reception of long frames for VLAN */
		lp->recv_giants = 1;
	}


	DEBUG1("silicon %d (%d), pwa %02x%02x%02x-%03d, connector type%s%s%s%s\n",
		lp->controller, lp->eeprom.eeprom_controller,
		lp->eeprom.eeprom_pwa[1], lp->eeprom.eeprom_pwa[0],
		lp->eeprom.eeprom_pwa[3], lp->eeprom.eeprom_pwa[2],
		lp->eeprom.eeprom_connectors & I82558_ED_CONN_RJ45 ? " RJ45" : "",
		lp->eeprom.eeprom_connectors & I82558_ED_CONN_BNC ? " BNC" : "",
		lp->eeprom.eeprom_connectors & I82558_ED_CONN_AUI ? " AUI" : "",
		lp->eeprom.eeprom_connectors & I82558_ED_CONN_MII ? " MII" : "");


#if 0
	DEBUG1("board assy %4.4x%2.2x-%3.3d,  connectors",
		lp->eeprom[8], lp->eeprom[9]>>8, lp->eeprom[9] & 0xff);
	for(i = 0; zin_connectors[i]; i++)
		if(lp->eeprom[5] & (1<<i)) {
			DEBUG1(" %s", zin_connectors[i]);
		}
	DEBUG1("\n");
	i = (lp->eeprom[6]>>8) & 0x3f;
	DEBUG1("phy %s at address %d\n", i < zin_phy_types ? zin_phy_type[i] : "???", lp->eeprom[6] & 0x1f);

	lp->phy_addr = lp->eeprom[6] & 0x1f;

	lp->phy_serial = (lp->eeprom[6] & 0x8000) ? 1 : 0;
#endif

	for(i = 0; i < 4; i++) {
		DEBUG1("phy[%2d] = 0x%04x\n", i, zin_read_phy_reg(ppa, i));
	}

	zin_media_init(ppa);

	DEBUG1("hal_hdwr_init(%d) exit\n", (int)ppa);

	/* fixme: do this before media_init? */
	lp->hardware_initialized = 1;

	return(0);
}




uint32_t
pdk_hal_init(ppa)
int32_t ppa;
{
	local_adapter_t *lp;
	pdk_desc_handle_t *handlep;
	void *vaddr;
	uint32_t paddr;
	int32_t err;

	DEBUG3("pdk_hal_init(%d): enter\n", (int)ppa);

	/* If there is already a local adapter structure */
	if ((lp = local_adapters[ppa]) != 0) {
		pdk_hal_deinit(ppa);
	}

	/* Get a local adapter structure */
	local_adapters[ppa] = (local_adapter_t *)pdk_mem_zalloc(ppa, sizeof(local_adapter_t));

	if ((lp = local_adapters[ppa]) == 0) {
		return(PDK_NO_MEM);
	}

	lp->ppa = ppa;
	lp->link_status_up = 0;

/* fixme: check out */
	lp->free_tx_descriptor_count = zin_num_txd;
	lp->tx_lwm = zin_num_txd;
	lp->rx_lwm = zin_num_rxd;
	lp->tx_pending_threshold = (zin_num_txd + 15)>>4;


	/* Allocate locks */

	if ((lp->rx_descriptor_lock = pdk_lock_alloc(ppa, PDK_LOCK_INTR)) == 0) {
		pdk_hal_deinit(ppa);
		return(PDK_SYS_ERR);
	}

	if ((lp->tx_descriptor_lock = pdk_lock_alloc(ppa, PDK_LOCK_INTR)) == 0) {
		pdk_hal_deinit(ppa);
		return(PDK_SYS_ERR);
	}


	/* get configuration and control block storage */

	handlep = pdk_mem_desc_alloc(ppa, 1, sizeof(i82558_union_t)*2);
	if (handlep == 0) {
		pdk_hal_deinit(ppa);
		return(PDK_SYS_ERR);
	}

	lp->cb_buffer_handlep = handlep;

	handlep = pdk_mem_desc_get_pmap(handlep, &vaddr, &paddr, &err);
	if (err) {
		pdk_hal_deinit(ppa);
		return(PDK_SYS_ERR);
	}
	lp->cb_buffer = (i82558_union_t*)vaddr;
	lp->cb_buffer_paddr = paddr;

	DEBUG2("pdk_hal_init(%d): gen cb 0x%x 0x%x\n", (int)ppa, vaddr, paddr);


	/* get statistics buffer */

	handlep = pdk_mem_desc_alloc(ppa, 1, sizeof(i82558_stats_t)+256);
	if (handlep == 0) {
		pdk_hal_deinit(ppa);
		return(PDK_SYS_ERR);
	}
	lp->stats_buffer_handlep = handlep;

	handlep = pdk_mem_desc_get_pmap(handlep, &vaddr, &paddr, &err);
	if (err) {
		pdk_hal_deinit(ppa);
		return(PDK_SYS_ERR);
	}
	lp->stats_buffer = (i82558_stats_t*)vaddr;
	lp->stats_buffer_paddr = paddr;

	DEBUG2("pdk_hal_init(%d): stats 0x%x 0x%x\n", (int)ppa, vaddr, paddr);


	/* setup rfa list */

	lp->rfa_head_bp = NULL;
	lp->rfa_tail_bp = NULL;
	lp->rfa_head_paddr = 0;


	/* transmit control blocks */

	handlep = pdk_mem_desc_alloc(ppa, 1, zin_num_txd * sizeof(txld_t));
	if (handlep == 0) {
		pdk_hal_deinit(ppa);
		return(PDK_SYS_ERR);
	}

	lp->txld_space_handlep = handlep;

	handlep = pdk_mem_desc_get_pmap(handlep, &vaddr, &paddr, &err);
	if (err) {
		pdk_hal_deinit(ppa);
		return(PDK_SYS_ERR);
	}
	lp->txld_space = (txld_t*)vaddr;
	lp->txld_space_paddr = paddr;

	DEBUG2("pdk_hal_init(%d): tx cb 0x%x 0x%x\n", (int)ppa, vaddr, paddr);

	err = hal_hdwr_init(ppa);


	DEBUG3("pdk_hal_init(%d): exit %d\n", (int)ppa, err);

	return(err);
}




uint32_t
pdk_hal_deinit(ppa)
int32_t ppa;
{
	local_adapter_t *lp;

	DEBUG2("pdk_hal_deinit(%d): enter\n", (int)ppa);

	if ((lp = local_adapters[ppa]) == 0)
		return(PDK_OUT_STATE);

	if (lp->state == LOCAL_ADAPTER_RUNNING) {
		DEBUG1("pdk_hal_deinit(%d):  Adapter is running - stopping before deiniting\n", (int)ppa);
		pdk_hal_ctl(ppa, PDK_HAL_STOP_AND_RESET, 0);
	}

	/* Clear lp from local adapter structure */
	local_adapters[ppa] = 0;

#ifdef PDK_HAL_SPEEDO
	if (lp->hal_speedometer_timeout_id != 0) {
		pdk_untimeout(lp->hal_speedometer_timeout_id);
		lp->hal_speedometer_timeout_id = 0;
	}
#endif

	if(lp->cb_buffer_handlep) {
		pdk_mem_desc_free(lp->cb_buffer_handlep);
	}

	if(lp->stats_buffer_handlep) {
		pdk_mem_desc_free(lp->stats_buffer_handlep);
	}

	if(lp->txld_space_handlep) {
		pdk_mem_desc_free(lp->txld_space_handlep);
	}

	if(lp->rx_descriptor_lock) {
		pdk_lock_free(lp->rx_descriptor_lock);
	}

	if(lp->tx_descriptor_lock) {
		pdk_lock_free(lp->tx_descriptor_lock);
	}

	pdk_mem_free(ppa, (uchar *) lp, sizeof(local_adapter_t));

	DEBUG2("pdk_hal_deinit(%d): exit\n", (int)ppa);
	return(0);
}




ZIN_LOCAL int32_t
pdk_hal_first_start(ppa)
int32_t ppa;
{
	local_adapter_t *lp = local_adapters[ppa];
	uint32_t err;

	DEBUG2("pdk_hal_first_start(%d): enter lp=%x\n", (int)ppa, lp);

	/* Init hardware if not done already */

/* $$$ fixme use lp->hardware_initialized */
#if 0
	if (!lp->factory_mac_address[0] && !lp->factory_mac_address[1]
		&& !lp->factory_mac_address[2])	/* Check for OID == 0 */
#endif

	if(!lp->hardware_initialized) {
		err = hal_hdwr_init(ppa);
		if (err) {
			DEBUG2("pdk_hal_first_start(%d): hdwr_init error\n", (int)ppa);
			return(err);
		}
	} else {

#if 0
		/* Take the board out of snooze and sleep modes */
		pdk_reg_config_put32(ppa, CFDA, 0);	/* Needed for hot swap */

		/* Reset chip */
		pdk_reg_put32(ppa, CSR0, CSR0_SWR);
		pdk_time_delay(2);
#endif
	}

	DEBUG2("pdk_hal_first_start(%d): exit lp=%x\n", (int)ppa, lp);

	return(0);
}




/* start the interface. called with tx and rx lock held */

ZIN_LOCAL uint32_t
pdk_hal_start(ppa, do_loopback)
int32_t ppa;
int32_t do_loopback;
{
	local_adapter_t *lp;
	uint32_t	err;
#if 0
	i82558_config_t	*ccb;
#endif
	txld_t		*txp;
	int		i;


	DEBUG1("pdk_hal_start(%d): enter, do_loopback=%d\n", (int)ppa, do_loopback);

	if ((lp = local_adapters[ppa]) == 0)
		return(PDK_OUT_STATE);

	if (lp->state == LOCAL_ADAPTER_RELEASED) {
		DEBUG2("pdk_hal_start(%d): RELEASED, will not be started\n", (int)ppa);
		return(0);
	}

	if (lp->state == LOCAL_ADAPTER_ALLOCATED) {
		if ((err = pdk_hal_first_start(ppa)) != 0) {
			DEBUG1("pdk_hal_start(%d): first_start error\n", (int)ppa);
			return(err);
		}
		lp->state = LOCAL_ADAPTER_STOPPED;
	}

	if (lp->state != LOCAL_ADAPTER_STOPPED) {
		DEBUG1("pdk_hal_start(%d): Out of state(%d)\n", (int)ppa, lp->state);
		return(PDK_OUT_STATE);
	}

#if 0
	if (lp->rx_enqueue_here != lp->rx_dequeue_here) {
		DEBUG1("pdk_hal_start(%d): Out of state, receive buffers present\n", (int)ppa);
		return(PDK_OUT_STATE);
	}
	if (lp->tx_enqueue_here != lp->tx_dequeue_here) {
		DEBUG1("pdk_hal_start(%d): Out of state, transmit buffers present\n", (int)ppa);
		return(PDK_OUT_STATE);
	}
#endif

	/* reset chip */

	zin_reset_chip(ppa);

	/* init CBL and RFA address mode */

	zin_scb_wait(ppa);
	pdk_reg_put32(ppa, I82558_GP_REG, 0);
	zin_scb_cmd(ppa, I82558_CMD_CU_BASE);
	zin_scb_wait(ppa);
	zin_scb_cmd(ppa, I82558_CMD_RU_BASE);


	/* load statistics buffer physical address */

	zin_scb_wait(ppa);
	pdk_reg_put32(ppa, I82558_GP_REG, lp->stats_buffer_paddr);
	zin_scb_cmd(ppa, I82558_CMD_CU_DUMP_LA);


#if 0
	/* config control block */

	ccb = &lp->config_template;
	zin_build_config(lp, &lp->config_template);

	if(zin_sync_command(ppa, (char*)ccb, sizeof(i82558_config_t))) {
		pdk_hal_stop(ppa);
		return(PDK_HW_ERR);
	}
#endif


	/* transmit control blocks */

	txp = lp->txld_space;
	pdk_bzero(txp, sizeof(txld_t) * zin_num_txd);

	for (i = 0; i < zin_num_txd; i++) {
		txld_t *ttxp = &txp[i];
		txld_t *ntxp = &txp[(i + 1) % zin_num_txd];

		ttxp->tcb.tcb_status = pdk_htobs(I82558_CB_STATUS_C | I82558_CB_STATUS_OK);
		ttxp->tcb.tcb_command = pdk_htobs(I82558_CB_COMMAND_NOP);
		ttxp->tcb.tcb_link_addr = pdk_htobl(zin_tcb_paddr(lp, &ntxp->tcb));
		ttxp->tcb.tcb_tbd_array_addr = 	pdk_htobl(zin_tcb_paddr(lp, &ttxp->tbd[lp->la_tbd_array_start]));
		ttxp->next = ntxp;

		DEBUG3("txp%02d  0x%08x 0x%08x  0x%08x 0x%08x 0x%08x\n",
			i, ttxp, lp->txld_space_paddr + ((char*)ttxp) - ((char*)txp),
			ttxp->tcb.tcb_link_addr, ttxp->tcb.tcb_tbd_array_addr, ttxp->next);
	}

	lp->txld_first = lp->txld_last = lp->txld_space;
	lp->tx_queued = 0;


	/* configure chip */

	if(zin_config_update(ppa)) {
		pdk_hal_stop(ppa);
		return(PDK_HW_ERR);
	}


	/* the receive frame area is set up when we get the buffers */


#if 0
				move this to before descriptor ring setup
	if (do_loopback) {
figure out lock situation!
		if (zxe_loopback_test(ppa) != 0) {
			DEBUG1("pdk_hal_start(%d): Loopback failed\n", (int)ppa);
			pdk_hal_stop(ppa);

			zxe_operational_led(ppa, 0);
			zxe_internal_fault_led(ppa, 1);
			return(PDK_HW_ERR);
		}
	}
#endif

	/* setup media */

	if(zin_media_config(ppa)) {
		DEBUG1("pdk_hal_start(%d): Media configuration failed\n", (int)ppa);
		pdk_hal_stop(ppa);
		return(PDK_HW_ERR);
	}

	lp->state = LOCAL_ADAPTER_RUNNING;

	/* start timer */

	if((lp->timer_id = pdk_timeout(zin_timer, (void*)ppa, TIMEOUT_DELAY)) == NULL) {
		pdk_hal_stop(ppa);
		/* fixme: error code */
		return(PDK_NO_MEM);
	}


	/* enable interrupts */

	pdk_reg_put8(ppa, I82558_IC_REG, 0);

	DEBUG1("pdk_hal_start(%d): exit\n", (int)ppa);
	return(0);
}




/*
 * Restart the adapter.  Keep the receive buffers.  Recycle any transmit
 * buffers.
 */
int32_t
pdk_hal_restart(ppa)
int32_t ppa;
{
	DEBUG2("pdk_hal_restart(%d)\n", (int)ppa);
#if 0
	local_adapter_t *lp;
	desc_t *rx_dequeue_here;
	desc_t *rx_enqueue_here;
	desc_t *tdd;
	uint32_t rx_num_bufs;
	int32_t temp_debug;
	int32_t err;

	DEBUG2("pdk_hal_restart(%d)\n", (int)ppa);

	if ((lp = local_adapters[ppa]) == 0)
		return(PDK_OUT_STATE);

	if (lp->state != LOCAL_ADAPTER_RUNNING) {
		DEBUG1("pdk_hal_restart(%d): Out of state(%d)\n", (int)ppa, lp->state);
		return(PDK_OUT_STATE);
	}

#if 0
	/* Stop the adapter */
	SET_CSR7(ppa,0);
	pdk_reg_put32(ppa, CSR0, CSR0_SWR);
	pdk_time_delay(2);
#endif

	/* Save the receive buffers */
	zin_acquire_rx_lock(lp);
	zin_acquire_tx_lock(lp);

	rx_dequeue_here = lp->rx_dequeue_here;
	rx_enqueue_here = lp->rx_enqueue_here;
	rx_num_bufs = lp->rx_num_bufs;
	lp->rx_dequeue_here = lp->rx_enqueue_here;

	DEBUG3("pdk_hal_restart(%d): 0x%x 0x%x 0x%x\n", (int)ppa, rx_dequeue_here, rx_enqueue_here, rx_num_bufs);

	pdk_hal_stop(ppa);

	tdd = rx_dequeue_here;
	while (tdd != rx_enqueue_here) {
		tdd->RDES0 = 0;
		tdd = tdd->NEXT;
	}

	if ((temp_debug = pdk_hal_debug) != 0)
		pdk_hal_debug = 1;

	if ((err = pdk_hal_start(ppa, 0)) != 0) {
		DEBUG1("pdk_hal_restart(%d): Cannot start\n", (int)ppa);
		pdk_hal_debug = temp_debug;
		zin_release_tx_lock(lp);
		zin_release_rx_lock(lp);
		return(err);
	}
	pdk_hal_debug = temp_debug;

	/*
	 * Restore Receives after initialization
	 */
	while (rx_dequeue_here != lp->rx_queue) {
		rx_enqueue_here->BP = rx_dequeue_here->BP;
		rx_enqueue_here->RDES3 = rx_dequeue_here->RDES3;
		rx_enqueue_here->RDES2 = rx_dequeue_here->RDES2;
		rx_enqueue_here->RDES1 = rx_dequeue_here->RDES1 & ~RDES1_RER;
		if (rx_enqueue_here->NEXT == lp->rx_queue)
			rx_enqueue_here->RDES1 |= RDES1_RER;
		rx_enqueue_here = rx_enqueue_here->NEXT;
		rx_dequeue_here = rx_dequeue_here->NEXT;
	}
	lp->rx_enqueue_here = rx_enqueue_here;
	lp->rx_num_bufs = rx_num_bufs;

	tdd = rx_dequeue_here;
	while (tdd != rx_enqueue_here) {
		tdd->RDES0 = RDES0_OWN;
		tdd = tdd->NEXT;
	}
	DEBUG3("pdk_hal_restart(%d):  CSR5 0x%x\n", (int)ppa, pdk_reg_get32(ppa, CSR5));

	zin_release_tx_lock(lp);
	zin_release_rx_lock(lp);

#ifdef PDK_HAL_SPEEDO
	if (lp->speedometer_poll_rate != 0) {
		if (lp->hal_speedometer_timeout_id == 0)
			lp->hal_speedometer_timeout_id = pdk_timeout(hal_speedometer_timeout, (void *)ppa, HAL_SPEEDOMETER_TICKS);
	}
#endif
	
#endif
	return(0);
}



/* stop the interface. called with tx and rx lock held */

ZIN_LOCAL uint32_t pdk_hal_stop(int32_t ppa)
{
	local_adapter_t *lp;
	txld_t *txp;
	int i;
	pdk_buf_t *bp;

	DEBUG2("pdk_hal_stop(%d) enter\n", (int)ppa);

	if ((lp = local_adapters[ppa]) == 0) {
		DEBUG2("pdk_hal_stop(%d) no lp\n", (int)ppa);
		return(PDK_OUT_STATE);
	}

	if (lp->state == LOCAL_ADAPTER_RELEASED) {
		DEBUG2("znb%d is RELEASED, already stopped\n",ppa);
		return(0);				/* Don't stop this one again */
	}

	if((lp->state != LOCAL_ADAPTER_RUNNING) && (lp->state != LOCAL_ADAPTER_STOPPED)) {
		DEBUG2("pdk_hal_stop(%d): Out of state(%d)\n", (int)ppa, lp->state);
		return(PDK_OUT_STATE);
	}

#if 0
	/*
	 * Cancel stats updater.
	 */
	untimeout(zin_tick, sc, lp->stat_ch);
#endif
	if(lp->timer_id) {
		pdk_untimeout(lp->timer_id);
		lp->timer_id = NULL;
	}

#if 0
	zin_acquire_tx_lock(lp);
#endif

	/* reset chip */

	zin_reset_chip(ppa);

	lp->state = LOCAL_ADAPTER_STOPPED;

#if 0
	zin_release_tx_lock(lp);
#endif

	/* fixme: passify phy */

	zin_media_init(ppa);

	if(lp->link_status_up) {
		(*lp->layer_routines.pdk_status)(ppa, PDK_STATUS_LINK_DOWN, 0);
	}

#if 0
	zin_acquire_tx_lock(lp);
#endif

	lp->link_status_up = 0;
	lp->current_media_type = lp->configured_media_type;

	/* release transmit buffers */

	DEBUG2("pdk_hal_stop(%d) release tx buffers\n", (int)ppa);
	DEBUG3("pdk_hal_stop(%d) tx_q=%d first=0x%x last=0x%x\n", (int)ppa, lp->tx_queued, lp->txld_first, lp->txld_last);
	for(txp = lp->txld_space, i = 0; txp && (i < zin_num_txd); i++, txp++) {
		DEBUG4("pdk_hal_stop(%d) txp=0x%x\n", (int)ppa, txp);
		if(txp->bp != NULL) {
			DEBUG4("pdk_hal_stop(%d) bp=0x%x txp=0x%x\n", (int)ppa, txp->bp, txp);
			pdk_buf_unmap(txp->bp);
			(*lp->layer_routines.pdk_xmit_complete)(lp->ppa, txp->bp, PDK_XMIT_STATUS_FLUSHING);
			txp->bp = NULL;
		}
	}
	lp->tx_queued = 0;
	/* fixme: null out txld_first/last? */
	lp->free_tx_descriptor_count = zin_num_txd;

	/* release receive buffers */

	DEBUG2("pdk_hal_stop(%d) release rx buffers\n", (int)ppa);
	while((bp = lp->rfa_head_bp) != NULL) {
		DEBUG4("pdk_hal_stop(%d) bp=0x%x\n", (int)ppa, bp);
		lp->rfa_head_bp = bp->b_link;
		bp->b_link = NULL;
		pdk_buf_unmap(bp);
		pdk_buf_free(bp);
		}
	lp->rfa_head_bp = NULL;
	lp->rfa_tail_bp = NULL;
	lp->rfa_head_paddr = 0;
	lp->rx_num_bufs = 0;

#if 0
	zin_release_tx_lock(lp);
#endif

	DEBUG1("pdk_hal_stop(%d): exit\n", (int)ppa);
	return(0);
}




uint32_t
pdk_hal_iac(ppa)
int32_t ppa;
{
	local_adapter_t *lp;
	uint8_t isr_value;

	lp = local_adapters[ppa];

	if ((lp == 0) || (lp->state != LOCAL_ADAPTER_RUNNING)) {
		return(0);
	}

	if((isr_value = pdk_reg_get8(ppa, I82558_IS_REG)) != 0) {

		/* disable interrupts */
		pdk_reg_put8(ppa, I82558_IC_REG, I82558_IC_MASK);

		/* acknowledge */
		pdk_reg_put8(ppa, I82558_IS_REG, isr_value);

		DEBUG4("pdk_hal_iac(%d): isr_value 0x%x\n", (int)ppa, isr_value);

#if 0
		if (isr_value & CSR5_RI)
			lp->rx_icount++;
		if (isr_value & CSR5_TI)
			lp->tx_icount++;
		if (isr_value & ~(CSR5_RI|CSR5_TI))
			lp->other_icount++;
#endif
	}

	return((uint32_t)isr_value);
}




uint32_t pdk_hal_intr(int32_t ppa, uint32_t isr_value)
{
	local_adapter_t *lp;
/*	uint16_t phy_isr_value; */

	DEBUG3("pdk_hal_intr(%d): isr_value 0x%x\n", (int)ppa, isr_value);

	if ((lp = local_adapters[ppa]) != 0) {

		if(isr_value & (I82558_IS_FR | I82558_IS_RNR)) {
			zin_process_recv_complete(ppa, lp, isr_value);
		}

		if(isr_value & (I82558_IS_CXTNO | I82558_IS_CNA)) {
			DEBUG3("pdk_hal_intr(%d): tx interrupt\n", (int)ppa);
			zin_acquire_tx_lock(lp);
			zin_process_xmit_complete(lp);
			zin_release_tx_lock(lp);
		}

		if(isr_value & I82558_IS_MDI) {
			zin_mdi_async_read_complete(ppa, lp);
		}

		/* enable interrupts */
		pdk_reg_put8(ppa, I82558_IC_REG, 0);
	}
	else {
/* fixme: turn off interrupt in hw ??? */
		DEBUG2("pdk_hal_intr(%d): lp==NULL?\n", (int)ppa);
#if 0
		pdk_reg_put32(ppa, CSR7, CSR7_CFG);
#endif
		}

	return(0);
}




uint32_t
pdk_hal_xmit(ppa, bp)
int32_t ppa;
pdk_buf_t *bp;
{
	local_adapter_t	*lp;
	pdk_map_t	*mp;
	uint32_t	length;
	uint32_t	paddr;
	int32_t		err;
	txld_t		*txp;
	int		fragment;


	DEBUG3("pdk_hal_xmit(%d): enter\n", (int)ppa);

	if ((lp = local_adapters[ppa]) == 0) {
		DEBUG2("pdk_hal_xmit(%d): local_adapter is zero\n", (int)ppa);
		return(PDK_LINK_DOWN);
	}

	zin_acquire_tx_lock(lp);

/* fixme: did i comment this out? when the link detection wasn't working? */
#if 1
	if(!lp->link_status_up) {
		DEBUG2("pdk_hal_xmit(%d): local_adapter link status down\n", (int)ppa);
		zin_release_tx_lock(lp);
		return(PDK_LINK_DOWN);
	}
#endif

	if(lp->tx_queued + 1 >= zin_num_txd) {
		DEBUG3("pdk_hal_xmit(%d): ring full\n", (int)ppa);
		zin_release_tx_lock(lp);
		return(PDK_NO_ROOM);
	}

#if ZIN_DUMP_BUFFERS
	pdk_buf_dump("tx_bp", bp);
#endif
#if ZIN_DUMP_PACKETS
	pdk_print_ether_packet("TX:", pdk_buf_rptr(bp), 999);
#endif

	/* Map this pdk transmit buffer */
	if ((mp = pdk_buf_map(bp, PDK_BUF_XMIT, &err)) == 0) {
		DEBUG2("pdk_hal_xmit(%d): pdk_buf_map failed - err %d\n", (int)ppa, err);
		zin_release_tx_lock(lp);
		return(err);
	}

	txp = lp->txld_last->next;
	DEBUG3("pdk_hal_xmit(%d): txp=0x%x last=0x%x\n", (int)ppa, txp, lp->txld_last);
	fragment = 0;
	txp->length = 0;

	while(mp) {
		mp = pdk_buf_get_pmap(mp, &paddr, &length, &err);
		if (err) {
			DEBUG2("pdk_hal_xmit(%d): pdk_buf_get_pmap failed\n", (int)ppa);
			pdk_buf_unmap(bp);
			zin_release_tx_lock(lp);
			return(err);
		}

		DEBUG4("pdk_hal_xmit(%d): frag%d 0x%08x %d\n", (int)ppa, fragment, paddr, length);

		txp->tbd[fragment].tbd_addr = pdk_htobl(paddr);
		txp->tbd[fragment].tbd_size = pdk_htobl(length);
		txp->length += length;
		fragment++;

		/* fixme: pullup */
		if(fragment >= (int)TXLD_NUM_TBD) break;
	}

	txp->bp = bp;
	txp->tcb.tcb_tbd_number = fragment;
	txp->tcb.tcb_status = 0;
	txp->tcb.tcb_tx_threshold = 64; /* 512 bytes */
	txp->tcb.tcb_command = pdk_htobs(I82558_CB_COMMAND_SMTX | I82558_CB_COMMAND_FMTX
		| I82558_CB_COMMAND_S | I82558_CB_COMMAND_I);

	/* guarantee in order... */
	SYS_INSTR_SYNC();

	lp->txld_last->tcb.tcb_command &= ~pdk_htobs(I82558_CB_COMMAND_S);
	lp->txld_last = txp;

	if(lp->tx_queued++ == 0)
		lp->txld_first = txp;

	DEBUG3("pdk_hal_xmit(%d): tx_q=%d first=0x%x last=0x%x\n", (int)ppa, lp->tx_queued, lp->txld_first, lp->txld_last);

	zin_scb_wait(ppa);
	zin_scb_cmd(ppa, I82558_CMD_CU_RESUME);

	zin_release_tx_lock(lp);

	return(0);
}




ZIN_LOCAL void
zin_process_xmit_complete(lp)
local_adapter_t *lp;
{
	txld_t		*txp;

	DEBUG3("zin_process_xmit_complete(%d): enter tx_q=%d first=0x%x last=0x%x\n", lp->ppa, lp->tx_queued, lp->txld_first, lp->txld_last);
	txp = lp->txld_first;

	while(lp->tx_queued && (txp->tcb.tcb_status & pdk_htobs(I82558_CB_STATUS_C))) {

		DEBUG4("zin_process_xmit_complete(%d): txp=0x%x\n", lp->ppa, txp);
		if(txp->bp != NULL) {
			pdk_buf_unmap(txp->bp);
			(*lp->layer_routines.pdk_xmit_complete)(lp->ppa, txp->bp, 0);
			txp->bp = NULL;
			lp->Statistics.hs_tx_bytes += txp->length;
			lp->Statistics.hs_tx_packets++;
		}
		lp->tx_queued--;
		txp = txp->next;
	}

	lp->txld_first = txp;
	DEBUG3("zin_process_xmit_complete(%d): exit  tx_q=%d first=0x%x last=0x%x\n", lp->ppa, lp->tx_queued, lp->txld_first, lp->txld_last);
}




/* zin_enqueue_recv_buf -- enqueue a receive buffer */

ZIN_LOCAL uint32_t
zin_enqueue_recv_buf(int32_t ppa, pdk_buf_t *bp, local_adapter_t *lp)
{
	pdk_map_t	*mp;
	uint32_t	paddr, length;
	int32_t		err;
	uint8_t		*vaddr;
	i82558_rfa_t	*rfa, *tail_rfa;
	int		start_ru;

#if ZIN_DUMP_BUFFERS
	pdk_buf_dump("rxnew", bp);
#endif

	/* First check that adapter is still running */
	if (lp->state != LOCAL_ADAPTER_RUNNING) {
		DEBUG2("pdk_hal_recv(%d): adapter not running\n", (int)ppa);
		return(PDK_OUT_STATE);
	}

	/* Map this pdk receive buffer */
	if ((mp = pdk_buf_map(bp, PDK_BUF_RECV, &err)) == 0) {
		DEBUG1("pdk_hal_recv(%d): pdk_buf_map failed\n", (int)ppa);
		return(err);
	}

	mp = pdk_buf_get_pmap(mp, &paddr, &length, &err);
	if (err) {
		pdk_buf_unmap(bp);
		DEBUG1("pdk_hal_recv(%d): pdk_buf_get_pmap failed\n", (int)ppa);
		return(err);
	}

	if (mp != 0) {
		pdk_buf_unmap(bp);
		DEBUG1("pdk_hal_recv(%d): Multi fragmented receive buffer\n", (int)ppa);
		return(PDK_BAD_PARAM);
	}

	vaddr = pdk_buf_rptr(bp);

	rfa = (i82558_rfa_t*)(vaddr + RFA_OFFSET);
	rfa->rfa_size = pdk_htobs(length - sizeof(i82558_rfa_t) - RFA_OFFSET);
	rfa->rfa_status = 0;
	rfa->rfa_control = pdk_htobs(I82558_RFA_CONTROL_EL);
	rfa->rfa_actual_count = 0;
	rfa->rfa_link_addr = I82558_CB_NULL;
	rfa->rfa_rbd_addr = I82558_CB_NULL;
#if ZIN_BUFFER_MARK
	*((uint32_t*)(vaddr + sizeof(i82558_rfa_t) + RFA_OFFSET + 0)) = 0;
	*((uint32_t*)(vaddr + sizeof(i82558_rfa_t) + RFA_OFFSET + 4)) = 0;
	*((uint32_t*)(vaddr + sizeof(i82558_rfa_t) + RFA_OFFSET + 8)) = 0x86868686;
	*((uint32_t*)(vaddr + sizeof(i82558_rfa_t) + RFA_OFFSET + 12)) = 0;
#endif

	DEBUG4("pdk_hal_recv(%d): head=0x%08x tail=0x%08x\n", (int)ppa, lp->rfa_head_bp, lp->rfa_tail_bp);
	DEBUG4("pdk_hal_recv(%d): bp=0x%08x pa=0x%08x len=%d rfa=0x%08x\n", (int)ppa, bp, paddr, length, rfa);

	start_ru = 0;
	bp->b_link = NULL;

	if (lp->rfa_head_bp != NULL) {
		/* append to end of list */
		lp->rfa_tail_bp->b_link = bp;
		vaddr = pdk_buf_rptr(lp->rfa_tail_bp);
		tail_rfa = (i82558_rfa_t*)(vaddr + RFA_OFFSET);
		tail_rfa->rfa_link_addr = pdk_htobl(paddr + RFA_OFFSET);
		/* guarantee in order... */
		SYS_INSTR_SYNC();
		tail_rfa->rfa_control = 0;
	}
	else {
		lp->rfa_head_bp = bp;
		lp->rfa_head_paddr = paddr + RFA_OFFSET;
		start_ru = 1;
	}
	lp->rfa_tail_bp = bp;

	lp->rx_num_bufs++;
	if (lp->rx_num_bufs > lp->rx_hwm)
		lp->rx_hwm = lp->rx_num_bufs;

	if(start_ru) {
		/* note: we hold the rx lock and acquire the tx lock. make
		   sure we never hold the tx lock and acquire the rx lock
		   or we will die in smp deadlock hell. */
		DEBUG4("pdk_hal_recv(%d): starting ru\n", (int)ppa);
		zin_acquire_tx_lock(lp);
		zin_scb_wait(ppa);
		pdk_reg_put32(ppa, I82558_GP_REG, lp->rfa_head_paddr);
		zin_scb_cmd(ppa, I82558_CMD_RU_START);
		zin_release_tx_lock(lp);
		}

#if 0
/* fixme: ??? */
	descriptor_flush(lp, H_READ);
#endif

	return(0);
}




/*
 * pdk_hal_recv -- supply a receive buffer to be enqueued for DMA
 *
 * gets called from pal receive code and sal buffer mgmt code
 *
 * this assumes that bp->b_len is set up (PDK_ETHERRECV_SIZE)
 */

uint32_t
pdk_hal_recv(ppa, bp)
int32_t ppa;
pdk_buf_t *bp;
{
	local_adapter_t	*lp;
	int32_t		err;

	if ((lp = local_adapters[ppa]) == 0) {
		return(PDK_OUT_STATE);
	}

	zin_acquire_rx_lock(lp);

	err = zin_enqueue_recv_buf(ppa, bp, lp);

	zin_release_rx_lock(lp);

	return(err);
}




ZIN_LOCAL void
zin_process_recv_complete(int32_t ppa, local_adapter_t *lp, uint32_t isr_value)
{
	pdk_buf_t	*bp;
	uint8_t		*rptr;
	i82558_rfa_t	*rfa;
	uint32_t	frame_size;

	zin_acquire_rx_lock(lp);

	DEBUG4("zin_process_recv_complete(%d): rfa_head=0x%08x\n", (int)ppa, lp->rfa_head_bp);

	while((bp = lp->rfa_head_bp) != NULL) {

		if(bp->b_link == NULL) {
			/* last in list, keep for now */
			break;
		}

		rptr = pdk_buf_rptr(bp);
		rfa = (i82558_rfa_t*)(rptr + RFA_OFFSET);

		DEBUG4("zin_process_recv_complete(%d): bp=0x%08x rfa=0x%08x 0x%08x st=0x%04x 0x%04x\n", (int)ppa, bp, rfa, lp->rfa_head_paddr, rfa->rfa_status, pdk_htobs(rfa->rfa_status));

		if(!(rfa->rfa_status & pdk_htobs(I82558_RFA_STATUS_C)))
			/* not fully baked yet */
			break;

		lp->rfa_head_bp = bp->b_link;
		if((lp->rfa_head_paddr = pdk_htobl(rfa->rfa_link_addr)) == -1)
			/* should never happen as we always leave at
			   least one receive area buffer on the list */
			lp->rfa_head_paddr = 0;

		zin_release_rx_lock(lp);

		bp->b_link = NULL;
		pdk_buf_unmap(bp);

		frame_size = pdk_htobs(rfa->rfa_actual_count) & I82558_RFA_AC_MASK;
		DEBUG4("zin_process_recv_complete(%d): bp=0x%08x va=0x%08x rfa=0x%08x fs=%d\n", (int)ppa, bp, rptr, rfa, frame_size);

#if 0
		if((rfa->rfa_status & I82558_RFA_STATUS_CE) && not promiscuous
			/* possible on 557 only */
			zin_acquire_rx_lock(lp);
			/* do not count twice: ++lp->Statistics.hs_rx_crc; */
			/* requeue buffer -- must be called with lock held */
			zxe_set_recv_buf(ppa, bp, lp);
			continue;
#endif


		/* pdk_print_ether_packet("RX0", rptr + sizeof(i82558_rfa_t) + RFA_OFFSET, frame_size); */
#if ZIN_DUMP_BUFFERS
		pdk_buf_dump("rx_bp", bp);
#endif
		pdk_buf_reserve(bp, sizeof(i82558_rfa_t) + RFA_OFFSET);
		pdk_buf_set_len(bp, frame_size);

#if ZIN_DUMP_PACKETS
		pdk_print_ether_packet("RX:", rptr + sizeof(i82558_rfa_t) + RFA_OFFSET, frame_size);
#endif
#ifdef PDK_HAL_SPEEDO
		lp->rx_bytes += frame_size;
#endif
		lp->Statistics.hs_rx_bytes += frame_size;

		/* Give buffer to upper level */
		/* fixme: receive packet status */
		(*lp->layer_routines.pdk_recv_complete)(ppa, bp, 0);

		/* Update the statistics based on Receive status */
		++lp->Statistics.hs_rx_packets;

		/* acquire lock before checking list pointers again */
		zin_acquire_rx_lock(lp);
	}

	/* beware: this possibly leaves with lp->rfa_head_bp == NULL and
	   lp->rfa_tail_bp pointing to the last used pdk_buf_t. */

	zin_release_rx_lock(lp);

	if(isr_value & I82558_IS_RNR) {
		DEBUG4("zin_process_recv_complete(%d): (re)starting ru at 0x%08x\n", (int)ppa, lp->rfa_head_paddr);
		zin_acquire_tx_lock(lp);
		zin_scb_wait(ppa);
		pdk_reg_put32(ppa, I82558_GP_REG, lp->rfa_head_paddr);
		zin_scb_cmd(ppa, I82558_CMD_RU_START);
		zin_release_tx_lock(lp);
		}
}




uint32_t
hal_get_state(ppa, hsp)
int32_t ppa;
hal_state_t *hsp;
{
	local_adapter_t	*lp;
	pdk_hal_stat_t stat;

	pdk_bzero(hsp, sizeof(hal_state_t));

	hsp->hs_version = HS_VERSION;

	if ((ppa < 0) || (ppa >= NUM_PORTS)) {
		hsp->hs_state = HS_STATE_NOT_INIT;
		return(0);
	}

	lp = local_adapters[ppa];
	if (lp == 0) {
		hsp->hs_state = HS_STATE_NOT_INIT;
		return(0);
	}

	hal_get_stats(ppa, &stat);
	
	switch(lp->state) {
		case LOCAL_ADAPTER_RUNNING:
			hsp->hs_state = HS_STATE_RUNNING;
			break;
		case LOCAL_ADAPTER_RELEASED:
			hsp->hs_state = HS_STATE_RELEASED;
			break;
		default:
			hsp->hs_state = HS_STATE_STOPPED;
	}

	switch(lp->configured_media_type) {
		case ZXE_MEDIA_AUTO:
			hsp->hs_link_cfg = HS_LINK_AUTO;
			break;

		case ZXE_MEDIA_TP:
			hsp->hs_link_cfg = HS_LINK_TP;
			break;

		case ZXE_MEDIA_BNC:
			hsp->hs_link_cfg = HS_LINK_BNC;
			break;

		case ZXE_MEDIA_AUI:
			hsp->hs_link_cfg = HS_LINK_AUI;
			break;

		case ZXE_MEDIA_TPFD:
			hsp->hs_link_cfg = HS_LINK_TPFD;
			break;

		case ZXE_MEDIA_TX:
			hsp->hs_link_cfg = HS_LINK_TX;
			break;

		case ZXE_MEDIA_TXFD:
			hsp->hs_link_cfg = HS_LINK_TXFD;
			break;

		case ZXE_MEDIA_T4:
			hsp->hs_link_cfg = HS_LINK_T4;
			break;

		case ZXE_MEDIA_AUTO_COMPLETE:
			hsp->hs_link_cfg = HS_LINK_AUTO;
			break;
	}


	switch(lp->current_media_type) {
		case ZXE_MEDIA_AUTO:
			hsp->hs_link_state = HS_LINK_DOWN;
			break;

		case ZXE_MEDIA_TP:
			hsp->hs_link_state = HS_LINK_TP;
			break;

		case ZXE_MEDIA_BNC:
			hsp->hs_link_state = HS_LINK_BNC;
			break;

		case ZXE_MEDIA_AUI:
			hsp->hs_link_state = HS_LINK_AUI;
			break;

		case ZXE_MEDIA_TPFD:
			hsp->hs_link_state = HS_LINK_TPFD;
			break;

		case ZXE_MEDIA_TX:
			hsp->hs_link_state = HS_LINK_TX;
			break;

		case ZXE_MEDIA_TXFD:
			hsp->hs_link_state = HS_LINK_TXFD;
			break;

		case ZXE_MEDIA_T4:
			hsp->hs_link_state = HS_LINK_T4;
			break;

		case ZXE_MEDIA_AUTO_COMPLETE:
			hsp->hs_link_state = HS_LINK_DOWN;
			break;
	}

	if(!lp->link_status_up) {
		hsp->hs_link_state = HS_LINK_DOWN;
#if 0
		hsp->hs_fault =
		(lp->led_local_settings & (FAILOVER_FAULT_LED|EXT_FAULT_LED))?HS_FAULT_EXTERNAL:
		((lp->led_local_settings & (INT_FAULT_LED))?HS_FAULT_INTERNAL:0);
#endif
	}
	hsp->hs_tx_descriptor_count = lp->tx_descriptor_count;
	hsp->hs_rx_descriptor_count = lp->rx_descriptor_count;
	hsp->hs_tx_icount = lp->tx_icount;
	hsp->hs_rx_icount = lp->rx_icount;
	hsp->hs_other_icount = lp->other_icount;

	hsp->hs_tx_resources_max = zin_num_txd-1;
	hsp->hs_tx_resources = zin_num_txd - lp->free_tx_descriptor_count;
	hsp->hs_tx_lwm = lp->tx_lwm-1;
	lp->tx_lwm = zin_num_txd;
	hsp->hs_tx_depleted = lp->tx_depleted;

	hsp->hs_tx_packets = stat.hs_tx_packets;
	hsp->hs_tx_err_packets = stat.hs_tx_err_packets;
	hsp->hs_tx_one_collision = stat.hs_tx_one_collision;
	hsp->hs_tx_more_collisions = stat.hs_tx_more_collisions;
	hsp->hs_tx_max_collisions = stat.hs_tx_max_collisions;
	hsp->hs_tx_late_collisions = stat.hs_tx_late_collisions;
	hsp->hs_tx_carrier_loss = stat.hs_tx_carrier_loss;
	hsp->hs_tx_no_carrier = stat.hs_tx_no_carrier;
	hsp->hs_tx_jabber = stat.hs_tx_jabber;
	hsp->hs_tx_underrun = stat.hs_tx_underrun;
	hsp->hs_tx_overrun = stat.hs_tx_overrun;
	hsp->hs_tx_deferred = stat.hs_tx_deferred;
	hsp->hs_rx_packets = stat.hs_rx_packets;

	hsp->hs_rx_resources_max = lp->rx_hwm;
	hsp->hs_rx_resources = lp->rx_num_bufs;
	hsp->hs_rx_lwm = lp->rx_lwm;
	lp->rx_lwm = lp->rx_hwm;
#if 0
/* fixme: count interrupt RU not ready */
	lp->rx_depleted += pdk_reg_get32(ppa, CSR8) & 0x1ffff;
#endif
	hsp->hs_rx_depleted = lp->rx_depleted;

	hsp->hs_rx_err_packets = stat.hs_rx_err_packets;
	hsp->hs_rx_collisions = stat.hs_rx_collisions;
	hsp->hs_rx_mii_error = stat.hs_rx_mii_error;
	hsp->hs_rx_crc = stat.hs_rx_crc;
	hsp->hs_rx_descriptor_error = stat.hs_rx_descriptor_error;
	hsp->hs_rx_frame_too_long = stat.hs_rx_frame_too_long;
	hsp->hs_rx_runt_frame = stat.hs_rx_runt_frame;
	hsp->hs_rx_alignment_error = stat.hs_rx_alignment_error;
	hsp->hs_rx_overrun = stat.hs_rx_overrun;

	return(0);
}

ZIN_LOCAL uint32_t
hal_get_stats(ppa, stat)
int32_t ppa;
pdk_hal_stat_t *stat;
{
	local_adapter_t	*lp;


	if ((lp = local_adapters[ppa]) == 0) {
		return(PDK_OUT_STATE);
	}

	if (lp->tx_descriptor_lock) {
		zin_acquire_tx_lock(lp);
	}

	zin_dump_counters(ppa);

	if (lp->tx_descriptor_lock) {
		zin_release_tx_lock(lp);
	}

	pdk_bcopy(&lp->Statistics, stat, sizeof(pdk_hal_stat_t));
	return(0);
}


ZIN_LOCAL uint32_t
hal_set_stats(ppa, stat)
int32_t ppa;
pdk_hal_stat_t *stat;
{
	local_adapter_t	*lp;


	if ((lp = local_adapters[ppa]) == 0) {
		return(PDK_OUT_STATE);
	}

	pdk_bcopy(stat, &lp->Statistics, sizeof(pdk_hal_stat_t));
	return(0);
}


#ifdef PDK_HAL_SPEEDO

uint32_t		tx_bytes=0;
uint32_t		rx_bytes=0;
uint32_t
hal_get_speed(ppa, hsp)
int32_t ppa;
hal_speedometer_t *hsp;
{
	local_adapter_t	*lp;
/*	pdk_hal_stat_t stat; */

	pdk_bzero(hsp, sizeof(hal_speedometer_t));

	hsp->hsp_version = HSP_VERSION;

	lp = local_adapters[ppa];
	if (lp == 0) {
		DEBUG1("hal_get_speed NOT INIT\n");
		hsp->hsp_state = HS_STATE_NOT_INIT;
		return(0);
	}

	if (lp->state == LOCAL_ADAPTER_RUNNING)
		hsp->hsp_state = HS_STATE_RUNNING;
	else
		hsp->hsp_state = HS_STATE_STOPPED;

	pdk_bcopy(&lp->hsp,hsp,sizeof(hal_speedometer_t));
	hsp->hsp_poll_rate = lp->speedometer_poll_rate;
	hsp->hsp_version = HSP_VERSION;

	return(0);
}

uint32_t
hal_set_speedometer_params(ppa, hsp)
int32_t ppa;
hal_speedometer_t *hsp;
{
	local_adapter_t	*lp;

	if (hsp->hsp_version != HSP_VERSION ) {
		return(PDK_BAD_PARAM);
	}
	if ( hsp->hsp_poll_rate	== 0 ) {
  		pdk_hal_ctl(ppa, PDK_HAL_STOP_SPEEDOMETER_POLL, 0);
		return(0);
	}

	if ( hsp->hsp_poll_rate	== -1 ) {
		/* fixme: hsp_poll_rate unsigned? when does it get to -1?
		   netblaster hal does the same. */
		/* XXX deal with this */
	}

	lp = local_adapters[ppa];
	if (lp == 0) {
		DEBUG1("hal_get_speed NOT INIT\n");
		hsp->hsp_state = HS_STATE_NOT_INIT;
		return(0);
	}

	if (lp->state == LOCAL_ADAPTER_RUNNING)
		hsp->hsp_state = HS_STATE_RUNNING;
	else
		hsp->hsp_state = HS_STATE_STOPPED;

	switch(hsp->notification_type) {
		case PSDB_GREATER_THAN_OR_EQUAL:
			pdk_bcopy(hsp,&lp->set_params_gte_hsp,sizeof(hal_speedometer_t));
			break;
		case PSDB_LESS_THAN:
			pdk_bcopy(hsp,&lp->set_params_lt_hsp,sizeof(hal_speedometer_t));
			break;
	}

  	pdk_hal_ctl(ppa,PDK_HAL_SET_SPEEDOMETER_POLL_RATE_AND_START,(void *)hsp->hsp_poll_rate);

	return(0);
}

void hal_psdb_speedometer_notification(int32_t ppa,uint32_t notification_type);

void
hal_psdb_speedometer_notification(ppa,notification_type)
int32_t ppa;
uint32_t notification_type;
{
	local_adapter_t	*lp;
	pdk_psdb_hal_speedometer_notification_t pftn;
	hal_speedometer_t *hsp=&pftn.phsn_speed;
	uint32_t err;

	lp = local_adapters[ppa];

	pftn.phsn_header.ppr_header.Size = sizeof(pftn);
	pftn.phsn_header.ppr_header.Type = PSDB_TYPE_HAL_SPEEDOMETER_NOTIFICATION;
	pftn.phsn_header.ppr_ppa = ppa;
	/* update record with current values */
	hsp->notification_type=notification_type;

	err = hal_get_speed(ppa, hsp);
	if (err)
		return;
	else {
		register uint32_t *lp, *record_lim;

		lp = (uint32_t *) &pftn;
		record_lim = (uint32_t *)((char *)lp + pftn.phsn_header.ppr_header.Size);
		while (lp < record_lim)
			*lp++ = psdb_ntop(*lp);
	}

	(*lp->layer_routines.pdk_status)(ppa, PDK_PSDB, &pftn);
}

void
hal_speedometer_timeout(arg)
void *arg;
{
/*	uint32_t tmp; */
	local_adapter_t	*lp;
	int32_t ppa=(int32_t)arg;
	lp = local_adapters[ppa];
	
	if (lp->speedometer_poll_rate == 0) {
		lp->hal_speedometer_timeout_id = 0;
		return;
	}

	/* do rate calculations: current rate = (current-last)/interval */

	lp->hsp.hsp_tx_pps = 
		HAL_SPEEDOMETER_RATE(lp->Statistics.hs_tx_packets-lp->prev_Statistics.hs_tx_packets);
	lp->hsp.hsp_rx_pps = 
		HAL_SPEEDOMETER_RATE(lp->Statistics.hs_rx_packets-lp->prev_Statistics.hs_rx_packets);

	lp->hsp.hsp_tx_ips = 
		HAL_SPEEDOMETER_RATE(lp->tx_icount-lp->prev_tx_icount);

	lp->hsp.hsp_rx_ips = 
		HAL_SPEEDOMETER_RATE(lp->rx_icount-lp->prev_rx_icount);

	lp->hsp.hsp_tx_eps = 
		HAL_SPEEDOMETER_RATE(lp->Statistics.hs_tx_err_packets-lp->prev_Statistics.hs_tx_err_packets);

	lp->hsp.hsp_rx_eps = 
		HAL_SPEEDOMETER_RATE(lp->Statistics.hs_rx_err_packets-lp->prev_Statistics.hs_rx_err_packets);

#define ONE_MBPS 1000000
#define MBPS (ONE_MBPS)

#if 0
	if (lp->LinkSpeed > 0) {

		tmp = HAL_SPEEDOMETER_RATE(lp->tx_bytes);
		tmp = (tmp*(100/lp->LinkSpeed)+MBPS/16)/(MBPS/8);
		lp->hsp.hsp_tx_pctutil = tmp;

		tmp  = HAL_SPEEDOMETER_RATE(lp->rx_bytes);
		tmp = (tmp*(100/lp->LinkSpeed)+MBPS/16)/(MBPS/8);
		lp->hsp.hsp_rx_pctutil = tmp;
	
	}
#endif

	/* update maximums */

	if (lp->hsp.hsp_tx_pps > lp->hsp.hsp_tx_pps_hwm)
		lp->hsp.hsp_tx_pps_hwm = lp->hsp.hsp_tx_pps;
 
	if (lp->hsp.hsp_rx_pps > lp->hsp.hsp_rx_pps_hwm)
		lp->hsp.hsp_rx_pps_hwm = lp->hsp.hsp_rx_pps;

	if (lp->hsp.hsp_tx_ips > lp->hsp.hsp_tx_ips_hwm)
		lp->hsp.hsp_tx_ips_hwm = lp->hsp.hsp_tx_ips;
 
	if (lp->hsp.hsp_rx_ips > lp->hsp.hsp_rx_ips_hwm)
		lp->hsp.hsp_rx_ips_hwm = lp->hsp.hsp_rx_ips;

	if (lp->hsp.hsp_tx_pctutil > 99)
		lp->hsp.hsp_tx_pctutil = 99;
	
	if (lp->hsp.hsp_rx_pctutil > 99)
		lp->hsp.hsp_rx_pctutil = 99;

	if (lp->hsp.hsp_tx_pctutil > lp->hsp.hsp_tx_pctutil_hwm)
		lp->hsp.hsp_tx_pctutil_hwm = lp->hsp.hsp_tx_pctutil;

	if (lp->hsp.hsp_rx_pctutil > lp->hsp.hsp_rx_pctutil_hwm)
		lp->hsp.hsp_rx_pctutil_hwm = lp->hsp.hsp_rx_pctutil;

#define PSDB_ALARM_NULL 0

/* check for greater than alarm conditions */
	{
	uint32_t alarm_condition=0;

	if (lp->set_params_gte_hsp.notification_type == PSDB_GREATER_THAN_OR_EQUAL ) {

		if ( ( lp->hsp.hsp_rx_pctutil >= lp->set_params_gte_hsp.hsp_rx_pctutil )
			&& ( lp->set_params_gte_hsp.hsp_rx_pctutil != PSDB_ALARM_NULL ) )
			alarm_condition=1;
	
		if ( ( lp->hsp.hsp_tx_pctutil >= lp->set_params_gte_hsp.hsp_tx_pctutil ) 
			&& ( lp->set_params_gte_hsp.hsp_tx_pctutil != PSDB_ALARM_NULL ) )
			alarm_condition=1;	       
	
		if ( (lp->hsp.hsp_rx_pps >= lp->set_params_gte_hsp.hsp_rx_pps )
			&& ( lp->set_params_gte_hsp.hsp_rx_pps != PSDB_ALARM_NULL ) )
			alarm_condition=1;
	
		if ( (lp->hsp.hsp_tx_pps >= lp->set_params_gte_hsp.hsp_tx_pps )
			&& ( lp->set_params_gte_hsp.hsp_tx_pps !=  PSDB_ALARM_NULL ) )
			alarm_condition=1;

		if ( (lp->hsp.hsp_rx_eps >= lp->set_params_gte_hsp.hsp_rx_eps )
			&& ( lp->set_params_gte_hsp.hsp_rx_eps != PSDB_ALARM_NULL ) )
			alarm_condition=1;
	
		if ( (lp->hsp.hsp_tx_eps >= lp->set_params_gte_hsp.hsp_tx_eps )
			&& ( lp->set_params_gte_hsp.hsp_tx_eps !=  PSDB_ALARM_NULL ) )
			alarm_condition=1;

		if ( (lp->hsp.hsp_rx_ips >= lp->set_params_gte_hsp.hsp_rx_ips )
			&& ( lp->set_params_gte_hsp.hsp_rx_ips != PSDB_ALARM_NULL ) )
			alarm_condition=1;
	
		if ( (lp->hsp.hsp_tx_ips >= lp->set_params_gte_hsp.hsp_tx_ips )
			&& ( lp->set_params_gte_hsp.hsp_tx_ips !=  PSDB_ALARM_NULL ) )
			alarm_condition=1;

		if (alarm_condition)
			hal_psdb_speedometer_notification(ppa,PSDB_GREATER_THAN_OR_EQUAL);
	
		}
	}

/* check for less than or equal to alarm conditions */

	{
	uint32_t alarm_condition=0;
	if (lp->set_params_gte_hsp.notification_type == PSDB_LESS_THAN) {

		if ( ( lp->hsp.hsp_rx_pctutil < lp->set_params_lt_hsp.hsp_rx_pctutil )
			&& ( lp->set_params_lt_hsp.hsp_rx_pctutil != PSDB_ALARM_NULL ) )
			alarm_condition=1;
	
		if ( ( lp->hsp.hsp_tx_pctutil < lp->set_params_lt_hsp.hsp_tx_pctutil )
			&& ( lp->set_params_lt_hsp.hsp_tx_pctutil != PSDB_ALARM_NULL ) )
			alarm_condition=1;
	
		if ( ( lp->hsp.hsp_rx_pps < lp->set_params_lt_hsp.hsp_rx_pps )
			&& ( lp->set_params_lt_hsp.hsp_rx_pps != PSDB_ALARM_NULL ) )
			alarm_condition=1;
	
		if ( ( lp->hsp.hsp_tx_pps < lp->set_params_lt_hsp.hsp_tx_pps )
			&& ( lp->set_params_lt_hsp.hsp_tx_pps != PSDB_ALARM_NULL ) )
			alarm_condition=1;

		if ( (lp->hsp.hsp_rx_eps < lp->set_params_lt_hsp.hsp_rx_eps )
			&& ( lp->set_params_lt_hsp.hsp_rx_eps != PSDB_ALARM_NULL ) )
			alarm_condition=1;
	
		if ( (lp->hsp.hsp_tx_eps < lp->set_params_lt_hsp.hsp_tx_eps )
			&& ( lp->set_params_lt_hsp.hsp_tx_eps !=  PSDB_ALARM_NULL ) )
			alarm_condition=1;

		if ( (lp->hsp.hsp_rx_ips < lp->set_params_lt_hsp.hsp_rx_ips )
			&& ( lp->set_params_lt_hsp.hsp_rx_ips != PSDB_ALARM_NULL ) )
			alarm_condition=1;
	
		if ( (lp->hsp.hsp_tx_ips < lp->set_params_lt_hsp.hsp_tx_ips )
			&& ( lp->set_params_lt_hsp.hsp_tx_ips !=  PSDB_ALARM_NULL ) )
			alarm_condition=1;
	 
		if (alarm_condition)
			hal_psdb_speedometer_notification(ppa,PSDB_LESS_THAN);
	
		}
	}

	/* now remember the current totals */

	lp->prev_Statistics.hs_tx_packets = lp->Statistics.hs_tx_packets;
	lp->prev_Statistics.hs_rx_packets = lp->Statistics.hs_rx_packets;
	lp->prev_Statistics.hs_tx_err_packets = lp->Statistics.hs_tx_err_packets;
	lp->prev_Statistics.hs_rx_err_packets = lp->Statistics.hs_rx_err_packets;
	lp->prev_tx_icount = lp->tx_icount;
	lp->prev_rx_icount = lp->rx_icount;
	if (lp->tx_bytes > tx_bytes) tx_bytes=lp->tx_bytes;
	if (lp->rx_bytes > rx_bytes) rx_bytes=lp->rx_bytes;
	lp->tx_bytes=0;
	lp->rx_bytes=0;

	lp->hal_speedometer_timeout_id = pdk_timeout(hal_speedometer_timeout, (void *)ppa, HAL_SPEEDOMETER_TICKS);
}


#endif



uint32_t
pdk_hal_ctl(ppa, opcode, param)
int32_t ppa;
uint32_t opcode;
void *param;
{
	local_adapter_t	*lp;
	uint32_t	err;
#if 0
	uint32_t	local_settings;
#endif

	if(opcode != PDK_HAL_GET_NUM_RECEIVE_BUFS) {
		/* not called at interrupt time */
		DEBUG2("pdk_hal_ctl(%d) enter opcode=%d\n", (int)ppa, opcode);
	}

	lp = local_adapters[ppa];

	err = 0;

	switch(opcode) {
		case PDK_PSDB:
		case PDK_PSDB_INLINE: {
			pdk_psdb_rhdr_t *prp;
			uint32_t record_type;

			prp = (pdk_psdb_rhdr_t *) param;

			record_type = psdb_pton(prp->ppr_header.Type);
			ppa = (int32_t) psdb_pton(prp->ppr_ppa);

			err = hal_psdb(prp);

			/* Inline calls return the error, they do not create
			 * a PSDB status upcall.
			 */
			if (opcode == PDK_PSDB_INLINE)
				break;

			hal_psdb_return(lp, record_type, err);

			if ((record_type == PSDB_TYPE_HAL_GET_STATE) && !err) {
				hal_psdb_return_state(lp, (pdk_psdb_hal_state_t *) prp);
			}
#ifdef PDK_HAL_SPEEDO
			if ((record_type == PSDB_TYPE_HAL_SPEEDOMETER_GET) && !err) {
				hal_psdb_return_speed(lp, (pdk_psdb_hal_speedometer_get_t *) prp);
			}
#endif
			/* fixme: this is the behavior of the znb code, ok? */
			err = 0;
			break; }

		case PDK_HAL_EXT_FAULT:
			if (lp == 0)
				break;
			/* fixme:  */
			/* zxe_external_fault_led(ppa, (int32_t) param, 1); */
			break;

		case PDK_HAL_MTU_SIZE:
			/* fixme: return regular media mtu or max possible? */
			err = PDK_ETHERMTU;
			break;

		case PDK_HAL_RECEIVE_INIT:
			err = pdk_buf_receive_init(ppa);
			break;

		case PDK_HAL_RECEIVE_DEINIT:
			pdk_buf_receive_deinit(ppa);
			break;

		case PDK_HAL_GET_NUM_RECEIVE_BUFS:
			*((uint32_t *) param) = lp->rx_num_bufs;
			break;

		case PDK_HAL_ENABLE_INTERRUPTS:
/* fixme: not tested */
			zin_enable_interrupts(ppa);
			break;

		case PDK_HAL_DISABLE_INTERRUPTS:
/* fixme: not tested */
			zin_disable_interrupts(ppa);
			break;

		case PDK_HAL_GET_INFLIGHT:
			/* fixme: ??? */
			break;

		case PDK_HAL_GETID:
			/* fixme: assign id for 55x ? */
			break;

		case PDK_HAL_SETID:
			/* fixme: ??? */
			break;

		case PDK_HAL_RESTART:
			/* fixme:  */
			pdk_hal_restart(ppa);
			break;

		case PDK_HAL_INIT_AND_START:
			if (lp->rx_descriptor_lock) {
				zin_acquire_rx_lock(lp);
			}

			if (lp->tx_descriptor_lock) {
				zin_acquire_tx_lock(lp);
			}

			err = (uint32_t) pdk_hal_start(ppa,1);

			if (lp->tx_descriptor_lock) {
				zin_release_tx_lock(lp);
			}

			if (lp->rx_descriptor_lock) {
				zin_release_rx_lock(lp);
			}
			break;

		case PDK_HAL_STOP_AND_RESET:
			if (lp->rx_descriptor_lock) {
				zin_acquire_rx_lock(lp);
			}

			if (lp->tx_descriptor_lock) {
				zin_acquire_tx_lock(lp);
			}

			pdk_hal_stop(ppa);
			lp->state = LOCAL_ADAPTER_ALLOCATED;

			if (lp->tx_descriptor_lock) {
				zin_release_tx_lock(lp);
			}

			if (lp->rx_descriptor_lock) {
				zin_release_rx_lock(lp);
			}
			pdk_time_delay(1);

			/* Need to do buffer recycling ?????? */
			break;

		case PDK_HAL_DEBUG:
			DEBUG0("pdk_hal_ctl(%d): debug set to %d\n", (int)ppa, param);
			pdk_hal_debug = (int32_t) param;
			break;

		case PDK_HAL_REVLEVEL:
			if (lp == 0) {
				err = PDK_OUT_STATE;
				break;
				}
			/* fixme: ??? */
			break;

		case PDK_HAL_GET_COUNTERS:
			err = hal_get_stats(ppa, (pdk_hal_stat_t *) param);
			break;

		case PDK_HAL_SET_COUNTERS:
			err = hal_set_stats(ppa, (pdk_hal_stat_t *) param);
			break;

		case PDK_HAL_GET_HDWRADDR:
			DEBUG2("pdk_hal_ctl(%d) get_hdwraddr\n", (int)ppa);
			if (lp == 0)
				err = PDK_OUT_STATE;
			else
				pdk_bcopy(lp->factory_mac_address, (uchar*)param, PDK_ETHERADDRL);
			break;

		case PDK_HAL_GET_PHYSADDR:
			DEBUG2("pdk_hal_ctl(%d) get_physaddr\n", (int)ppa);
			if (lp == 0)
				err = PDK_OUT_STATE;
			else
				pdk_bcopy(lp->current_mac_address, (uchar*)param, PDK_ETHERADDRL);
			break;

		case PDK_HAL_SET_PHYSADDR:
			DEBUG2("pdk_hal_ctl(%d) set_physaddr\n", (int)ppa);
			if (lp == 0) {
				err = PDK_OUT_STATE;
				break;
			}
			pdk_bcopy((uchar*)param, lp->current_mac_address, PDK_ETHERADDRL);

			if (lp->tx_descriptor_lock) {
				zin_acquire_tx_lock(lp);
			}
			err = 0;
			/* If not running address will be setup and loaded
			 * on PDK_HAL_INIT_AND_START.
			 */
			if (lp->state == LOCAL_ADAPTER_RUNNING) {
				err = zin_station_address_update(lp->ppa);
			}
			if (lp->tx_descriptor_lock) {
				zin_release_tx_lock(lp);
			}
			break;

		case PDK_HAL_SET_PROMISCUOUS_PHYS:
			if (lp == 0) {
				err = PDK_OUT_STATE;
				break;
			}
			lp->promiscuous_multicast = 1;
			lp->promiscuous_physical = 1;
			if (lp->state == LOCAL_ADAPTER_RUNNING) {
				zin_recv_mode_update(ppa);
			}
			break;

		case PDK_HAL_SET_PROMISCUOUS_MULTI:
			if (lp == 0) {
				err = PDK_OUT_STATE;
				break;
			}
			lp->promiscuous_multicast = 1;
			lp->promiscuous_physical = 0;
			if (lp->state == LOCAL_ADAPTER_RUNNING) {
				zin_recv_mode_update(ppa);
			}
			break;

		case PDK_HAL_CLEAR_PROMISCUOUS:
			if (lp == 0) {
				err = PDK_OUT_STATE;
				break;
			}
			lp->promiscuous_multicast = 0;
			lp->promiscuous_physical = 0;
			if (lp->state == LOCAL_ADAPTER_RUNNING) {
				zin_recv_mode_update(ppa);
			}
			break;

		case PDK_HAL_ADD_MULTICAST:
			if (lp == 0) {
				err = PDK_OUT_STATE;
			}
			else if (lp->multicast_address_count >= I82558_MAX_MCA) {
				err = PDK_NO_ROOM;
			}
			else {
				err = zin_add_multicast_address(lp->ppa, (uchar *) param);
			}
			break;

		case PDK_HAL_DEL_MULTICAST:
			if (lp == 0) {
				err = PDK_OUT_STATE;
			}
			else if (lp->multicast_address_count == 0) {
				err = PDK_NOT_FOUND;
			}
			else {
				err = zin_del_multicast_address(lp->ppa, (uchar *) param);
			}
			break;

		case PDK_HAL_CLEAR_MULTICAST:
			if (lp == 0) {
				err = PDK_OUT_STATE;
				break;
			}
			lp->multicast_address_count = 0;
			if (lp->tx_descriptor_lock) {
				zin_acquire_tx_lock(lp);
			}
			if (lp->state == LOCAL_ADAPTER_RUNNING) {
				err = zin_multicast_update(ppa);
			}
			if (lp->tx_descriptor_lock) {
				zin_release_tx_lock(lp);
			}
			break;

		case PDK_HAL_DUMPCAM:
            		break;

		case PDK_HAL_DUMPRING:
            		break;

       		case PDK_HAL_LED:
            		break;

		case PDK_HAL_SET_MEDIATYPE:
			if (lp == 0) {
				err = PDK_OUT_STATE;
			}
			else if (((int32_t) param) > ZXE_MEDIA_MAX) {
				err = PDK_BAD_PARAM;
			}
			else {
				lp->configured_media_type = (int32_t) param;
				/* fixme: should the media type be set in hw? */
			}
			break;

		case PDK_HAL_GET_MEDIATYPE:
			if (lp == 0)
				err = PDK_OUT_STATE;
			else
				*((uint32_t*)param) = lp->current_media_type;
			break;

		case PDK_HAL_SET_TX_PENDING_THRESHOLD:
			lp->tx_pending_threshold = (int32_t) param;
			lp->tx_pending_count = (int32_t) 0;
			break;

		case PDK_HAL_HOTSWAP:
			if (lp == 0) {
				err = PDK_OUT_STATE;
			}
			else if ((int32_t) param == 0) {
				/* card gone */
				if (lp->state == LOCAL_ADAPTER_ALLOCATED) {
					lp->hardware_initialized = 0;
					lp->state = LOCAL_ADAPTER_RELEASED;
					lp->factory_mac_address[0] =
					lp->factory_mac_address[1] =
					lp->factory_mac_address[2] = 0;
				}
				else {
					err = PDK_OUT_STATE;
				}
			}
			else {
				/* card inserted */
				if (lp->state == LOCAL_ADAPTER_RELEASED) {
					lp->state = LOCAL_ADAPTER_ALLOCATED;
				}
				else {
					err = PDK_OUT_STATE;
				}
			}
			break;

#ifdef PDK_HAL_SPEEDO
		case PDK_HAL_SET_SPEEDOMETER_POLL_RATE_AND_START:
			/* input is in ms */
			DEBUG2("SPEEDOMETER start: PPA: %d\n",ppa);
			if (param == 0) {
				DEBUG1("pdk_hal_ctl: ZERO poll interval\n");
				err = PDK_BAD_PARAM;
				break;
			}
			lp->speedometer_poll_rate = (uint32_t)param;
			if (lp->hal_speedometer_timeout_id == 0) {
				lp->hal_speedometer_timeout_id = pdk_timeout(hal_speedometer_timeout, (void *)ppa, HAL_SPEEDOMETER_TICKS);
			}
			break;

		case PDK_HAL_STOP_SPEEDOMETER_POLL:
			if (lp->hal_speedometer_timeout_id != 0) {
				pdk_untimeout(lp->hal_speedometer_timeout_id);
				lp->hal_speedometer_timeout_id = 0;
			}
			lp->speedometer_poll_rate = 0;
			break;
#endif
		default:
			DEBUG2("pdk_hal_ctl(%d) bad_param\n", (int)ppa);
			err = PDK_BAD_PARAM;
	}

	if(opcode != PDK_HAL_GET_NUM_RECEIVE_BUFS) {
		/* not called at interrupt time */
		DEBUG2("pdk_hal_ctl(%d) exit err=%d\n", (int)ppa, err);
	}

	return(err);
}




void
hal_psdb_return(lp, type, return_val)
local_adapter_t *lp;
uint32_t type;
uint32_t return_val;
{
	int32_t ppa;
	pdk_psdb_rhdr_t *prp;
	pdk_psdb_success_t ps;
	pdk_psdb_error_t pe;
	
	ppa = lp->ppa;

	if (return_val == 0) {
		ps.ps_header.ppr_header.Size = psdb_ntop(sizeof(ps));
		ps.ps_header.ppr_header.Type = psdb_ntop(PSDB_TYPE_SUCCESS);
		ps.ps_header.ppr_ppa = psdb_ntop(ppa);
		ps.ps_success_type = psdb_ntop(type);

		prp = &ps.ps_header;
	}
	else {
		pe.pe_header.ppr_header.Size = psdb_ntop(sizeof(pe));
		pe.pe_header.ppr_header.Type = psdb_ntop(PSDB_TYPE_ERROR);
		pe.pe_header.ppr_dev_id = psdb_ntop(ppa);
		pe.pe_error_type = psdb_ntop(type);
		pe.pe_error = psdb_ntop(return_val);

		prp = &pe.pe_header;
	}

	(*lp->layer_routines.pdk_status)(ppa, PDK_PSDB, (void *) prp);
}

void
hal_psdb_return_state(lp, phsp)
local_adapter_t *lp;
pdk_psdb_hal_state_t *phsp;
{
	DEBUG2("hal_psdb_return_state(%d): Called\n", lp->ppa);

	(*lp->layer_routines.pdk_status)(lp->ppa, PDK_PSDB, (void *) phsp);
}

#ifdef PDK_HAL_SPEEDO
void
hal_psdb_return_speed(lp, phsp)
local_adapter_t *lp;
pdk_psdb_hal_speedometer_get_t *phsp;
{
	DEBUG2("hal_psdb_return_speed(%d): Called\n", lp->ppa);

	(*lp->layer_routines.pdk_status)(lp->ppa, PDK_PSDB, (void *) phsp);

}
#endif

ZIN_LOCAL uint32_t
hal_psdb(record)
pdk_psdb_rhdr_t *record;
{
	uint32_t record_type, record_size;
	uint32_t *lp;
	int32_t ppa;
	int32_t i;

	record_type = psdb_pton(record->ppr_header.Type);
	record_size = psdb_pton(record->ppr_header.Size);

	lp = (uint32_t *) record;
	for (i=0;i<(int32_t)(record_size/sizeof(uint32_t));i++)
		lp[i] = psdb_pton(lp[i]);

	ppa = (int32_t) record->ppr_ppa;

	/* Let hal_get_state check the ppa and fill in structure */
	if (record_type != PSDB_TYPE_HAL_GET_STATE) {
		if ((ppa < 0) || (ppa >= NUM_PORTS))
			return(PSDB_PPA_OUT_RANGE);
	}

	switch(record_type) {
		case PSDB_TYPE_HAL_GET_STATE: {
			pdk_psdb_hal_state_t *phsp;
			uint32_t err;

			phsp = (pdk_psdb_hal_state_t *) record;

			if (record_size != sizeof(pdk_psdb_hal_state_t))
				return(PSDB_INCORRECT_SIZE);

			err = hal_get_state(ppa, &phsp->phs_state);
			if (err)
				return(err);
			else {
				register uint32_t *lp, *record_lim;

				lp = (uint32_t *) record;
				record_lim = (uint32_t *)((char *)lp + record->ppr_header.Size);
				while (lp < record_lim)
					*lp++ = psdb_ntop(*lp);
			}
			return(0);
		}
		case PSDB_TYPE_HAL_FAULT_NOTIFICATION_SET:
		 	if (local_adapters[ppa] == 0)
				return(PSDB_PPA_OUT_RANGE);

			local_adapters[ppa]->fault_notification = 1;
			return(0);

		case PSDB_TYPE_HAL_FAULT_NOTIFICATION_CLEAR:
		 	if (local_adapters[ppa] == 0)
				return(PSDB_PPA_OUT_RANGE);

			local_adapters[ppa]->fault_notification = 0;
			return(0);

#ifdef PDK_HAL_SPEEDO
		case PSDB_TYPE_HAL_SPEEDOMETER_GET: {
			pdk_psdb_hal_speedometer_get_t *phsp;
			uint32_t err;

			phsp = (pdk_psdb_hal_speedometer_get_t *) record;

			if (record_size != sizeof(pdk_psdb_hal_speedometer_get_t))
				return(PSDB_INCORRECT_SIZE);

			err = hal_get_speed(ppa, &phsp->phsg_speed);
			if (err)
				return(err);
			else {
				register uint32_t *lp, *record_lim;

				lp = (uint32_t *) record;
				record_lim = (uint32_t *)((char *)lp + record->ppr_header.Size);
				while (lp < record_lim)
					*lp++ = psdb_ntop(*lp);
			}
			return(0);
		}
		case PSDB_TYPE_HAL_SPEEDOMETER_NOTIFICATION: {
			/* XXX get alarm thresholds here */
			return(0);
		}
		case PSDB_TYPE_HAL_SPEEDOMETER_SET_PARAMS: {
			pdk_psdb_hal_speedometer_set_params_t *phsp;

			phsp = (pdk_psdb_hal_speedometer_set_params_t *) record;

			if (record_size != sizeof(pdk_psdb_hal_speedometer_set_params_t))
				return(PSDB_INCORRECT_SIZE);

			return(hal_set_speedometer_params(ppa, &phsp->phss_speed));
		}
#endif
		default:
			DEBUG3("hal_psdb:  Record type 0x%x unknown\n", record_type);
			break;
	}

	return(PSDB_PROTOCOL_ERROR);
}




ZIN_LOCAL int32_t
zin_add_multicast_address(ppa, addr)
int32_t ppa;
uchar *addr;
{
	int32_t i;
	local_adapter_t *lp;
	uint32_t err;

	DEBUG1("zin_add_multicast_address(%d): %02X%02X%02X%02X%02X%02X\n",
		ppa, addr[0], addr[1], addr[2], addr[3], addr[4], addr[5]);

	if ((lp = local_adapters[ppa]) == 0) {
		return(PDK_OUT_STATE);
	}

	if (lp->tx_descriptor_lock) {
		zin_acquire_tx_lock(lp);
	}

	for (i=0; i<lp->multicast_address_count; i++) {
		if (pdk_bcmp(lp->multicast_addresses[i], addr, PDK_ETHERADDRL) == 0) {
			/* already registered */
			if (lp->tx_descriptor_lock) {
				zin_release_tx_lock(lp);
			}
			return(0);
		}
	}

	pdk_bcopy(addr, lp->multicast_addresses[lp->multicast_address_count], PDK_ETHERADDRL);
	lp->multicast_address_count++;

	err = 0;

	if (lp->state == LOCAL_ADAPTER_RUNNING) {
		err = zin_multicast_update(ppa);
	}

	if (lp->tx_descriptor_lock) {
		zin_release_tx_lock(lp);
	}

	return(err);
}

ZIN_LOCAL int32_t
zin_del_multicast_address(ppa, addr)
int32_t ppa;
uchar *addr;
{
	int32_t i;
	int32_t		err;
	local_adapter_t *lp;

	DEBUG1("zin_del_multicast_address(%d): %02X%02X%02X%02X%02X%02X\n",
		(int)ppa, addr[0], addr[1], addr[2], addr[3], addr[4], addr[5]);

	if ((lp = local_adapters[ppa]) == 0) {
		return(PDK_OUT_STATE);
	}

	if (lp->multicast_address_count == 0)
		return(PDK_NOT_FOUND);

	if (lp->tx_descriptor_lock) {
		zin_acquire_tx_lock(lp);
	}

	for (i=0; i<lp->multicast_address_count; i++) {
		if (pdk_bcmp(lp->multicast_addresses[i], addr, PDK_ETHERADDRL) == 0) {
			lp->multicast_address_count--;
			/* Copy the last over the entry being removed */
			pdk_bcopy(lp->multicast_addresses[lp->multicast_address_count],
				lp->multicast_addresses[i], PDK_ETHERADDRL);
			err = 0;
			if (lp->state == LOCAL_ADAPTER_RUNNING) {
				err = zin_multicast_update(ppa);
			}
			if (lp->tx_descriptor_lock) {
				zin_release_tx_lock(lp);
			}
			return(err);
		}
	}

	if (lp->tx_descriptor_lock) {
		zin_release_tx_lock(lp);
	}

	return(PDK_NOT_FOUND);
}





