/*==========================================================================
**
**	./pdk/src/drivers/sal/sal_buf.c
**
**	System abstraction layer for Linux
**
**==========================================================================
**
**	(c) Copyright 1998 ZNYX Corporation
**	All Rights Reserved.
**
**==========================================================================
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**==========================================================================
**
**	david	03/03/99	Created
**	jr	09/15/01	OA version, rptr/wptr code added,
**				namespace cleanup
**	jr	02/18/02	added pdk_buf_reserve()
**
**==========================================================================
**
**	Notes:
**
**	As of linux 2.4.7, this is how the skb pointers are used
**	in non-cloned or linearized skbufs:
**
**	head	start of buffer space
**	data	start of data in buffer, data >= head	b_rptr
**	tail	end of data in buffer, tail >= data	b_wptr
**	end	end of buffer, end >= tail
**	len	size of data in buffer, len == tail - data
**
**==========================================================================
*/

#include <pdk_port.h>
#include <pdk_global.h>
#include <pdk_types.h>
#include <pdk_constant.h>

#include <pdk_mem.h>
#include <pdk_buf.h>
#include <pdk_lock.h>
#include <pdk_queue.h>
#include <pdk_linux.h>
#include <pdk_errno.h>
#include <pdk_lib.h>
#include <pdk_br554.h>

#include <pdk_print.h>

#include <asm/bitops.h>
#include <asm/atomic.h>

#define NUM_BPS (TXD_SIZE+NUM_RECV_BUFS + 128)

typedef struct pdk_buf_header {
	uint32_t		bh_state;
#define BH_DEINIT 0
#define BH_FLUSHING 1
#define BH_DEINITTING 2
#define BH_RUNNING 4
#define BH_INITED 8
	pdk_buf_t		*bh_header_space;
	pdk_buf_t		*bh_header_freeq;
	uint32_t		bh_inflight;
	pdk_lock_t		*bh_lock;
	uint32_t		hist[NUM_BPS+1];
} pdk_buf_header_t;

pdk_buf_header_t pdk_buf_array[NUM_PORTS];

/*extern int32_t sal_debug;*/

#define DEBUG1 if (pdk_sal_debug>0) pdk_print
#define DEBUG2 if (pdk_sal_debug>1) pdk_print
#define DEBUG3 if (pdk_sal_debug>2) pdk_print

#define PDK_GET_BUF(hp, bp) { \
	pdk_lock_acquire((hp)->bh_lock); \
	(hp)->hist[(hp)->bh_inflight++]++; \
	bp = (pdk_buf_t *) pdk_q_deq((pdk_q_t **) &(hp)->bh_header_freeq); \
	pdk_lock_release((hp)->bh_lock); \
}

#ifdef NOT
#define PDK_FREE_BUF(hp, bp) { \
	__label__ here; \
	\
	if (bp->b_skb != 0) {\
		DEBUG1("FREE: %p: skb %#x\n", &&here, bp->b_skb); \
	} \
	pdk_lock_acquire((hp)->bh_lock); \
	(hp)->bh_inflight--; \
	pdk_q_enq((pdk_q_t **) &(hp)->bh_header_freeq, (pdk_q_t *) (bp)); \
	pdk_lock_release((hp)->bh_lock); \
}
#endif


extern __inline__ void PDK_FREE_BUF(pdk_buf_header_t *hp, pdk_buf_t *bp)
{
	int32_t here;

	if (bp->b_skb != 0) {
		DEBUG1("FREE: %p: skb %#x\n", &here, (uint32_t)bp->b_skb);
	}
	pdk_lock_acquire((hp)->bh_lock);
	(hp)->bh_inflight--;
	pdk_q_enq((pdk_q_t **) &(hp)->bh_header_freeq, (pdk_q_t *) (bp));
	pdk_lock_release((hp)->bh_lock);
}

#if 0
uint32_t lockheld = 0;

/* DBG: debug skb's */

void check_bh_lkcnt(ppa)
int32_t ppa;
{
	pdk_buf_header_t *hp = &pdk_buf_array[ppa];

	if (atomic_read(&(hp->bh_lock->lkcnt)) > 0) {
		lockheld++;

		/* Used to come here when buffers were being lost from freeq.
		   With lock changes, freeq is OK, but we *still* come here
		   a lot.

		DEBUG1("LOCK held before znb_int32_tr %d.\n",
			atomic_read(&(hp->bh_lock->lkcnt)) );
		*/
	}
}

void freeqinfo(ppa)
int32_t ppa;
{
	pdk_buf_header_t *hp = &pdk_buf_array[ppa];
	pdk_buf_t	*buf_queue = 0;
	pdk_buf_t	*bp;
	int32_t freebps=0;
	
	pdk_lock_acquire((hp)->bh_lock);

	while ((bp = (pdk_buf_t *) pdk_q_deq((pdk_q_t **) &hp->bh_header_freeq)) != 0) {
		freebps++;
		pdk_q_enq((pdk_q_t **) &buf_queue, (pdk_q_t *)  bp);
	}

	while ((bp = (pdk_buf_t *) pdk_q_deq((pdk_q_t **) &buf_queue)) != 0) {
		pdk_q_enq((pdk_q_t **) &hp->bh_header_freeq, (pdk_q_t *)  bp);
	}

	pdk_lock_release((hp)->bh_lock);
	DEBUG1("FREEq: %d bps, %d inflight  %d lockheld\n",
		freebps, hp->bh_inflight, lockheld);
}
#endif

/*
#define NEWRX	0
#define NEWTX	1
#define TXING	2
#define RXING	3
#define RXHAL	4
#define RXERR	5
#define	TXDONE	6
#define	RXDONE	7
*/
static char *bp_labels[] = {"NEWRX", "NEWTX", "TXING", "RXING", "RXHAL",
			"RXERR", "TXDONE", "RXDONE"};

void pdk_dumpskb(ppa, id, bp, skbp)
int32_t ppa;
char *id;
pdk_buf_t *bp;
struct sk_buff *skbp;
{
	pdk_buf_header_t *hp = &pdk_buf_array[ppa];
	uint32_t *n = hp->hist;
	int32_t i;

	if (id == NULL) {
		for(i=0; i<NUM_BPS; i++) {
			if ((i & 0x7)==0)
				DEBUG1("\n%3d:",i+1);
			DEBUG1(" %8d",n[i]);
		}
		DEBUG1("\n");
		for(i=0; i<NUM_BPS; i++) {
			n[i] = 0;
		}
		return;
	}

	DEBUG1("%s (bp %#x skb %#x): h %#x  d %#x  t %#x  e %#x\n",
		id, (uint32_t)bp, (uint32_t)skbp, (uint32_t)skbp->head, (uint32_t)skbp->data, (uint32_t)skbp->tail, (uint32_t)skbp->end);
	DEBUG1("    b_len %#x  skblen: %#x  bufsize: %#x  %s\n",
		bp->b_len, skbp->len, skbp->truesize, bp_labels[bp->history]);
}


uint32_t
pdk_buf_global_init()
{
	pdk_bzero(pdk_buf_array, sizeof(pdk_buf_array));
	return(0);
}

uint32_t
pdk_buf_global_deinit()
{
	return(0);
}

uint32_t
pdk_buf_init(ppa)
int32_t ppa;
{
	pdk_buf_header_t *hp;
	pdk_buf_t *bp;
	int32_t i;

	hp = &pdk_buf_array[ppa];

	DEBUG3("pdk_buf_init(%d): state %d\n", ppa, hp->bh_state);

	if (hp->bh_state != BH_DEINIT)
		return(PDK_OUT_STATE);

	DEBUG3("pdk_buf_init(%d): header_freeq 0x%x\n", ppa, (uint32_t)hp->bh_header_freeq);

	if (hp->bh_header_freeq != 0)
		return(PDK_OUT_STATE);

	hp->bh_lock = pdk_lock_alloc(ppa, PDK_LOCK_INTR);
	if (hp->bh_lock == 0)
		return(PDK_NO_MEM);

	hp->bh_header_space = (pdk_buf_t *) pdk_mem_zalloc(ppa, sizeof(pdk_buf_t) * NUM_BPS);
	bp = (pdk_buf_t *) hp->bh_header_space;

	if (bp == 0)
		return(PDK_NO_MEM);

	for (i=0;i<NUM_BPS;i++) {
		bp->b_ppa = ppa;

		PDK_FREE_BUF(hp, bp);

		bp++;
	}

	pdk_lock_acquire(hp->bh_lock);
	hp->bh_inflight = 0;
	hp->bh_state = BH_RUNNING;
	pdk_lock_release(hp->bh_lock);

	return(0);
}

uint32_t
pdk_buf_deinit(ppa)
int32_t ppa;
{
	pdk_buf_header_t *hp;
	pdk_lock_t *bh_lock;
	pdk_buf_t *bp;
	uchar *bh_header_space;

	hp = &pdk_buf_array[ppa];

	if (hp->bh_state == BH_DEINIT)
		return(0);

	bh_lock = hp->bh_lock;
	bh_header_space = (uchar *)hp->bh_header_space;

	DEBUG3("pdk_buf_deinit(%d): Inflight %d\n", ppa, hp->bh_inflight);

	pdk_lock_acquire(bh_lock);

	if (hp->bh_inflight != 0)
		hp->bh_state = BH_DEINITTING;

	if (hp->bh_state == BH_DEINITTING) {
		pdk_lock_release(bh_lock);
		while ((bp = (pdk_buf_t *) pdk_q_deq((pdk_q_t **) &hp->bh_header_freeq)) != 0) {
			pdk_buf_free(bp);
		}
		return(0);
	}

	DEBUG3("pdk_buf_deinit(%d): Completed deinit\n", ppa);

	hp->bh_state = BH_DEINIT;
	hp->bh_header_freeq = 0;

	pdk_lock_release(bh_lock);

	pdk_mem_free(ppa, bh_header_space, sizeof(pdk_buf_t) * NUM_BPS);

	pdk_lock_free(bh_lock);

	return(0);
}


extern int32_t zxe_debug;
extern int32_t pdk_hal_debug;


/*
 * If skb is non-zero, it is attached to a pdk_buf_t structure.  Otherwise
 * a new one is allocated.
 */
pdk_buf_t *
pdk_buf_alloc(ppa, size, skb)
int32_t ppa;
uint32_t size;
struct sk_buff *skb;
{
	pdk_buf_header_t *hp;
	pdk_buf_t *bp;

	hp = &pdk_buf_array[ppa];

	PDK_GET_BUF(hp, bp);

	if (bp == 0) {
		pdk_lock_acquire(hp->bh_lock);
		hp->bh_inflight--;
		pdk_lock_release(hp->bh_lock);

		DEBUG2("pdk_buf_alloc(%d): freeq empty, inflight %d\n", ppa, hp->bh_inflight);
		/* dump the inflight histogram */
		pdk_dumpskb(ppa, NULL, 0, 0); 

		return(0);
	}

	bp->b_type = PDK_BUF_TX;
	if (skb != 0) {
		bp->b_skb = skb;
		bp->history = NEWTX;
		return(bp);
	}

	if ( (bp->b_skb = alloc_skb(size, GFP_ATOMIC)) == NULL) {
		DEBUG1("pdk_buf_alloc(%d): alloc_skb failed.\n", ppa);
		PDK_FREE_BUF(hp, bp);
		return(NULL);
	}

	bp->history = NEWRX;
	return(bp);
}

#if (LINUX_VERSION_CODE < 0x020400)
#define	pci_map_single(_dev, _va, _len, _dir) (uint32_t)virt_to_bus((_va))
#define	pci_unmap_single(_dev, _pa, _len, _dir)
#endif

void pdk_buf_unmap(bp)
pdk_buf_t *bp;
{
	if (bp->b_type == PDK_BUF_RX)
		pci_unmap_single(znb_pdev[bp->b_ppa], bp->b_padr,bp->b_len, PCI_DMA_FROMDEVICE);
	else
		pci_unmap_single(znb_pdev[bp->b_ppa], bp->b_padr,bp->b_skb->len, PCI_DMA_TODEVICE);
}



/* receive_deinit() is followed by hal_stop(),
 * which buf_free() all receive bp's
 */
void
pdk_buf_free(bp)
pdk_buf_t *bp;
{
	pdk_buf_header_t *hp;
	uint32_t bh_state = 0;
	int32_t ppa;

	ppa = bp->b_ppa;
	hp = &pdk_buf_array[ppa];

	DEBUG3("pdk_buf_free(%d): state %d  bp=0x%x  type=%d  skb=0x%x\n", ppa, hp->bh_state, bp, bp->b_type, (uint32_t)bp->b_skb);

	pdk_lock_acquire(hp->bh_lock);

	if (hp->bh_state == BH_FLUSHING || hp->bh_state == BH_DEINITTING) {
		DEBUG2("pdk_buf_free(%d): deinitting %d %d\n", ppa, hp->bh_state, hp->bh_inflight);
		if (bp->b_skb != 0) {
			dev_free_skb(bp->b_skb);
			bp->b_skb = 0;
		}

		pdk_q_enq((pdk_q_t **) &hp->bh_header_freeq, (pdk_q_t *) bp);

		if (--hp->bh_inflight == 0) {
			bh_state = hp->bh_state;
		}

		pdk_lock_release(hp->bh_lock);

		if (bh_state != 0) {
			DEBUG3("pdk_buf_free(%d): calling pdk_buf_receive_deinit\n", ppa);
			pdk_buf_receive_deinit(ppa);
			if (bh_state == BH_DEINITTING) {
				DEBUG3("pdk_buf_free(%d): calling pdk_buf_deinit\n", ppa);
				pdk_buf_deinit(ppa);
			} else {
				DEBUG3("pdk_buf_free(%d): setting state=running\n", ppa);
				/* bh_state as in pdk_buf_init() */
				pdk_lock_acquire(hp->bh_lock);
				hp->bh_state = BH_RUNNING;
				pdk_lock_release(hp->bh_lock);
			}
		}
		return;
	}

	/* zxe_recv_complete() must call pdk_buf_free()
	 * after saving off mp and clearing b_skb.
	 */

	pdk_lock_release(hp->bh_lock);

	if (bp->b_skb != 0) {
		dev_free_skb(bp->b_skb);
		bp->b_skb = 0;
	}

	PDK_FREE_BUF(hp, bp);

	if (bp->b_type == PDK_BUF_RX) {
		if ((bp = pdk_buf_alloc(ppa, PDK_ETHERRECV_SIZE, 0)) != 0) {
			DEBUG3("pdk_buf_free(%d): refilling rx bp=0x%x\n", ppa, bp);
			bp->b_type = PDK_BUF_RX;
			bp->b_len = PDK_ETHERRECV_SIZE;
			(*zxe_array[ppa]->za_llr.pdk_recv)(ppa, bp);
		}
		else
			DEBUG1("pdk_buf_free(%d): buf_alloc fail\n", ppa);
	}
}

pdk_map_t *
pdk_buf_get_pmap(mapp, paddrp, lenp, errp)
pdk_map_t *mapp;
uint32_t *paddrp;
uint32_t *lenp;
int32_t *errp;
{
	dma_addr_t dma_handle;

	/* for RLK, bp has a header ptr. Want to return that paddr
	 * on first pmap call, skb paddr on second call
	 * works, but why ever use tail?  Won't tail == data for RX
	 * but maybe not for TX?
	*/
	if (mapp->b_type == PDK_BUF_RX) {
		dma_handle = pci_map_single(znb_pdev[mapp->b_ppa], mapp->b_skb->tail,mapp->b_len, PCI_DMA_FROMDEVICE);
		*lenp = (uint32_t) mapp->b_len;
		DEBUG3("pmap b_len: 0x%x\n", mapp->b_len);
	}
	else {
		dma_handle = pci_map_single(znb_pdev[mapp->b_ppa], mapp->b_skb->data,mapp->b_skb->len, PCI_DMA_TODEVICE);
		*lenp = (uint32_t) mapp->b_skb->len;
		DEBUG3("pmap skb len: 0x%x\n", mapp->b_skb->len);
	}

/*	*paddrp = pdk_ntbridge_dma_map(mapp->b_ppa,(uint32_t)dma_handle);*//*hamid*/
	*paddrp = (uint32_t)dma_handle;

	*errp=0;

	return(0);
}


int32_t
pdk_buf_receive_init(ppa)
int32_t ppa;
{
	pdk_buf_header_t *hp;
	pdk_buf_t *buf_queue = 0;
	pdk_buf_t *bp = 0;
	int32_t i;

	hp = &pdk_buf_array[ppa];

	DEBUG3("pdk_buf_receive_init(%d): initting.\n", ppa);

	pdk_lock_acquire(hp->bh_lock);

	if (hp->bh_state & BH_INITED) {
		DEBUG2("pdk_buf_receive_init(%d): Already initted\n", ppa);
		pdk_lock_release(hp->bh_lock);
		return(0);
	}

	if ( !(hp->bh_state & BH_RUNNING) ) {
		DEBUG1("pdk_buf_receive_init(%d): Out of state %d\n", ppa, hp->bh_state);
		pdk_lock_release(hp->bh_lock);
		return(PDK_OUT_STATE);
	}

	hp->bh_state |= BH_INITED;

	pdk_lock_release(hp->bh_lock);

	for (i=0;i<NUM_RECV_BUFS;i++) {
		if ((bp = pdk_buf_alloc(ppa, PDK_ETHERRECV_SIZE, 0)) == 0)
			break;

		bp->b_type = PDK_BUF_RX;
		bp->b_len = PDK_ETHERRECV_SIZE;

		pdk_q_enq((pdk_q_t **)&buf_queue, (pdk_q_t *)bp);
	}

	if (bp != 0) {
		while ((bp = (pdk_buf_t *)pdk_q_deq((pdk_q_t **)&buf_queue)) != 0)
			(*zxe_array[ppa]->za_llr.pdk_recv)(ppa, bp);

		return(0);
	}

	DEBUG1("pdk_buf_receive_init(%d):  pdk_buf_alloc failed, %d\n", ppa, i);

	pdk_lock_acquire(hp->bh_lock);
	hp->bh_state = BH_DEINITTING;
	pdk_lock_release(hp->bh_lock);

	while ((bp = (pdk_buf_t *)pdk_q_deq((pdk_q_t **)&buf_queue)) != 0)
		pdk_buf_free(bp);

	return(PDK_NO_MEM);
}


int32_t
pdk_buf_receive_deinit(ppa)
int32_t ppa;
{
	pdk_buf_header_t *hp;

	hp = &pdk_buf_array[ppa];

	if (hp->bh_state == BH_DEINIT) {
		return(0);
	}

	pdk_lock_acquire(hp->bh_lock);

	DEBUG2("pdk_buf_receive_deinit(%d): state %d\n", ppa, hp->bh_state);

	if ( !(hp->bh_state & BH_INITED) ) {
		DEBUG2("pdk_buf_receive_deinit(%d): Not initted\n", ppa);
		pdk_lock_release(hp->bh_lock);
		return(0);
	}

	/* Stage one of deinit - cause recycle packets to be freed */
	if (hp->bh_inflight != 0) {
		if (hp->bh_state & BH_RUNNING)
			hp->bh_state = BH_FLUSHING;

		DEBUG2("pdk_buf_receive_deinit(%d): Inflight %d state %d\n", ppa, hp->bh_inflight, hp->bh_state);
	}

	pdk_lock_release(hp->bh_lock);
	return(0);
}



void
pdk_buf_reserve(bp, len)
pdk_buf_t *bp;
uint32_t len;
{
	skb_reserve(bp->b_skb, len);
	return(0);
}



/* set_len called by receive thread, so setting b_len made sense.
 * TX thread already has skb->len set, so pmap works.
 * Except for load_setup_buf(), which ignores the length from pmap,
 * and fills the desc length with SETUP_FRAME_SIZE explicitly.
 * Does rlk mean b_len or skb-)len should be set here depending on b_type?
 * Added type check.  Hal loopback stuff needs it.
 */
uint32_t
pdk_buf_set_len(bp, len)
pdk_buf_t *bp;
uint32_t len;
{
	if (bp->b_skb == 0) {
		DEBUG1("b_skb is ZERO\n");
	}
	if (bp->b_type == PDK_BUF_TX) {
		/* jr This works only because none of the pdk code used
		   the tail ptr this far. Use the proper skb inlines. */
#if 0
		bp->b_skb->len = len;
#endif
		skb_put(bp->b_skb, len - bp->b_skb->len);
		DEBUG3("set_len skb len: 0x%x\n", bp->b_skb->len);
	}
	else {
		bp->b_len = len;
		DEBUG3("set_len b_len: 0x%x\n", bp->b_len);
	}
	return(0);
}


/*
SEE pdk_buf.h
uchar *pdk_buf_wptr(bp)
*/

uchar *
pdk_buf_rptr(bp)
pdk_buf_t *bp;
{
	return(bp->b_skb->data);
}


pdk_buf_t *
pdk_buf_pullup(bp, len)
pdk_buf_t *bp;
uint32_t len;
{
	return(bp);
}

uint32_t
pdk_buf_contiguous(bp, flags, p, lenp)
pdk_buf_t *bp;
uint32_t flags;
uchar *p;
uint32_t *lenp;
{
        uchar *limit;

	if (flags == PDK_BUF_READ)
		limit = bp->b_skb->end;
	else /* PDK_BUF_WRITE */
		limit = bp->b_skb->tail;
		limit = bp->b_skb->end; /* TMHXXX */
	/* head is absolute beginning of buffer,
	 * but data is start of unused.  Depending on usage, p may be head.
	 */
	if ((p < bp->b_skb->head) || (p > limit))
		return(PDK_BAD_PARAM);
	*lenp = limit - p;

	return(0);
}


/*
TODO TMHXXX avoid copy: make first buffer part of list header+data then
allocate seperate headers, and seperate fragments by pointing skb->data
to the first buffer.  Modify bp header to support this and not freeing first
bp until all is complete.  Support of multiple pmap calls is required.
Preallocation is also a nifty idea.
*/

pdk_buf_t *
pdk_buf_fragment(ppa, bp, num_frags, frag_len, header_len)
int32_t ppa;
pdk_buf_t *bp;
int32_t num_frags;
int32_t frag_len;
int32_t header_len;
{
	pdk_buf_header_t *hp;
	pdk_buf_t *frag_q, *frag_bp=0;
	uchar *datap;
	uint32_t len=frag_len+header_len;
	uint32_t total_len=0; /* rounded sum of all fragments */
	uint32_t orig_len; /* true sum of all fragments */
	uint32_t last_len; /* actual size of last fragment */
	uint32_t *lenp=0;
	int i;

	hp = &pdk_buf_array[ppa];
	frag_q = 0;
	datap = pdk_buf_rptr(bp);
	orig_len=pdk_buf_get_len(bp);

	for (i=0; i<num_frags; i++) {
		frag_bp = pdk_buf_alloc(ppa, PDK_ETHERRECV_SIZE, 0);
		if (frag_bp == 0) {
			DEBUG1("pdk_buf_fragment(%d): bp freeq empty\n", ppa);
			goto buf_frag_fail1;
		}
		lenp=(uint32_t *)pdk_buf_rptr(frag_bp);
		*lenp=len;
		frag_bp->b_skb->len = len;
		pdk_bcopy(datap, &frag_bp->b_skb->data[header_len], frag_len);
		datap += frag_len;
		total_len += frag_len;
		pdk_q_enq((pdk_q_t **) &frag_q, (pdk_q_t *) frag_bp);
	}
	last_len = frag_len - (total_len - orig_len);
	frag_bp->b_skb->len = last_len + header_len;
	*lenp=frag_bp->b_skb->len;
	pdk_buf_free(bp);
	return(frag_q);

buf_frag_fail1:
	if (frag_q) {
		while ((frag_bp = (pdk_buf_t *) pdk_q_deq((pdk_q_t **) &frag_q)) != 0) {
			pdk_buf_unmap(frag_bp);
			pdk_buf_free(frag_bp);
		}
	}

	pdk_buf_unmap(bp);

	return(0);
}

pdk_map_t *
pdk_buf_map(bp, flags, errp)
pdk_buf_t *bp;
uint32_t flags;
int32_t *errp;
{
	return((pdk_map_t *)bp);
}



uint32_t
pdk_buf_get_len(bp)
	pdk_buf_t *bp;
{
	return(bp->b_skb->len);
}



/* move forward read ptr -- pull off/consume data from head of buffer */

uint32_t
pdk_buf_movef_rptr(bp, len)
pdk_buf_t *bp;
uint32_t len;
{
#if 0
	/* move the data ptr? */
	if (bp->b_skb->data + len > bp->b_skb->end)
		return(PDK_BAD_PARAM);

	bp->b_skb->data += len;
	bp->b_skb->len -= len;
	return(0);
#endif
	return(skb_pull(bp->b_skb, len) ? 0 : PDK_BAD_PARAM);
}

/* move backward read ptr -- push to/prepend at head of buffer */

uint32_t
pdk_buf_moveb_rptr(bp, len)
pdk_buf_t *bp;
uint32_t len;
{
#if 0
	if (bp->b_skb->data - len < bp->b_skb->head)
		return(PDK_BAD_PARAM);

	bp->b_skb->data -= len;
	bp->b_skb->len += len;
#endif
	if(len > skb_headroom(bp->b_skb))
		return(PDK_BAD_PARAM);
	skb_push(bp->b_skb, len);
	return(0);
}

/* move forward write ptr -- append data at end of buffer */

uint32_t
pdk_buf_movef_wptr(bp, len)
pdk_buf_t *bp;
uint32_t len;
{
	skb_put(bp->b_skb, len);
	return(0);
}

/* move backward write ptr -- cut off data at end of buffer */

uint32_t
pdk_buf_moveb_wptr(bp, len)
pdk_buf_t *bp;
uint32_t len;
{
	if (bp->b_skb->len < len)
		return(PDK_BAD_PARAM);
	/* skb_trim(skb, new_length) */
	skb_trim(bp->b_skb, bp->b_skb->len - len);
	return(0);
}


#if NOT
pdk_buf_t *
pdk_buf_link(bp, bp1)
pdk_buf_t *bp;
pdk_buf_t *bp1;
{
	return(bp);
}

uint32_t
pdk_buf_copy(fbp, tbp)
pdk_buf_t *fbp;
pdk_buf_t *tbp;
{
}

#endif /* NOT */

