/*==========================================================================
**
**	./pdk/src/drivers/hal/hal.c
**
**	Hardware abstraction layer
**
**==========================================================================
**
**	(c) Copyright 1998, 1999 ZNYX Corporation
**	All Rights Reserved.
**
**==========================================================================
**
**	Modification History:
**	sw	04/08/98 	Created
**	ram	10/05/98 	Performance optimized.  Non-chain mode.
**	ram	03/31/99	Added 470 support.  Cleaned out legacy product
**	ram&mmc	05/27/99	Changed 21143 autonegotiation to interrupt. 
**				Added BM5208 PHY super-isolate and reset. 
**				Much cleanup
**	mmc 06/14/99		Support PHY reset on 470.
**	mmc 01/27/00    Added pdk_hal_version string.
**	hamid 03/25/02    Addec TX byte count for the ifconfig and SNMP
**==========================================================================
**
**	Notes:
**	The following Symbols can be defined in pdk_port.h:
**
**	TX_COMPLETE_CHECK_THRESHOLD - transmit complete processing threshold
** 	FREE_TXD_THRESHOLD -  Interrupt if this number of xmit ds are consumed
**			Set to zero to interrupt on every xmit complete
**	DYNAMIC_THRESHOLD - ??
**	MAXIMUM_MULTICAST_ADDRESSES - defaults to 14 if not defined
**	SYS_DESC_SYNC - to enable calls to pdk_mem_desc_sync and descriptor_sync
**	PDK_BUF_SYNC - to enable calls to pdk_buf_sync	
**	QNX4 - to generate far references and FP_OFF
**	PDK_NO_DEBUG - to not generate pdk_print statements
**	DEBUG_CHECK_RING - to generate code to check descriptor rings at init.
**	LEVELED_INTERRUPTS - to allow checking for receives on all ports of a
**		multiport card before other conditions.
**	HAL_AVOID_ISR_READ - to test for receive complete by comparing dequeue
**		and enqueue pointers, avoiding a read of the chip isr reg.
**	SHADOW_CSR7 - keep track of interrupt mask register, if 0, don't
**		check device for interrupts in pdk_hal_iac
**	LOOPBACK_21143 - Do loopback test for 21143 with 6611 type PHY.
**	
**==========================================================================
*/

#include <pdk_port.h>
#include <pdk_global.h>
#include <pdk_types.h>
#include <pdk_hton.h>

#include <pdk_lib.h>
#include <pdk_buf.h>
#include <pdk_print.h>
#include <pdk_queue.h>
#include <pdk_lock.h>
#include <pdk_mem.h>
#include <pdk_time.h>
#include <pdk_errno.h>
#include <pdk_reg.h>
#include <pdk_hal.h>

#include <pdk_layer.h>
#include <pdk_constant.h>

#include <pdk_br554.h>
#include <pdk_psdb.h>

#include "deviceid.h"
#include "dec21040.h"
#include "dp83840.h"
#include "ics1890.h"
#include "bcm5000.h"

#ifndef	SYS_INSTR_SYNC
#define	SYS_INSTR_SYNC()
#endif

#define SROM_SIZE 128

#define H_READ  1
#define H_WRITE 2

#define	H_PROMISCUOUS_PHYSICAL	1
#define	H_PROMISCUOUS_MULTICAST	2
#define	H_BROADCAST		4

#define	SETUP_FRAME_SIZE	192
#define	SETUP_BUF_ADDR_SIZE	12
#define	MAX_SETUP_BUF_ADDRS	16
#define	SETUP_BUF_ADDR_ADD	1
#define	SETUP_BUF_ADDR_DEL	2

#ifdef LOOPBACK_21143
#define ZXE_LOOPBACK_TPINT 11
#define ZXE_LOOPBACK_TXINT 12
#define ZXE_LOOPBACK_TXEXT 13
#endif

#ifdef SYS_DESC_SYNC
#undef SYS_DESC_SYNC
#define	SYS_DESC_SYNC(handlep, deqaddr, enqaddr, type) (pdk_mem_desc_sync(handlep, (char *)deqaddr, (char *) enqaddr, type));
#define	DESCRIPTOR_SYNC(lp, flag) descriptor_sync(lp, flag)
#else
#define	SYS_DESC_SYNC(handlep, deqaddr, enqaddr, type)
#define	DESCRIPTOR_SYNC(lp, flag)
#endif

#ifdef PDK_BUF_SYNC
#undef PDK_BUF_SYNC
#define PDK_BUF_SYNC(bp, len, flags) (pdk_buf_sync(bp, len, flags))
#else
#define PDK_BUF_SYNC(bp, len, flags)
#endif

#ifndef	TX_COMPLETE_CHECK_THRESHOLD
#define	TX_COMPLETE_CHECK_THRESHOLD 4
#endif


/* Hardware abstraction layer interface */

char *pdk_hal_version="ZNYX HAL Version M11 01";

#ifndef RXD_SIZE
#define RXD_SIZE	64
#endif
#ifndef TXD_SIZE
#define TXD_SIZE	64
#endif
int32_t num_receive_desc = RXD_SIZE;
int32_t num_transmit_desc = TXD_SIZE;
int32_t zxe_loopback_delay = 1000;

#ifndef FREE_TXD_THRESHOLD
#define FREE_TXD_THRESHOLD (num_transmit_desc >> 2)
#endif

/* Ring descriptor structure */

typedef struct {
	uint32_t		tdes0;
	uint32_t		tdes1;
	uint32_t		tdes2;
	uint32_t		tdes3;
} txd_t;

typedef struct {
	uint32_t		rdes0;
	uint32_t		rdes1;
	uint32_t		rdes2;
	uint32_t		rdes3;
} rxd_t;

#ifdef QNX4
#define QNX_FAR far
#define QNX_OFF FP_OFF
#else
#define QNX_FAR
#define QNX_OFF
#endif

/* # of int32_t words to pad -> 128/4 - 4 */
#ifndef DESC_PAD_SIZE
#define DESC_PAD_SIZE 28
#endif /* DESC_PAD_SIZE */
typedef struct desc_s {
	union {
		rxd_t r;
		txd_t t;
	} d;
	union {
		struct {
			struct desc_s QNX_FAR	*next;
			pdk_buf_t	*bp;
		} ee;
		uint32_t pad[DESC_PAD_SIZE];
	} e;
	
} desc_t;

#define desc_t    desc_t QNX_FAR

/* Descriptor skip length must be set to 2 int32_t words (for next and bp) */
#define CSR0_DSL_VALUE (DESC_PAD_SIZE<<2)


#define RDES0	d.r.rdes0
#define RDES1	d.r.rdes1
#define RDES2	d.r.rdes2
#define RDES3	d.r.rdes3

#define TDES0	d.t.tdes0
#define TDES1	d.t.tdes1
#define TDES2	d.t.tdes2
#define TDES3	d.t.tdes3

#define NEXT	e.ee.next
#define BP	e.ee.bp

#define MAX_TX_FRAGS	64

#define NIC_ISR_RECEIVE 	(CSR5_RPS|CSR5_RI)
#define NIC_ISR_TRANSMIT	(CSR5_UNF|CSR5_TPS|CSR5_TI)
#define NIC_ISR_OTHER		(CSR5_SE|CSR5_LNF|CSR5_GTE|CSR5_GPI|CSR5_LC|CSR5_LNP_ANC)
/*
#define NIC_ISR_OTHER		(CSR5_SE|CSR5_LNF|CSR5_GTE|CSR5_GPI)
*/

#define NIC_ISR_BITS		(NIC_ISR_RECEIVE|NIC_ISR_TRANSMIT|NIC_ISR_OTHER)

#define CSR7_CFG		(CSR7_NIM|CSR7_AIM|NIC_ISR_BITS)


/* Adapter description structure */
typedef struct zxe_adapterinfo {
	uint32_t	AdapterId;
	uint32_t	NicType;
	uint32_t	NumChannels;
	uint32_t	PhyId;
	uint32_t	AIFlags;
} zxe_adapterinfo_t;

#define AI_LEDS           1
#define AI_SHARED_PHYLOCK 2
#define AI_SHARED_LEDS    4
#define AI_GEP0_LPBK      8
#define AI_GEP2_PHYRESET 16
#define AI_SHARED_SROM   32


#define LOCAL_ADAPTER_ALLOCATED 0
#define LOCAL_ADAPTER_STOPPED 1
#define LOCAL_ADAPTER_RUNNING 2
#define LOCAL_ADAPTER_RELEASED 3

#define OPERATIONAL_LED 0x1
#define INT_FAULT_LED	0x2
#define EXT_FAULT_LED	0x4
#define FAILOVER_FAULT_LED	0x8

typedef struct _local_adapter {
	uint32_t		state;
	uchar			*srom_image;

	int32_t			srom_image_valid;

	uchar			*setup_buf;

	pdk_desc_handle_t	*rx_descriptor_handlep;
	uint32_t		rx_paddr_base;

	desc_t			*rx_queue;
	desc_t			*rx_dequeue_here;
	desc_t			*rx_enqueue_here;

	int32_t			rx_num_bufs;
	int32_t			rx_lwm;
	int32_t			rx_hwm;
	uint32_t		rx_depleted;
	uint32_t		rx_descriptor_count;
	uint32_t		rx_icount;

	pdk_desc_handle_t	*tx_descriptor_handlep;
	uint32_t		tx_paddr_base;

	desc_t			*tx_queue;
	desc_t			*tx_dequeue_here;
	desc_t			*tx_enqueue_here;

	int32_t			free_tx_descriptor_count;
	int32_t			tx_lwm;
	int32_t			tx_depleted;
	uint32_t		tx_descriptor_count;
	uint32_t		tx_icount;

	int32_t			tx_pending_count; /* for periodic interrupts */
	uint32_t		tx_pending_threshold; /* interrupt when this number is hit */
	uint32_t		other_icount;
	
	pdk_lock_t		*rx_descriptor_lock;
	pdk_lock_t		*tx_descriptor_lock;

	uint32_t		led_local_settings;

	uint32_t		pdk_xmit_complete_check;
	uint32_t		fault_notification;

	uint32_t		AdapterId;

	zxe_adapterinfo_t	Info;

	uchar			PermanentNetworkAddress[PDK_ETHERADDRL];
	uchar			CurrentNetworkAddress[PDK_ETHERADDRL];

	uint32_t		PhyType;

	uint32_t		TxFifoUnderflowGauge;
#ifdef DYNAMIC_THRESHOLD
	uint32_t		TxFifoThreshold;
#endif
	uint32_t		OperationModeBase;
	uint32_t		OperationMode;
#ifdef SHADOW_CSR7
	uint32_t		csr7;
#endif
	uint32_t		LinkSpeed;
	uint32_t		ConfiguredMediaType;
	uint32_t		CurrentMediaType;
	uint32_t		AutoNegotiatedMediaType;
	uint32_t		PhyControl;
#define MEDIA_CONNECTED 1
#define MEDIA_DISCONNECTED 2
	uint32_t		MediaConnectState;
	uint32_t		PollTimeout;
	uint32_t		PacketFilter;
	uint32_t		RxTimeoutCounter;
	uint32_t		TxBufferThreshold;
	uint32_t		TxTimeoutCounter;
	int32_t			MCAddressCount;
#ifndef ZXE_MAXIMUM_MULTICAST_ADDRESSES
#define ZXE_MAXIMUM_MULTICAST_ADDRESSES 14
#endif
	uchar			MCAddresses[ZXE_MAXIMUM_MULTICAST_ADDRESSES][PDK_ETHERADDRL];

	/* adapter statistics */

	pdk_hal_stat_t		Statistics;
	int32_t			ppa;
	pdk_layer_routines_t	layer_routines;
#ifdef PDK_HAL_SPEEDO
	void			*hal_speedometer_timeout_id;
	pdk_hal_stat_t		prev_Statistics;/* previous stats */
	hal_speedometer_t	hsp;
	hal_speedometer_t	set_params_gte_hsp;	/* greater than */
	hal_speedometer_t	set_params_lt_hsp;	/* less than or equal */
	uint32_t		prev_rx_icount;
	uint32_t		prev_tx_icount;
	uint32_t		tx_bytes;
	uint32_t		rx_bytes;
	uint32_t		speedometer_poll_rate;
#endif
} local_adapter_t;

local_adapter_t *local_adapters[NUM_PORTS];

/* MEDIA */

/*
 * Transmit FIFO configurations for DEC 21140
 */

uint32_t	TxFifoConfigurations[] = {
	CSR6_TR_72 | CSR6_MDR, CSR6_TR_96 | CSR6_MDR, CSR6_TFF_128,
	CSR6_TFF_256, CSR6_TFF_512, CSR6_TFF_1024,
	CSR6_TFF_1024 | CSR6_FP };
#define NUM_TXFIFOCONFIG	7

/* Adapter description structure */

zxe_adapterinfo_t ZxePhyInformation[] = {
	{	ADAPTER_ZX345, DEVICE_DEC21140, 1, DEVICE_DP83840, 0 },
	{	ADAPTER_ZX345F, DEVICE_DEC21140, 1, DEVICE_DP83840, 0 },
	{	ADAPTER_ZX345X2, DEVICE_DEC21140, 1, DEVICE_ICS1890, 0 },
	{	ADAPTER_ZX346, DEVICE_DEC21140, 4, DEVICE_ICS1890, 0 },
	{	ADAPTER_ZX348, DEVICE_DEC21140, 2, DEVICE_DP83840, 0 },
	{	ADAPTER_ZX345Q, DEVICE_DEC21143, 1, DEVICE_QS6611, 0 },
	{	ADAPTER_ZX346Q, DEVICE_DEC21143, 4, DEVICE_QS6611, 0 },
	{	ADAPTER_ZX348Q, DEVICE_DEC21143, 2, DEVICE_QS6611, 0 },
	{	ADAPTER_ZX414, DEVICE_DEC21143, 4, DEVICE_QS6611, 0 },
	{	ADAPTER_ZX412, DEVICE_DEC21143, 2, DEVICE_QS6611, 0 },
	{	ADAPTER_ZX212, DEVICE_DEC21143, 2, DEVICE_QS6611, 0 },
	{	ADAPTER_ZX214, DEVICE_DEC21143, 4, DEVICE_QS6611, 0 },
	{	ADAPTER_ZX442, DEVICE_DEC21143, 2, DEVICE_QS6611, 0 },
	{	ADAPTER_ZX444, DEVICE_DEC21143, 4, DEVICE_QS6611, 0 },
	{	ADAPTER_ZX474, DEVICE_DEC21143, 4, DEVICE_BCM5208, AI_LEDS|AI_SHARED_PHYLOCK|AI_SHARED_LEDS|AI_SHARED_SROM },
	{	ADAPTER_ZX476, DEVICE_DEC21143, 6, DEVICE_BCM5208, AI_LEDS|AI_SHARED_PHYLOCK|AI_SHARED_LEDS|AI_SHARED_SROM},
	{	ADAPTER_ZX478, DEVICE_DEC21143, 8, DEVICE_BCM5208, AI_LEDS|AI_SHARED_PHYLOCK|AI_SHARED_LEDS|AI_SHARED_SROM },
	{	ADAPTER_ZX222, DEVICE_DEC21143, 2, DEVICE_BCM5208, AI_GEP2_PHYRESET },
	{	ADAPTER_ZX374, DEVICE_DEC21143, 4, DEVICE_QS6611, AI_LEDS | AI_GEP0_LPBK},
	{	ADAPTER_ZX372, DEVICE_DEC21143, 2, DEVICE_QS6611, AI_LEDS | AI_GEP0_LPBK},
	{	ADAPTER_ZX244, DEVICE_DEC21143, 4, DEVICE_QS6611, AI_LEDS | AI_GEP0_LPBK},
	{	ADAPTER_ZX424, DEVICE_DEC21143, 4, DEVICE_QS6611, AI_GEP0_LPBK},
	{	ADAPTER_ZX422, DEVICE_DEC21143, 2, DEVICE_QS6611, AI_GEP0_LPBK},
	{ 0 }
};

typedef struct local_static {
	pdk_lock_t	*PhyLock;
	uint32_t	PhyAddress;
	uint32_t	led_shadow;
} local_static_t;

local_static_t local_statics[NUM_PORTS];

#define	PHYLOCK(ppa) { \
	pdk_lock_t *_PhyLock = local_statics[ppa].PhyLock; \
	if (_PhyLock) pdk_lock_acquire(_PhyLock); \
}
#define	PHYUNLOCK(ppa) { \
	pdk_lock_t *_PhyLock = local_statics[ppa].PhyLock; \
	if (_PhyLock) pdk_lock_release(_PhyLock); \
}

int32_t zxe_hs_active = 1;

#ifdef SHADOW_CSR7
#define	SET_CSR7(ppa,val) pdk_shadow_put32(ppa,CSR7,val,&(lp->csr7));
#else
#define SET_CSR7(ppa,val) pdk_reg_put32(ppa,CSR7,val)
#endif

uint32_t pdk_hal_iac(int32_t ppa);
uint32_t pdk_hal_intr(int32_t ppa, uint32_t iac);
uint32_t pdk_hal_xmit(int32_t ppa, pdk_buf_t *bp);
uint32_t pdk_hal_recv(int32_t ppa, pdk_buf_t *bp);
void hal_psdb_return_state(local_adapter_t *lp, pdk_psdb_hal_state_t *phsp);
#ifdef PDK_HAL_SPEEDO
/* convert ms to usec */
#define HAL_SPEEDOMETER_TICKS (lp->speedometer_poll_rate*1000)
#define HAL_SPEEDOMETER_RATE(val) \
	((((uint32_t) (val)) < (0xffffffff/1000)) ? \
		 ((((uint32_t) (val))*1000)/lp->speedometer_poll_rate) : \
		 ((((uint32_t) (val))/lp->speedometer_poll_rate)*1000))
void hal_psdb_return_speed(local_adapter_t *lp, pdk_psdb_hal_speedometer_get_t *phsp);
void hal_speedometer_timeout(void *arg);
#endif
void hal_psdb_return(local_adapter_t *lp, uint32_t type, uint32_t return_val);
uint32_t hal_psdb(pdk_psdb_rhdr_t *record);
uint32_t hal_get_state(int32_t ppa, hal_state_t *hsp);
uint32_t hal_get_stats(int32_t ppa, pdk_hal_stat_t *stat);

#ifdef PDK_HAL_SPEEDO
uint32_t hal_get_speed(int32_t ppa, hal_speedometer_t *hsp);
uint32_t hal_set_speedometer_params(int32_t ppa, hal_speedometer_t *hsp);
#endif

static pdk_layer_routines_t my_routines = {
	pdk_hal_iac, /* pdk_iac */
	pdk_hal_intr, /* pdk_intr */
	pdk_hal_xmit, /* pdk_xmit */
	pdk_hal_recv, /* pdk_recv */
	pdk_hal_ctl, /* pdk_ctl */
	0, /* pdk_xmit_complete */
	0, /* pdk_recv_complete */
	0  /* pdk_status */
};


/* Local routines */

uint32_t get_network_address(int32_t ppa);
void ZxeStartTxAndRx(int32_t ppa);
int32_t ZxePreConfigureMedia(int32_t ppa);
int32_t ZxeConfigureMedia(int32_t ppa);
uint32_t ZxeSetupAdapterInfo(int32_t ppa);
void process_recv_complete(int32_t ppa);
void process_xmit_complete(local_adapter_t *lp);
void ZxeMediaPoll(int32_t ppa, uint32_t isr_value);
uint32_t pdk_modify_setup_buf(int32_t ppa, uchar *addr, uint32_t flags);
uint32_t load_setup_buf(int32_t ppa);
void init_setup_buf(int32_t ppa);
uint32_t pdk_hal_stop(int32_t ppa);
int32_t add_multicast_address(int32_t ppa, uchar *addr);
int32_t del_multicast_address(int32_t ppa, uchar *addr);
void zxe_operational_led(int32_t ppa, int32_t on);
void zxe_internal_fault_led(int32_t ppa, int32_t on);
void zxe_external_fault_led(int32_t ppa, int32_t on, int32_t failover);

int32_t ZxeReadPHY(int32_t ppa, int32_t Addr, uint16_t *Ptr);
int32_t ZxeWritePHY(int32_t ppa, int32_t Addr, uint16_t Word);

int32_t zxe_loopback_test(int32_t ppa);
int32_t zxe_test_fault(int32_t ppa);

int32_t pdk_hal_debug = 1;
#ifdef PDK_NO_DEBUG
#define DEBUG0
#define DEBUG1
#define DEBUG2
#define DEBUG3
#define DEBUG4
#else
#define DEBUG0                      pdk_print
#define DEBUG1 if (pdk_hal_debug>0) pdk_print
#define DEBUG2 if (pdk_hal_debug>1) pdk_print
#define DEBUG3 if (pdk_hal_debug>2) pdk_print
#define DEBUG4 if (pdk_hal_debug>3) pdk_print
#endif

pdk_layer_routines_t *
pdk_hal_register(ppa, up, down)
int32_t ppa;
pdk_layer_routines_t *up;
pdk_layer_routines_t *down;
{
	local_adapter_t *lp;

	if ((lp = local_adapters[ppa]) == 0) {
		DEBUG1("pdk_hal_register(%d):  Local adapter not initialized\n", ppa);
		return(0);
	}

	if (down) {
		DEBUG1("pdk_hal_register(%d):  No layers should be below the HAL\n", ppa);
		return(0);
	}

	if (up) {
		lp->layer_routines.pdk_xmit_complete = up->pdk_xmit_complete;
		lp->layer_routines.pdk_recv_complete = up->pdk_recv_complete;
		lp->layer_routines.pdk_status = up->pdk_status;
	}

	return(&my_routines);
}

uint32_t
pdk_hal_global_init()
{
	int_t i;

	DEBUG4("pdk_hal_global_init\n");

	for (i=0; i<NUM_PORTS; i++) {
		local_adapters[i] = 0;
		local_statics[i].PhyLock = 0;
		local_statics[i].PhyAddress = -1;
		local_statics[i].led_shadow = 0;
	}

	return(0);
}

uint32_t
pdk_hal_global_deinit()
{
	DEBUG4("pdk_hal_global_deinit\n");
	return(0);
}

uint32_t
hal_hdwr_init(uint32_t ppa)
{
	local_adapter_t *lp;
	local_static_t *ls;
	pdk_lock_t *PhyLock;
	int32_t i,p;
	
	lp = local_adapters[ppa];

	/* Take the board out of snooze and sleep modes */
	pdk_reg_config_put32(ppa, CFDA, 0);

	/* Reset the board and turn off interrupts */
	pdk_reg_put32(ppa, CSR0, CSR0_SWR);
	pdk_time_delay(2);
	SET_CSR7(ppa, 0);

	/* Get network address and AdapterId from srom */
	if (get_network_address(ppa)) {
		DEBUG1("pdk_hdwr_init(%d): get_network_address failed\n", ppa);
		pdk_hal_deinit(ppa);
		return(PDK_OUT_STATE);
	}

	ZxeSetupAdapterInfo(ppa);

	/*
	 * Allocate the PhyLock if necessary
	 */
	ls = &local_statics[ppa];

	if (ls->PhyLock == 0) {
		if ((PhyLock = pdk_lock_alloc(ppa, PDK_LOCK_INTR)) == 0) {
			pdk_hal_deinit(ppa);
			return(PDK_SYS_ERR);
		}

		if (lp->Info.AIFlags & AI_SHARED_PHYLOCK) {
		/* For the 470 series, use one lock for both broadcom phy(s) */
			int num_ports = lp->Info.NumChannels;
			for (i=0;i<num_ports;i++) {

				p = pdk_ntbridge_unit_ppa(ppa, i);

				if (p == -1)
					continue;

				local_statics[p].PhyLock = PhyLock;
			}
		} else
			ls->PhyLock = PhyLock;
	}

	if (lp->PhyType == DEVICE_ICS1890) {
		/* Inhibit transmission or negotiation */
		ZxeWritePHY(ppa, ICS1890_MIICR_REG, (uint16_t)ICS1890_MIICR_PWRDN);
	} 

	return(0);
}

uint32_t
pdk_hal_init(ppa)
int32_t ppa;
{
	local_adapter_t *lp;
	pdk_desc_handle_t *handlep;
	void QNX_FAR *vaddr;
	uint32_t fpaddr;
	desc_t *dd;
	desc_t *fdd;
	int32_t err;

	DEBUG4("pdk_hal_init(%d): Called\n", ppa);

	/* If there is already a local adapter structure */
	if ((lp = local_adapters[ppa]) != 0) {
		pdk_hal_deinit(ppa);
	}

	/* Get a local adapter structure */
	local_adapters[ppa] = (local_adapter_t *)pdk_mem_zalloc(ppa, sizeof(local_adapter_t));

	if ((lp = local_adapters[ppa]) == 0) {
		return(PDK_NO_MEM);
	}

	lp->ppa = ppa;
	lp->MediaConnectState = MEDIA_DISCONNECTED;
	lp->free_tx_descriptor_count = num_transmit_desc;
	lp->tx_lwm = num_transmit_desc;
	lp->rx_lwm = num_receive_desc;
	lp->tx_pending_threshold = (num_transmit_desc + 15)>>4;


	/* Allocate locks */
	if ((lp->rx_descriptor_lock = pdk_lock_alloc(ppa, PDK_LOCK_INTR)) == 0) {
		pdk_hal_deinit(ppa);
		return(PDK_SYS_ERR);
	}

	if ((lp->tx_descriptor_lock = pdk_lock_alloc(ppa, PDK_LOCK_INTR)) == 0) {
		pdk_hal_deinit(ppa);
		return(PDK_SYS_ERR);
	}

	/* Allocate setup buffer */ 
	if ((lp->setup_buf = (uchar *)pdk_mem_zalloc(ppa, SETUP_FRAME_SIZE)) == 0) {
		pdk_hal_deinit(ppa);
		return(PDK_NO_MEM);
	}

	/* Allocate SROM image */ 
	if ((lp->srom_image = (uchar *)pdk_mem_zalloc(ppa, SROM_SIZE)) == 0) {
		pdk_hal_deinit(ppa);
		return(PDK_NO_MEM);
	}

	/* Allocate receive descriptor space */
	handlep = pdk_mem_desc_alloc(ppa, 1, num_receive_desc * sizeof(*dd));
	if (handlep == 0) {
		pdk_hal_deinit(ppa);
		return(PDK_SYS_ERR);
	}

	/* Build receive ring */
	lp->rx_descriptor_handlep = handlep;

	handlep = pdk_mem_desc_get_pmap(handlep, &vaddr, &fpaddr, &err);
	if (err) {
		pdk_hal_deinit(ppa);
		return(PDK_SYS_ERR);
	}

	lp->rx_paddr_base = fpaddr;
	fdd = (desc_t *)vaddr;
	lp->rx_queue = fdd;
	lp->rx_enqueue_here = fdd;
	lp->rx_dequeue_here = fdd;

	for (dd=fdd; dd<(fdd+num_receive_desc); dd++)
		dd->NEXT = dd+1;

	dd = fdd+(num_receive_desc-1);
	dd->NEXT = fdd;
	dd->RDES1 = RDES1_RER;

	/* Allocate Transmit Descriptor space */
	handlep = pdk_mem_desc_alloc(ppa, 1, num_transmit_desc * sizeof(*dd));
	if (handlep == 0) {
		pdk_hal_deinit(ppa);
		return(PDK_SYS_ERR);
	}

	/* Build transmit ring */
	lp->tx_descriptor_handlep = handlep;

	handlep = pdk_mem_desc_get_pmap(handlep, &vaddr, &fpaddr, &err);
	if (err) {
		pdk_hal_deinit(ppa);
		return(PDK_SYS_ERR);
	}
	lp->tx_paddr_base = fpaddr;

	fdd = (desc_t *)vaddr;
	lp->tx_enqueue_here = fdd;
	lp->tx_dequeue_here = fdd;
	lp->tx_queue = fdd;

	for (dd=fdd; dd<(fdd+num_transmit_desc); dd++)
		dd->NEXT = dd+1;

	dd = fdd+(num_transmit_desc-1);
	dd->NEXT = fdd;
	dd->TDES1 = TDES1_TER;

	if (!zxe_hs_active) {	/* Host processor not active? */
		return(0);
	}

	return(hal_hdwr_init(ppa));
}

uint32_t
pdk_hal_deinit(ppa)
int32_t ppa;
{
	local_adapter_t *lp;
	pdk_lock_t *PhyLock;
	int32_t i, p;

	DEBUG2("pdk_hal_deinit(%d): Called\n", ppa);

	if ((lp = local_adapters[ppa]) == 0)
		return(PDK_OUT_STATE);

	if (lp->state == LOCAL_ADAPTER_RUNNING) {
		DEBUG1("pdk_hal_deinit(%d):  Adapter is running - stopping before deiniting\n", ppa);
		pdk_hal_ctl(ppa, PDK_HAL_STOP_AND_RESET, 0);
	}

	/* Clear lp from local adapter structure */
	local_adapters[ppa] = 0;

#ifdef PDK_HAL_SPEEDO
	if (lp->hal_speedometer_timeout_id != 0) {
		pdk_untimeout(lp->hal_speedometer_timeout_id);
		lp->hal_speedometer_timeout_id = 0;
	}
#endif

	/* Free setup buffer */ 
	if (lp->setup_buf) {
 		pdk_mem_free(ppa, lp->setup_buf, SETUP_FRAME_SIZE);
	}

	/* Free SROM image */ 
	if (lp->srom_image) {
		pdk_mem_free(ppa, lp->srom_image, SROM_SIZE);
	}

	if (lp->rx_descriptor_handlep)
		pdk_mem_desc_free(lp->rx_descriptor_handlep);

	if (lp->rx_descriptor_lock) {
		pdk_lock_free(lp->rx_descriptor_lock);
	}

	/* Free Transmit resources */
	if (lp->tx_descriptor_handlep)
		pdk_mem_desc_free(lp->tx_descriptor_handlep);

	if (lp->tx_descriptor_lock) {
		pdk_lock_free(lp->tx_descriptor_lock);
	}

	PhyLock = local_statics[ppa].PhyLock;

	if (lp->Info.AIFlags & AI_SHARED_PHYLOCK) {
		int num_ports = lp->Info.NumChannels;
		for (i=0;i<num_ports;i++) {
			p = pdk_ntbridge_unit_ppa(ppa, i);
			if (p == -1)
				continue;

			if (local_adapters[p] != 0)
				break;
		}

		/* Clear lock pointers */
		if (i == num_ports) {
			for (i = 0; i<num_ports;i++) {
				p = pdk_ntbridge_unit_ppa(ppa, i);
				if (p == -1)
					continue;

				local_statics[p].PhyLock = 0;
			}

			pdk_lock_free(PhyLock);
		}
	}
	else {
		local_statics[ppa].PhyLock = 0;
		if (PhyLock)
			pdk_lock_free(PhyLock);
	}

	pdk_mem_free(ppa, (uchar *) lp, sizeof(local_adapter_t));

	return(0);
}

void
hal_release_phy(uint32_t ppa)
{
	local_adapter_t *lp = local_adapters[ppa];

	if (lp->Info.AIFlags & AI_SHARED_PHYLOCK) {
		zxe_operational_led(ppa, 0); /* Releases reset, sets SUPER_ISOLATE */
	} else if (lp->Info.AIFlags & AI_GEP2_PHYRESET) {
		pdk_reg_put32(ppa, CSR15, CSR15_142_CWE|0x40000);	/* GEP2 output 0*/
	}
}

int32_t
pdk_hal_first_start(ppa)
int32_t ppa;
{
	local_adapter_t *lp = local_adapters[ppa];
	local_static_t *ls;
	pdk_lock_t *PhyLock;
	uint32_t err;

	/* Init hardware if not done already */

	if (!lp->PermanentNetworkAddress[0]
		&& !lp->PermanentNetworkAddress[1]
		&& !lp->PermanentNetworkAddress[2])	/* Check for OID == 0 */
	{
		err = hal_hdwr_init(ppa);
		if (err)
			return(err);
	} else {

		/* Take the board out of snooze and sleep modes */
		pdk_reg_config_put32(ppa, CFDA, 0);	/* Needed for hot swap */

		/* Reset chip */
		pdk_reg_put32(ppa, CSR0, CSR0_SWR);
		pdk_time_delay(2);
	}

	/* Make sure PHY reset is cleared and shared PHYs isolated */
	hal_release_phy(ppa);

	return(0);
}


uint32_t
pdk_hal_start(ppa, do_loopback)
int32_t ppa;
int32_t do_loopback;
{
	local_adapter_t *lp;
	uint32_t err;
	uint32_t pbl, cls, cfcs;
	uint16_t phy_isr_value;

	if ((lp = local_adapters[ppa]) == 0)
		return(PDK_OUT_STATE);

	if (lp->state == LOCAL_ADAPTER_RELEASED) {
		DEBUG2("znb%d is RELEASED, will not be started\n",ppa);
		return(0);				/* Don't start this one now */
	}

	if (lp->state == LOCAL_ADAPTER_ALLOCATED) {
		if ((err = pdk_hal_first_start(ppa)) != 0)
			return(err);

		lp->state = LOCAL_ADAPTER_STOPPED;
	}

	if (lp->state != LOCAL_ADAPTER_STOPPED) {
		DEBUG1("pdk_hal_start(%d): Out of state(%d)\n", ppa, lp->state);
		return(PDK_OUT_STATE);

	}

	if (lp->rx_enqueue_here != lp->rx_dequeue_here) {
		DEBUG1("pdk_hal_start(%d): Out of state, receive buffers present\n", ppa);
		return(PDK_OUT_STATE);
	}

	if (lp->tx_enqueue_here != lp->tx_dequeue_here) {
		DEBUG1("pdk_hal_start(%d): Out of state, transmit buffers present\n", ppa);
		return(PDK_OUT_STATE);
	}

	init_setup_buf(ppa);

	/* reset chip */
	pdk_reg_put32(ppa, CSR0, CSR0_SWR);
	pdk_time_delay(2);

	if (ZxePreConfigureMedia(ppa) == 0) {
		DEBUG1("pdk_hal_start(%d): PreMedia configuration failed\n", ppa);
		return(PDK_HW_ERR);
	}

	pbl = pdk_reg_config_get32(ppa, CFLT) & 0x3f;
	/* Translate cache line sizes of 8,16, or 32 to appropriate bit
	 * patterns of 01, 10, or 11 respectively */
	cls = pbl >> 3;
	if (cls == 0x4) {
		pbl = 0;
		cls = 0x3;
	}
	pbl <<= 9;					/* Set burst length to two cache lines */
	cls <<= 14;
	/* End translation */
	cfcs = pdk_reg_config_get32(ppa, CFCS);
	/* Translate PCI Memory Write and Invalidate bit to CSR0 equivalent - 
	 * 0x10 --> 0x01000000 */
	cfcs <<= 20;
	cfcs &= CSR0_WIE;
	DEBUG2("pdk_hal_start(%d): CLS/BURST bits 0x%x CFCS bits 0x%x\n",
				ppa, cls|pbl, cfcs);
	pdk_reg_put32(ppa, CSR0, cfcs|pbl|cls|CSR0_RME|CSR0_RLE|CSR0_DSL_VALUE);
	pdk_time_delay(1);
	SET_CSR7(ppa,0);
	pdk_time_delay(1);

	/* load ring addresses */
	pdk_reg_put32(ppa, CSR3, lp->rx_paddr_base);
	pdk_reg_put32(ppa, CSR4, lp->tx_paddr_base);

	lp->rx_enqueue_here = lp->rx_dequeue_here = lp->rx_queue;
	lp->tx_enqueue_here = lp->tx_dequeue_here = lp->tx_queue;

	lp->free_tx_descriptor_count = num_transmit_desc;
	lp->rx_num_bufs = 0;

#ifdef DYNAMIC_THRESHOLD
	lp->TxFifoThreshold = 600;
#endif
	lp->TxTimeoutCounter = 0;
	lp->RxTimeoutCounter = 0;
	/* configure media */

	if (ZxeConfigureMedia(ppa) == 0) {
		DEBUG1("pdk_hal_start(%d): Media configuration failed\n", ppa);
		pdk_hal_stop(ppa);
		return(PDK_HW_ERR);
	}

	/* load CSR6 */
	pdk_reg_put32(ppa, CSR6, lp->OperationModeBase);

	/* load timer */
	if (lp->PollTimeout)
		pdk_reg_put32(ppa, CSR11, lp->PollTimeout);

	if ((err = load_setup_buf(ppa)) != 0) {
		DEBUG1("pdk_hal_start(%d): load_setup_buf failed\n", ppa);
		return(err);
	}

	lp->OperationMode = lp->OperationModeBase | CSR6_ST;
	pdk_reg_put32(ppa, CSR6, lp->OperationMode);
	pdk_time_delay(100);

	if (do_loopback) {
		if (zxe_loopback_test(ppa) != 0) {
			DEBUG1("pdk_hal_start(%d): Loopback failed\n", ppa);
			pdk_hal_stop(ppa);

			zxe_operational_led(ppa, 0);
			zxe_internal_fault_led(ppa, 1);
			return(PDK_HW_ERR);
		}
	}
	else if (lp->PhyType == DEVICE_BCM5208) {
		/* restore the PHY */
		ZxeWritePHY(ppa, BCM5208_MIICR_REG, (uint16_t)BCM5208_MIICR_RESET);
		pdk_time_delay(2);
		ZxeWritePHY(ppa, BCM5208_MIICR_REG, (uint16_t)lp->PhyControl);
		if (lp->AdapterId == ADAPTER_ZX222)
			ZxeWritePHY(ppa, BCM5208_TXACR_REG,
				(uint16_t)(BCM5208_TXACR_FEF | BCM5208_TXACR_CIMDIS));
	}
	else if (lp->PhyType == DEVICE_ICS1890) {
		ZxeWritePHY(ppa, ICS1890_MIICR_REG, (uint16_t)lp->PhyControl);
	}

	zxe_internal_fault_led(ppa, 0);

	/* start tx and rx */
	ZxeStartTxAndRx(ppa);

	lp->MediaConnectState = MEDIA_DISCONNECTED;
	lp->state = LOCAL_ADAPTER_RUNNING;
	zxe_operational_led(ppa, 1);

	zxe_external_fault_led(ppa, 1, 0);	/* Will be cleared when link up */

	if (lp->PhyType == DEVICE_BCM5208) {
		/* enable PHY interrupts */
		ZxeWritePHY(ppa, BCM5208_IR_REG, (uint16_t)BCM5208_IR_INTEN);
		ZxeReadPHY(ppa, BCM5208_IR_REG, &phy_isr_value);
	}

	/* enable ints */
	SET_CSR7(ppa,CSR7_CFG);

	if (lp->PhyType == DEVICE_BCM5208) {
		ZxeReadPHY(ppa, BCM5208_IR_REG, &phy_isr_value);
	}
	return(0);
}

/*
 * Restart the adapter.  Keep the receive buffers.  Recycle any transmit
 * buffers.
 */
int32_t
pdk_hal_restart(ppa)
int32_t ppa;
{
	local_adapter_t *lp;
	desc_t *rx_dequeue_here;
	desc_t *rx_enqueue_here;
	desc_t *tdd;
	uint32_t rx_num_bufs;
	int32_t temp_debug;
	int32_t err;

	DEBUG2("pdk_hal_restart(%d)\n", ppa);

	if ((lp = local_adapters[ppa]) == 0)
		return(PDK_OUT_STATE);

	if (lp->state != LOCAL_ADAPTER_RUNNING) {
		DEBUG1("pdk_hal_restart(%d): Out of state(%d)\n", ppa, lp->state);
		return(PDK_OUT_STATE);
	}

	/* Stop the adapter */
	SET_CSR7(ppa,0);
	pdk_reg_put32(ppa, CSR0, CSR0_SWR);
	pdk_time_delay(2);

	/* Save the receive buffers */
	pdk_lock_acquire(lp->rx_descriptor_lock);
	pdk_lock_acquire(lp->tx_descriptor_lock);

	rx_dequeue_here = lp->rx_dequeue_here;
	rx_enqueue_here = lp->rx_enqueue_here;
	rx_num_bufs = lp->rx_num_bufs;
	lp->rx_dequeue_here = lp->rx_enqueue_here;

	DEBUG3("pdk_hal_restart(%d): 0x%x 0x%x 0x%x\n", ppa, QNX_OFF(rx_dequeue_here), QNX_OFF(rx_enqueue_here), rx_num_bufs);

	pdk_hal_stop(ppa);

	tdd = rx_dequeue_here;
	while (tdd != rx_enqueue_here) {
		tdd->RDES0 = 0;
		tdd = tdd->NEXT;
	}

	if ((temp_debug = pdk_hal_debug) != 0)
		pdk_hal_debug = 1;

	if ((err = pdk_hal_start(ppa, 0)) != 0) {
		DEBUG1("pdk_hal_restart(%d): Cannot start\n", ppa);
		pdk_hal_debug = temp_debug;
		pdk_lock_release(lp->tx_descriptor_lock);
		pdk_lock_release(lp->rx_descriptor_lock);
		return(err);
	}
	pdk_hal_debug = temp_debug;

	/*
	 * Restore Receives after initialization
	 */
	while (rx_dequeue_here != lp->rx_queue) {
		rx_enqueue_here->BP = rx_dequeue_here->BP;
		rx_enqueue_here->RDES3 = rx_dequeue_here->RDES3;
		rx_enqueue_here->RDES2 = rx_dequeue_here->RDES2;
		rx_enqueue_here->RDES1 = rx_dequeue_here->RDES1 & ~RDES1_RER;
		if (rx_enqueue_here->NEXT == lp->rx_queue)
			rx_enqueue_here->RDES1 |= RDES1_RER;
		rx_enqueue_here = rx_enqueue_here->NEXT;
		rx_dequeue_here = rx_dequeue_here->NEXT;
	}
	lp->rx_enqueue_here = rx_enqueue_here;
	lp->rx_num_bufs = rx_num_bufs;

	tdd = rx_dequeue_here;
	while (tdd != rx_enqueue_here) {
		tdd->RDES0 = RDES0_OWN;
		tdd = tdd->NEXT;
	}
#ifdef DEBUG_CHECK_RING
	check_ring(lp);
#endif
	DEBUG3("pdk_hal_restart(%d):  CSR5 0x%x\n", ppa, pdk_reg_get32(ppa, CSR5));

	pdk_lock_release(lp->tx_descriptor_lock);
	pdk_lock_release(lp->rx_descriptor_lock);

#ifdef PDK_HAL_SPEEDO
	if (lp->speedometer_poll_rate != 0) {
		if (lp->hal_speedometer_timeout_id == 0)
			lp->hal_speedometer_timeout_id = pdk_timeout(hal_speedometer_timeout, (void *)ppa, HAL_SPEEDOMETER_TICKS);
	}
#endif
	
	return(0);
}


uint32_t
pdk_hal_stop(ppa)
int32_t ppa;
{
	local_adapter_t *lp;
	desc_t *dd;

	DEBUG2("pdk_hal_stop(%d)\n", ppa);

	if ((lp = local_adapters[ppa]) == 0)
		return(PDK_OUT_STATE);

	if (lp->state == LOCAL_ADAPTER_RELEASED) {
		DEBUG2("znb%d is RELEASED, already stopped\n",ppa);
		return(0);				/* Don't stop this one again */
	}

	if ((lp->state != LOCAL_ADAPTER_RUNNING) &&
	    (lp->state != LOCAL_ADAPTER_STOPPED)) {

		DEBUG2("pdk_hal_stop(%d): Out of state(%d)\n", ppa, lp->state);
		return(PDK_OUT_STATE);
	}

	zxe_operational_led(ppa, 0);

	pdk_reg_put32(ppa, CSR0, CSR0_SWR);
	lp->state = LOCAL_ADAPTER_STOPPED;

	switch (lp->PhyType) {
		case  DEVICE_BCM5208:
			if (lp->AdapterId == ADAPTER_ZX222)
				ZxeWritePHY(ppa, BCM5208_TXACR_REG, (uint16_t)(BCM5208_TXACR_TXDIS | BCM5208_TXACR_CIMDIS));
			else
				ZxeWritePHY(ppa, BCM5208_AMPR_REG, (uint16_t)BCM5208_AMPR_SUPI);
			break;

		case DEVICE_ICS1890:
			ZxeWritePHY(ppa, ICS1890_MIICR_REG, (uint16_t)ICS1890_MIICR_PWRDN);
			break;
	} 

	pdk_time_delay(2);

	if (lp->MediaConnectState == MEDIA_CONNECTED) {
		(*lp->layer_routines.pdk_status)(ppa, PDK_STATUS_LINK_DOWN, 0);
	}

	lp->MediaConnectState = MEDIA_DISCONNECTED;
	lp->CurrentMediaType = lp->ConfiguredMediaType;

	zxe_external_fault_led(ppa, 0, 0);
	zxe_external_fault_led(ppa, 0, 1);

	/* Free Receive resources */
	for (dd=lp->rx_dequeue_here; dd != lp->rx_enqueue_here; dd=dd->NEXT) {
		if (dd->BP) {
			pdk_buf_unmap(dd->BP);
			pdk_buf_free(dd->BP);
		}
		dd->RDES0 = 0;
	}
	lp->rx_dequeue_here = lp->rx_enqueue_here;
	lp->rx_num_bufs = 0;

	/* Free Transmit resources */
	for (dd=lp->tx_dequeue_here; dd != lp->tx_enqueue_here; dd=dd->NEXT) {
		if (dd->BP) {
			pdk_buf_unmap(dd->BP);
			if (dd->TDES1 & TDES1_SET)
				pdk_buf_free(dd->BP);
			else
				(*lp->layer_routines.pdk_xmit_complete)(lp->ppa, dd->BP, PDK_XMIT_STATUS_FLUSHING);
			dd->BP = 0;
		}
		dd->TDES0 = 0;
	}
	lp->tx_dequeue_here = lp->tx_enqueue_here;
	lp->free_tx_descriptor_count = num_transmit_desc;

	return(0);
}

void
descriptor_flush(lp, flag)
local_adapter_t	*lp;
uint32_t flag;
{
	uint32_t *enqaddr, *deqaddr;
	pdk_desc_handle_t *handlep;

	if (flag == H_READ) {
		enqaddr = (uint32_t *)lp->rx_enqueue_here;
		deqaddr = (uint32_t *)lp->rx_dequeue_here;
		handlep = lp->rx_descriptor_handlep;

		if (enqaddr == deqaddr)
			return;

		SYS_DESC_SYNC(handlep, deqaddr, enqaddr, PDK_DESC_SYNC_FORDEV);

        	/* Issue receive poll command */

		pdk_reg_put32(lp->ppa, CSR2, (uint32_t)enqaddr);
	}
	else {
		enqaddr = (uint32_t *)lp->tx_enqueue_here;
		deqaddr = (uint32_t *)lp->tx_dequeue_here;
		handlep = lp->tx_descriptor_handlep;

		if (enqaddr == deqaddr)
			return;

		SYS_DESC_SYNC(handlep, deqaddr, enqaddr, PDK_DESC_SYNC_FORDEV);

        	/* Issue transmit poll command */

		pdk_reg_put32(lp->ppa, CSR1, (uint32_t)enqaddr);
	}
}

void
descriptor_sync(lp, flag)
local_adapter_t	*lp;
uint32_t flag;
{
	uint32_t *enqaddr, *deqaddr;
	pdk_desc_handle_t *handlep;

	if (flag == H_READ) {
		enqaddr = (uint32_t *)lp->rx_enqueue_here;
		deqaddr = (uint32_t *)lp->rx_dequeue_here;
		handlep = lp->rx_descriptor_handlep;
	}
	else {
		enqaddr = (uint32_t *)lp->tx_enqueue_here;
		deqaddr = (uint32_t *)lp->tx_dequeue_here;
		handlep = lp->tx_descriptor_handlep;
	}

	if (enqaddr == deqaddr)
		return;

	SYS_DESC_SYNC(handlep, deqaddr, enqaddr, PDK_DESC_SYNC_FORCPU);
}

#ifdef LEVELED_INTERRUPTS
/* This has not been tested very thoroughly.  Bob Miller */

/*
 * Leveled interrupts allow receives for an entire multiport card to
 * be tested prior to other conditions such as timers or link changes
 * Two levels currently supported, 0 and 1.
 */

int32_t level_calls = 0;
int32_t level_zero = 0;
int32_t level_zero_miss = 0;
int32_t level_one = 0;
int32_t level_spurious = 0;
int32_t level_bits[NUM_PORTS][32] = {0};

pdk_hal_leveled_iac(ppa, level)
int32_t ppa;
int32_t level;
{
	local_adapter_t *lp;
	uint32_t isr_value;
	uint32_t csr15;
	uint16_t phy_isr_value;
	int32_t i;

	level_calls++;

	lp = local_adapters[ppa];
	/* Avoid unneeded PCI accesses */
	if ((lp == 0) || (lp->state != LOCAL_ADAPTER_RUNNING)) {
		level_spurious++;
		return(0);
	}

	/* Look for receives to avoid CSR5 access */
	switch(level) {
		case 0:
			DESCRIPTOR_SYNC(lp, H_READ);
			if ((lp->rx_dequeue_here != lp->rx_enqueue_here) &&
			   ((lp->rx_dequeue_here->RDES0 & RDES0_OWN) == 0)) {
				SET_CSR7(ppa,0);
				pdk_reg_put32(ppa, CSR5, CSR5_RI);
				level_zero++;
				return(CSR5_RI);
			}

			level_zero_miss++;

			return(0);

		case 1:

			isr_value = pdk_reg_get32(ppa, CSR5); 
			if (isr_value & NIC_ISR_BITS)
				DEBUG4("pdk_hal_leveled_iac(%d): raw_isr_value 0x%x\n", ppa, isr_value);

			isr_value &= NIC_ISR_BITS;
			
			for (i=0;i<32;i++) {
				if ((isr_value>>i) & 1)
					level_bits[ppa][i]++;
			}

			if (isr_value != 0) {
				SET_CSR7(ppa,0);
				if (isr_value & CSR5_GPI) {
					csr15 = pdk_reg_get32(ppa, CSR15);
					/* Clear interupt conditions */
					ZxeReadPHY(ppa, BCM5208_IR_REG, &phy_isr_value);
					DEBUG2("pdk_hal_leveled_iac(%d): Phy interrupt 0x%x detected CSR15=0x%x\n", ppa, phy_isr_value, csr15);
					if ((csr15 & 0x10000) == 0) {
						DEBUG2("pdk_hal_leveled_iac(%d): Ignore phy interrupt transition to zero\n", ppa);
						pdk_reg_put32(ppa, CSR5, CSR5_GPI);
						isr_value &= ~CSR5_GPI;
						if (isr_value == 0) {
							SET_CSR7(ppa,CSR7_CFG);
							level_spurious++;
							return(0);
						}
					}
				}
				pdk_reg_put32(ppa, CSR5, isr_value);
			}

			if (isr_value == 0)
				level_spurious++;
			else
				level_one++;

			return(isr_value);

		default:
			DEBUG2("pdk_hal_leveled_iac(%d): Called with unsupported level %d\n", ppa, level);
			break;
	}
		
	return(0);
}

uint32_t
pdk_hal_iac(ppa)
int32_t ppa;
{
	uint32_t isr_value;

	if ((isr_value = pdk_hal_leveled_iac(ppa, 0)) == 0)
		isr_value = pdk_hal_leveled_iac(ppa, 1);

	return(isr_value);
}
#endif /* LEVELED_INTERRUPTS */

uint32_t
pdk_hal_iac(ppa)
int32_t ppa;
{
	local_adapter_t *lp;
	uint32_t isr_value;

	lp = local_adapters[ppa];
	/* Avoid unneeded PCI accesses */
#ifdef SHADOW_CSR7
	if ((lp == 0) || (lp->state != LOCAL_ADAPTER_RUNNING) || !lp->csr7)
#else
	if ((lp == 0) || (lp->state != LOCAL_ADAPTER_RUNNING))
#endif
	{
		return(0);
	}

#ifdef HAL_AVOID_ISR_READ
	{
		uint32_t soft_ret=0;

		/* Look for receives and xmits to avoid CSR5 access */
		DESCRIPTOR_SYNC(lp, H_READ);
		if ((lp->rx_dequeue_here != lp->rx_enqueue_here) &&
		   ((lp->rx_dequeue_here->RDES0 & RDES0_OWN) == 0)) {
			lp->rx_icount++;
			soft_ret |= CSR5_RI;
		}
		DESCRIPTOR_SYNC(lp, H_WRITE);
		if (( lp->tx_dequeue_here != lp->tx_enqueue_here) &&
		   ((lp->tx_dequeue_here->TDES0 & TDES0_OWN) == 0)) {
			if (soft_ret == 0) 
				lp->tx_icount++;

			soft_ret |= CSR5_TI;
		}
		/* do regs for both TI and RI here */
		if (soft_ret) {
			SET_CSR7(ppa, 0);
			pdk_reg_put32(ppa, CSR5, soft_ret);
			return(soft_ret);
		}
	}
#endif

	isr_value = pdk_reg_get32(ppa, CSR5); 
	if (isr_value & NIC_ISR_BITS)
		DEBUG4("pdk_hal_iac(%d): raw_isr_value 0x%x CSR12 0x%x\n", ppa, isr_value, pdk_reg_get32(ppa, CSR12));

	isr_value &= NIC_ISR_BITS;

	if (lp->Info.NicType != DEVICE_DEC21143)
		isr_value &= ~(CSR5_LNF|CSR5_LC|CSR5_LNP_ANC|CSR5_GPI);

	if (isr_value != 0) {
		SET_CSR7(ppa,0);
		pdk_reg_put32(ppa, CSR5, isr_value);

	}
	if (isr_value & CSR5_RI)
		lp->rx_icount++;

	if (isr_value & CSR5_TI)
		lp->tx_icount++;

	if (isr_value & ~(CSR5_RI|CSR5_TI))
		lp->other_icount++;

	return(isr_value);
}


uint32_t
pdk_hal_intr(ppa, isr_value)
int32_t ppa;
uint32_t isr_value;
{
	local_adapter_t *lp;
	uint16_t phy_isr_value;
	uint32_t csr15;

	DEBUG3("pdk_hal_intr(%d): isr_value 0x%x\n", ppa, isr_value);
	if ((lp = local_adapters[ppa]) != 0) {

		if ((lp->rx_dequeue_here != lp->rx_enqueue_here) &&
			(lp->rx_dequeue_here->RDES0 & RDES0_OWN) == 0) {
			process_recv_complete(ppa);
			pdk_reg_put32(ppa, CSR5, CSR5_RI);
		}

		DESCRIPTOR_SYNC(lp, H_READ);
		if ((lp->rx_dequeue_here != lp->rx_enqueue_here) &&
			(lp->rx_dequeue_here->RDES0 & RDES0_OWN) == 0) {
			process_recv_complete(ppa);
		}

		DESCRIPTOR_SYNC(lp, H_WRITE);
		if ((lp->tx_dequeue_here != lp->tx_enqueue_here) &&
			((lp->tx_dequeue_here->TDES0 & TDES0_OWN) == 0)) {

			pdk_lock_acquire(lp->tx_descriptor_lock);
			process_xmit_complete(lp);
			pdk_lock_release(lp->tx_descriptor_lock);
		}

		if (isr_value & CSR5_GPI) {
			csr15 = pdk_reg_get32(ppa, CSR15);
			if ((csr15 & 0x10000) == 0) {
				DEBUG4("pdk_hal_intr(%d): Ignore phy interrupt transition to zero\n", ppa);
				isr_value &= ~CSR5_GPI;
			} else {
				/* Clear interupt conditions */
				ZxeReadPHY(ppa, BCM5208_IR_REG, &phy_isr_value);
				DEBUG3("pdk_hal_intr(%d): Phy interrupt 0x%x detected CSR15=0x%x\n", ppa, phy_isr_value, csr15);
			}
		}

		if (isr_value & (CSR5_LNF|CSR5_LC|CSR5_LNP_ANC|CSR5_GTE|CSR5_GPI)) {
			ZxeMediaPoll(ppa,isr_value);
		}
		SET_CSR7(ppa,CSR7_CFG);
	} else
		pdk_reg_put32(ppa, CSR7, CSR7_CFG);

	return(0);
}


uint32_t
pdk_hal_xmit(ppa, bp)
int32_t ppa;
pdk_buf_t *bp;
{
	local_adapter_t	*lp;
	pdk_map_t	*mp;
	desc_t		*first_dd;
	desc_t		*dd;
	desc_t		*last_dd;
	uint32_t	length;
	uint32_t	paddr;
	int32_t		free_txds;
	int32_t		err;
	uint32_t	num_txds;
	uint32_t	pending;


	if ((lp = local_adapters[ppa]) == 0) {
		DEBUG2("pdk_hal_xmit(%d): local_adapter is zero\n", ppa);
		return(PDK_LINK_DOWN);
	}

	pdk_lock_acquire(lp->tx_descriptor_lock);

	if (lp->MediaConnectState != MEDIA_CONNECTED) {
		DEBUG2("pdk_hal_xmit(%d): local_adapter not CONNECTED\n", ppa);
		pdk_lock_release(lp->tx_descriptor_lock);
		return(PDK_LINK_DOWN);
	}

#ifdef NOT
	if ((length = pdk_buf_get_len(bp)) > PDK_ETHERMAX) {
		DEBUG2("pdk_hal_xmit(%d): buffer too long\n", ppa);
		pdk_lock_release(lp->tx_descriptor_lock);
		return(PDK_BAD_PARAM);
	}
#endif

	free_txds = lp->free_tx_descriptor_count;

	/* Always leave one */
	if (--free_txds == 0) {
		DEBUG2("pdk_hal_xmit(%d): Out of free descriptor entries\n", ppa);
		DESCRIPTOR_SYNC(lp, H_WRITE);
		/*  THIS IS A BUG:  pdk_lock_release(lp->tx_descriptor_lock); */
		process_xmit_complete(lp);
		free_txds = lp->free_tx_descriptor_count;
		if (--free_txds == 0) {
			pdk_lock_release(lp->tx_descriptor_lock);
			return(PDK_NO_ROOM);
		}
	}

	/* Map this pdk transmit buffer */
	if ((mp = pdk_buf_map(bp, PDK_BUF_XMIT, &err)) == 0) {
		DEBUG2("pdk_hal_xmit(%d): pdk_buf_map failed - err %d\n", ppa, err);
		pdk_lock_release(lp->tx_descriptor_lock);
		return(err);
	}

	/* Do the first.. */
	mp = pdk_buf_get_pmap(mp, &paddr, &length, &err);
		
	if (err) {
		DEBUG2("pdk_hal_xmit(%d): pdk_buf_get_pmap failed\n", ppa);
		pdk_buf_unmap(bp);
		pdk_lock_release(lp->tx_descriptor_lock);
		return(err);
	}

	first_dd = lp->tx_enqueue_here;
	first_dd->TDES2 = pdk_htobl(paddr);
	first_dd->TDES1 = pdk_htobl(length);
	first_dd->TDES1 |= TDES1_FS;
	if (first_dd->NEXT == lp->tx_queue)
		first_dd->TDES1 |= TDES1_TER;

	first_dd->BP = 0;
	DEBUG3("pdk_hal_xmit(%d): dd 0x%x(%d):", ppa, QNX_OFF(first_dd), length);

	last_dd = first_dd;
#ifdef PDK_HAL_SPEEDO
	lp->tx_bytes+=length;
#endif
	
	/* Now do the rest.. */
	for (dd = first_dd;;) {
		if (mp == 0)
			break;

		mp = pdk_buf_get_pmap(mp, &paddr, &length, &err);
		if (err) {
			DEBUG1("\npdk_hal_xmit(%d): pdk_buf_get_pmap failed\n", ppa);
			break;
		}

		DEBUG3(" 0x%x(%d)", QNX_OFF(dd), length);
		dd->TDES3 = pdk_htobl(paddr);
#ifdef PDK_HAL_SPEEDO
		lp->tx_bytes+=length;
#endif
		length <<= 11;
		dd->TDES1 |= pdk_htobl(length);

		if (mp == 0)
			break;

		dd = dd->NEXT;
		last_dd = dd;

		/* Always leave one */
		if (--free_txds == 0) {
			DEBUG3("\npdk_hal_xmit(%d): Out of free descriptor entries in middle of message\n", ppa);
			err = PDK_NO_ROOM;
			break;
		}

		mp = pdk_buf_get_pmap(mp, &paddr, &length, &err);
		if (err) {
			DEBUG1("\npdk_hal_xmit(%d): pdk_buf_get_pmap failed\n", ppa);
			break;
		}

#ifdef PDK_HAL_SPEEDO
		lp->tx_bytes+=length;
#endif
		dd->TDES2 = pdk_htobl(paddr);
		dd->TDES1 = pdk_htobl(length);
		if (dd->NEXT == lp->tx_queue)
			dd->TDES1 |= TDES1_TER;

		DEBUG3(" 0x%x(%d)", QNX_OFF(dd), length);
		dd->TDES0 = TDES0_OWN;
		dd->BP = 0;
	}

	/* No error?? */
	if (err == 0) {
		/* Update Tx descriptor list... */
		lp->tx_enqueue_here = last_dd->NEXT;
		lp->tx_enqueue_here->TDES0 = 0;

		lp->tx_descriptor_count += lp->free_tx_descriptor_count-free_txds;
	
		num_txds = lp->free_tx_descriptor_count-free_txds;
		if ((lp->free_tx_descriptor_count = free_txds) < lp->tx_lwm) {
			lp->tx_lwm = free_txds;
			if (lp->tx_lwm == 1)
				lp->tx_depleted++;
		}

		/* Interrupt when tx_pending_count xmit ds have been posted
		 * Pending is the number of currently consumed
		 * transmit resources
		 */
		pending = num_transmit_desc - free_txds;
		if (pending >= lp->tx_pending_threshold) { 
			if ( lp->tx_pending_count <= 0) {
				last_dd->TDES1 |= TDES1_IC;
				lp->tx_pending_count = lp->tx_pending_threshold; 
			}
		}
		lp->tx_pending_count -= num_txds;

		/* and last descriptor... */
		last_dd->TDES1 |= TDES1_LS;

		/* and save buffer */
		last_dd->BP = bp;

		/* guarantee in order... */
		SYS_INSTR_SYNC();

		/* and first descriptor... */
		first_dd->TDES0 = TDES0_OWN;

		DEBUG4(" -> deq 0x%x enq 0x%x\n", lp->tx_dequeue_here, lp->tx_enqueue_here);

		descriptor_flush(lp, H_WRITE);
	}
	else
		pdk_buf_unmap(bp);

	lp->pdk_xmit_complete_check++;
	if ((lp->pdk_xmit_complete_check%TX_COMPLETE_CHECK_THRESHOLD) == 0) {
		DESCRIPTOR_SYNC(lp, H_WRITE);
		process_xmit_complete(lp);
	}

	pdk_lock_release(lp->tx_descriptor_lock);

	return(err);
}


void
process_xmit_complete(lp)
local_adapter_t *lp;
{
	desc_t		*dd;
	uint32_t	tdes0;
	int32_t		free_txds;
	pdk_buf_t	*bp;
	int32_t		status;

	for (free_txds=lp->free_tx_descriptor_count, dd=lp->tx_dequeue_here;
					dd!=lp->tx_enqueue_here;
					dd=dd->NEXT, free_txds++) {
		/* Not done yet?? */
		if ((tdes0 = dd->TDES0) & TDES0_OWN) {
			break;
		}

		/* assume good status */
		status = 0;

		/* Setup buffer */
		if (dd->TDES1 & TDES1_SET) {
			if (dd->BP) {
				pdk_buf_unmap(dd->BP);
				pdk_buf_free(dd->BP);
				dd->BP = 0;
			}

			goto xmit_complete_cont;
		}

		/* Not last descriptor?? */
		if ((dd->TDES1 & TDES1_LS) == 0) {
			continue;
		}

		if (tdes0 == 0) {
			goto xmit_complete_cont;
		}

		/* Check for TX FIFO underflow */
		lp->TxTimeoutCounter = 0;

		++lp->Statistics.hs_tx_packets;
/*hamid added the following for the TX byte count*/
		lp->Statistics.hs_tx_bytes=lp->tx_bytes;

		/* Transmit failure?? */
		if (tdes0 & TDES0_ES) {
#ifdef PDK_HAL_SPEEDO
			++lp->Statistics.hs_tx_err_packets;
#endif
			status = PDK_XMIT_STATUS_ERR;

			if (tdes0 & TDES0_EC)
				++lp->Statistics.hs_tx_max_collisions;
			if (tdes0 & TDES0_LO)
				++lp->Statistics.hs_tx_carrier_loss;
			if (tdes0 & TDES0_LC)
				++lp->Statistics.hs_tx_late_collisions;
			if (tdes0 & TDES0_NC)
				++lp->Statistics.hs_tx_no_carrier;
			if (tdes0 & TDES0_TO)
				++lp->Statistics.hs_tx_jabber;
			if (tdes0 & TDES0_UF) {
				++lp->Statistics.hs_tx_underrun;
				lp->TxFifoUnderflowGauge += 1000;
			}

			DEBUG2("xmit_complete(%d): ERROR tdes0(0x%x)\n", lp->ppa, pdk_htobl(tdes0));
		}
		else {
			if (tdes0 & TDES0_CC_MASK) {
				DEBUG2("xmit_complete(%d): Collision tdes0(0x%x)\n", lp->ppa, pdk_htobl(tdes0));
				if ( ((tdes0 & TDES0_CC_MASK)>>TDES0_CC_SHIFT)==1 )
					++lp->Statistics.hs_tx_one_collision;
				else
					++lp->Statistics.hs_tx_more_collisions;
			}
			if (tdes0 & TDES0_DE) {
				DEBUG2("xmit_complete(%d): Deferred tdes0(0x%x)\n", lp->ppa, pdk_htobl(tdes0));
				++lp->Statistics.hs_tx_deferred;
			}

			if (lp->TxFifoUnderflowGauge)
				lp->TxFifoUnderflowGauge--;
		}

xmit_complete_cont:
		lp->tx_dequeue_here = dd->NEXT;
		lp->free_tx_descriptor_count = free_txds+1;
		if (dd->BP) {
			DEBUG3("xmit_complete(%d): enq 0x%x deq 0x%x\n", lp->ppa, QNX_OFF(lp->tx_enqueue_here), QNX_OFF(lp->tx_dequeue_here));
			bp = dd->BP;
			dd->BP = 0;
			pdk_buf_unmap(bp);
			(*lp->layer_routines.pdk_xmit_complete)(lp->ppa, bp, status);
		}
	}
}

uint32_t
zxe_set_recv_buf(int32_t ppa, pdk_buf_t *bp, local_adapter_t *lp)
{
	desc_t		*dd;
	pdk_map_t	*mp;
	uint32_t	paddr;
	uint32_t	length;
	int32_t		err;

	/* First check that adapter is still running */
	if (lp->state != LOCAL_ADAPTER_RUNNING) {
		DEBUG2("pdk_hal_recv(%d): adapter not running\n", ppa);
		return(PDK_OUT_STATE);
	}

	/* Map this pdk receive buffer */
	if ((mp = pdk_buf_map(bp, PDK_BUF_RECV, &err)) == 0) {
		DEBUG1("pdk_hal_recv(%d): pdk_buf_map failed\n", ppa);
		return(err);
	}

	mp = pdk_buf_get_pmap(mp, &paddr, &length, &err);
	if (err) {
		pdk_buf_unmap(bp);
		DEBUG1("pdk_hal_recv(%d): pdk_buf_get_pmap failed\n", ppa);
		return(err);
	}

	if (mp != 0) {
		pdk_buf_unmap(bp);
		DEBUG1("pdk_hal_recv(%d): Multi fragmented receive buffer\n", ppa);
		return(PDK_BAD_PARAM);
	}

	dd = lp->rx_enqueue_here;

	/* and save buffer */
	dd->BP = bp;

	dd->RDES2 = pdk_htobl(paddr);
	dd->RDES1 = pdk_htobl(length);
	if (dd->NEXT == lp->rx_queue)
		dd->RDES1 |= RDES1_RER;

	/* guarantee in order... */
	SYS_INSTR_SYNC();

	dd->RDES0 = RDES0_OWN;

	lp->rx_enqueue_here = dd->NEXT;
	lp->rx_num_bufs++;
	if (lp->rx_num_bufs > lp->rx_hwm)
		lp->rx_hwm = lp->rx_num_bufs;

	DEBUG4("pdk_hal_recv(%d): deq 0x%x enq 0x%x\n", ppa, QNX_OFF(lp->rx_dequeue_here), QNX_OFF(lp->rx_enqueue_here));

		descriptor_flush(lp, H_READ);

	return(0);
}

uint32_t
pdk_hal_recv(ppa, bp)
int32_t ppa;
pdk_buf_t *bp;
{
	local_adapter_t	*lp;
	int32_t		err;

	if ((lp = local_adapters[ppa]) == 0) {
		return(PDK_OUT_STATE);
	}

	pdk_lock_acquire(lp->rx_descriptor_lock);

	err = zxe_set_recv_buf(ppa, bp, lp);

	pdk_lock_release(lp->rx_descriptor_lock);

	return(err);
}

void
process_recv_complete(ppa)
int32_t ppa;
{
	local_adapter_t	*lp;
	desc_t		*dd;
	pdk_buf_t	*bp;
	uint32_t	frame_size;
	uint32_t	alignment_error;
	uint32_t	status;
	uint32_t	rdes0;

	lp = local_adapters[ppa];

	while (1) {
		pdk_lock_acquire(lp->rx_descriptor_lock);

		dd = lp->rx_dequeue_here;

		/* Empty ?? */
		if (dd == lp->rx_enqueue_here) {
			pdk_lock_release(lp->rx_descriptor_lock);
			break;
		}
		/* Not done yet?? */
		if ((rdes0 = dd->RDES0) & RDES0_OWN) {
			pdk_lock_release(lp->rx_descriptor_lock);
			break;
		}
		lp->RxTimeoutCounter = 0;

		lp->rx_dequeue_here = dd->NEXT;
		if (--lp->rx_num_bufs < lp->rx_lwm)
			lp->rx_lwm = lp->rx_num_bufs;

		lp->rx_descriptor_count++;

		bp = dd->BP;
		dd->BP = 0;

		pdk_lock_release(lp->rx_descriptor_lock);

		/* Must be last since receive buffers are max size */

		/* Unmap the buffer */
		pdk_buf_unmap(bp);

		/* Receive failure?? */
		if (rdes0 & RDES0_ES) {

			alignment_error = 0;
			++lp->Statistics.hs_rx_err_packets;

			if (rdes0 & RDES0_CE) {

				/* Only with CRC errors.. */
				if (rdes0 & RDES0_RE) {
					++lp->Statistics.hs_rx_mii_error;
				}
				else {
					++lp->Statistics.hs_rx_crc;
					if (rdes0 & RDES0_DB) {
						alignment_error++;
					}
				}
			}

			if (rdes0 & RDES0_LE) {
				++lp->Statistics.hs_rx_descriptor_error;
			}

			if (rdes0 & RDES0_TL) {
				++lp->Statistics.hs_rx_frame_too_long;
			}

			if (rdes0 & RDES0_RF) {
				++lp->Statistics.hs_rx_runt_frame;
				alignment_error = 0;
			}


			if (rdes0 & RDES0_CS) {
				++lp->Statistics.hs_rx_collisions;
				alignment_error = 0;
			}

			if (alignment_error) {
				++lp->Statistics.hs_rx_alignment_error;
			}

			if (rdes0 & RDES0_OF) {
				++lp->Statistics.hs_rx_overrun;
			}

			DEBUG2("recv_complete(%d): ERROR rdes0(0x%x) enq 0x%x deq 0x%x CSR5 0x%x CSR12 0x%x\n", ppa, pdk_htobl(rdes0), QNX_OFF(lp->rx_enqueue_here), QNX_OFF(lp->rx_dequeue_here), pdk_reg_get32(ppa, CSR5), pdk_reg_get32(ppa, CSR12));

			/* Give buffer back to adapter */
			pdk_hal_recv(ppa, bp);
		}
		else {

			frame_size = dd->RDES0 & RDES0_FL_MASK;
			frame_size = (pdk_htobl(frame_size)>>16) - 4;
			pdk_buf_set_len(bp, frame_size);
#ifdef PDK_HAL_SPEEDO
			lp->rx_bytes+=frame_size;
#endif

			status = 0;

			DEBUG3("recv_complete(%d): deq 0x%x len %d\n", ppa, QNX_OFF(lp->rx_dequeue_here), frame_size);

			PDK_BUF_SYNC(bp, frame_size, PDK_BUF_READ);

			/* Give buffer to upper level */
			(*lp->layer_routines.pdk_recv_complete)(ppa, bp, status);

			/* Update the statistics based on Receive status */
			++lp->Statistics.hs_rx_packets;

		}
	}
}

uint32_t
hal_get_state(ppa, hsp)
int32_t ppa;
hal_state_t *hsp;
{
	local_adapter_t	*lp;
	pdk_hal_stat_t stat;

	pdk_bzero(hsp, sizeof(hal_state_t));

	hsp->hs_version = HS_VERSION;

	if ((ppa < 0) || (ppa >= NUM_PORTS)) {
		hsp->hs_state = HS_STATE_NOT_INIT;
		return(0);
	}

	lp = local_adapters[ppa];
	if (lp == 0) {
		hsp->hs_state = HS_STATE_NOT_INIT;
		return(0);
	}

	hal_get_stats(ppa, &stat);
	
	switch(lp->state) {
		case LOCAL_ADAPTER_RUNNING:
			hsp->hs_state = HS_STATE_RUNNING;
			break;
		case LOCAL_ADAPTER_RELEASED:
			hsp->hs_state = HS_STATE_RELEASED;
			break;
		default:
			hsp->hs_state = HS_STATE_STOPPED;
	}

	switch(lp->ConfiguredMediaType) {
		case ZXE_MEDIA_AUTO:
			hsp->hs_link_cfg = HS_LINK_AUTO;
			break;

		case ZXE_MEDIA_TP:
			hsp->hs_link_cfg = HS_LINK_TP;
			break;

		case ZXE_MEDIA_BNC:
			hsp->hs_link_cfg = HS_LINK_BNC;
			break;

		case ZXE_MEDIA_AUI:
			hsp->hs_link_cfg = HS_LINK_AUI;
			break;

		case ZXE_MEDIA_TPFD:
			hsp->hs_link_cfg = HS_LINK_TPFD;
			break;

		case ZXE_MEDIA_TX:
			hsp->hs_link_cfg = HS_LINK_TX;
			break;

		case ZXE_MEDIA_TXFD:
			hsp->hs_link_cfg = HS_LINK_TXFD;
			break;

		case ZXE_MEDIA_T4:
			hsp->hs_link_cfg = HS_LINK_T4;
			break;

		case ZXE_MEDIA_AUTO_COMPLETE:
			hsp->hs_link_cfg = HS_LINK_AUTO;
			break;
	}


	switch(lp->CurrentMediaType) {
		case ZXE_MEDIA_AUTO:
			hsp->hs_link_state = HS_LINK_DOWN;
			break;

		case ZXE_MEDIA_TP:
			hsp->hs_link_state = HS_LINK_TP;
			break;

		case ZXE_MEDIA_BNC:
			hsp->hs_link_state = HS_LINK_BNC;
			break;

		case ZXE_MEDIA_AUI:
			hsp->hs_link_state = HS_LINK_AUI;
			break;

		case ZXE_MEDIA_TPFD:
			hsp->hs_link_state = HS_LINK_TPFD;
			break;

		case ZXE_MEDIA_TX:
			hsp->hs_link_state = HS_LINK_TX;
			break;

		case ZXE_MEDIA_TXFD:
			hsp->hs_link_state = HS_LINK_TXFD;
			break;

		case ZXE_MEDIA_T4:
			hsp->hs_link_state = HS_LINK_T4;
			break;

		case ZXE_MEDIA_AUTO_COMPLETE:
			hsp->hs_link_state = HS_LINK_DOWN;
			break;
	}

	if (lp->MediaConnectState == MEDIA_DISCONNECTED) {
		hsp->hs_link_state = HS_LINK_DOWN;
		hsp->hs_fault =
		(lp->led_local_settings & (FAILOVER_FAULT_LED|EXT_FAULT_LED))?HS_FAULT_EXTERNAL:
		((lp->led_local_settings & (INT_FAULT_LED))?HS_FAULT_INTERNAL:0);
	}
	hsp->hs_tx_descriptor_count = lp->tx_descriptor_count;
	hsp->hs_rx_descriptor_count = lp->rx_descriptor_count;
	hsp->hs_tx_icount = lp->tx_icount;
	hsp->hs_rx_icount = lp->rx_icount;
	hsp->hs_other_icount = lp->other_icount;

	hsp->hs_tx_resources_max = num_transmit_desc-1;
	hsp->hs_tx_resources = num_transmit_desc-lp->free_tx_descriptor_count;
	hsp->hs_tx_lwm = lp->tx_lwm-1;
	lp->tx_lwm = num_transmit_desc;
	hsp->hs_tx_depleted = lp->tx_depleted;

	hsp->hs_tx_packets = stat.hs_tx_packets;
	hsp->hs_tx_err_packets = stat.hs_tx_err_packets;
	hsp->hs_tx_one_collision = stat.hs_tx_one_collision;
	hsp->hs_tx_more_collisions = stat.hs_tx_more_collisions;
	hsp->hs_tx_max_collisions = stat.hs_tx_max_collisions;
	hsp->hs_tx_late_collisions = stat.hs_tx_late_collisions;
	hsp->hs_tx_carrier_loss = stat.hs_tx_carrier_loss;
	hsp->hs_tx_no_carrier = stat.hs_tx_no_carrier;
	hsp->hs_tx_jabber = stat.hs_tx_jabber;
	hsp->hs_tx_underrun = stat.hs_tx_underrun;
	hsp->hs_tx_overrun = stat.hs_tx_overrun;
	hsp->hs_tx_deferred = stat.hs_tx_deferred;
	hsp->hs_rx_packets = stat.hs_rx_packets;

	hsp->hs_rx_resources_max = lp->rx_hwm;
	hsp->hs_rx_resources = lp->rx_num_bufs;
	hsp->hs_rx_lwm = lp->rx_lwm;
	lp->rx_lwm = lp->rx_hwm;
	lp->rx_depleted += pdk_reg_get32(ppa, CSR8) & 0x1ffff;
	hsp->hs_rx_depleted = lp->rx_depleted;

	hsp->hs_rx_err_packets = stat.hs_rx_err_packets;
	hsp->hs_rx_collisions = stat.hs_rx_collisions;
	hsp->hs_rx_mii_error = stat.hs_rx_mii_error;
	hsp->hs_rx_crc = stat.hs_rx_crc;
	hsp->hs_rx_descriptor_error = stat.hs_rx_descriptor_error;
	hsp->hs_rx_frame_too_long = stat.hs_rx_frame_too_long;
	hsp->hs_rx_runt_frame = stat.hs_rx_runt_frame;
	hsp->hs_rx_alignment_error = stat.hs_rx_alignment_error;
	hsp->hs_rx_overrun = stat.hs_rx_overrun;

	return(0);
}

uint32_t
hal_get_stats(ppa, stat)
int32_t ppa;
pdk_hal_stat_t *stat;
{
	local_adapter_t	*lp;


	if ((lp = local_adapters[ppa]) == 0) {
		return(PDK_OUT_STATE);
	}

	pdk_bcopy(&lp->Statistics, stat, sizeof(pdk_hal_stat_t));
	return(0);
}


uint32_t
hal_set_stats(ppa, stat)
int32_t ppa;
pdk_hal_stat_t *stat;
{
	local_adapter_t	*lp;


	if ((lp = local_adapters[ppa]) == 0) {
		return(PDK_OUT_STATE);
	}

	pdk_bcopy(stat, &lp->Statistics, sizeof(pdk_hal_stat_t));
	return(0);
}

#ifdef PDK_HAL_SPEEDO

uint32_t		tx_bytes=0;
uint32_t		rx_bytes=0;
uint32_t
hal_get_speed(ppa, hsp)
int32_t ppa;
hal_speedometer_t *hsp;
{
	local_adapter_t	*lp;
	pdk_hal_stat_t stat;

	pdk_bzero(hsp, sizeof(hal_speedometer_t));

	hsp->hsp_version = HSP_VERSION;

	lp = local_adapters[ppa];
	if (lp == 0) {
		DEBUG1("hal_get_speed NOT INIT\n");
		hsp->hsp_state = HS_STATE_NOT_INIT;
		return(0);
	}

	if (lp->state == LOCAL_ADAPTER_RUNNING)
		hsp->hsp_state = HS_STATE_RUNNING;
	else
		hsp->hsp_state = HS_STATE_STOPPED;

	pdk_bcopy(&lp->hsp,hsp,sizeof(hal_speedometer_t));
	hsp->hsp_poll_rate = lp->speedometer_poll_rate;
	hsp->hsp_version = HSP_VERSION;

	return(0);
}

uint32_t
hal_set_speedometer_params(ppa, hsp)
int32_t ppa;
hal_speedometer_t *hsp;
{
	local_adapter_t	*lp;

	if (hsp->hsp_version != HSP_VERSION ) {
		return(PDK_BAD_PARAM);
	}
	if ( hsp->hsp_poll_rate	== 0 ) {
  		pdk_hal_ctl(ppa, PDK_HAL_STOP_SPEEDOMETER_POLL, 0);
		return(0);
	}

	if ( hsp->hsp_poll_rate	== -1 ) {
		/* XXX deal with this */
	}

	lp = local_adapters[ppa];
	if (lp == 0) {
		DEBUG1("hal_get_speed NOT INIT\n");
		hsp->hsp_state = HS_STATE_NOT_INIT;
		return(0);
	}

	if (lp->state == LOCAL_ADAPTER_RUNNING)
		hsp->hsp_state = HS_STATE_RUNNING;
	else
		hsp->hsp_state = HS_STATE_STOPPED;

	switch(hsp->notification_type) {
		case PSDB_GREATER_THAN_OR_EQUAL:
			pdk_bcopy(hsp,&lp->set_params_gte_hsp,sizeof(hal_speedometer_t));
			break;
		case PSDB_LESS_THAN:
			pdk_bcopy(hsp,&lp->set_params_lt_hsp,sizeof(hal_speedometer_t));
			break;
	}

  	pdk_hal_ctl(ppa,PDK_HAL_SET_SPEEDOMETER_POLL_RATE_AND_START,(void *)hsp->hsp_poll_rate);

	return(0);
}

void hal_psdb_speedometer_notification(int32_t ppa,uint32_t notification_type);

void
hal_psdb_speedometer_notification(ppa,notification_type)
int32_t ppa;
uint32_t notification_type;
{
	local_adapter_t	*lp;
	pdk_psdb_hal_speedometer_notification_t pftn;
	hal_speedometer_t *hsp=&pftn.phsn_speed;
	uint32_t err;

	lp = local_adapters[ppa];

	pftn.phsn_header.ppr_header.Size = sizeof(pftn);
	pftn.phsn_header.ppr_header.Type = PSDB_TYPE_HAL_SPEEDOMETER_NOTIFICATION;
	pftn.phsn_header.ppr_ppa = ppa;
	/* update record with current values */
	hsp->notification_type=notification_type;

	err = hal_get_speed(ppa, hsp);
	if (err)
		return;
	else {
		register uint32_t *lp, *record_lim;

		lp = (uint32_t *) &pftn;
		record_lim = (uint32_t *)((char *)lp + pftn.phsn_header.ppr_header.Size);
		while (lp < record_lim)
			*lp++ = psdb_ntop(*lp);
	}

	(*lp->layer_routines.pdk_status)(ppa, PDK_PSDB, &pftn);
}

void
hal_speedometer_timeout(arg)
void *arg;
{
	uint32_t tmp;
	local_adapter_t	*lp;
	int32_t ppa=(int32_t)arg;
	lp = local_adapters[ppa];
	
	if (lp->speedometer_poll_rate == 0) {
		lp->hal_speedometer_timeout_id = 0;
		return;
	}

	/* do rate calculations: current rate = (current-last)/interval */

	lp->hsp.hsp_tx_pps = 
		HAL_SPEEDOMETER_RATE(lp->Statistics.hs_tx_packets-lp->prev_Statistics.hs_tx_packets);
	lp->hsp.hsp_rx_pps = 
		HAL_SPEEDOMETER_RATE(lp->Statistics.hs_rx_packets-lp->prev_Statistics.hs_rx_packets);

	lp->hsp.hsp_tx_ips = 
		HAL_SPEEDOMETER_RATE(lp->tx_icount-lp->prev_tx_icount);

	lp->hsp.hsp_rx_ips = 
		HAL_SPEEDOMETER_RATE(lp->rx_icount-lp->prev_rx_icount);

	lp->hsp.hsp_tx_eps = 
		HAL_SPEEDOMETER_RATE(lp->Statistics.hs_tx_err_packets-lp->prev_Statistics.hs_tx_err_packets);

	lp->hsp.hsp_rx_eps = 
		HAL_SPEEDOMETER_RATE(lp->Statistics.hs_rx_err_packets-lp->prev_Statistics.hs_rx_err_packets);

#define ONE_MBPS 1000000
#define MBPS (ONE_MBPS)

	if (lp->LinkSpeed > 0) {

		tmp = HAL_SPEEDOMETER_RATE(lp->tx_bytes);
		tmp = (tmp*(100/lp->LinkSpeed)+MBPS/16)/(MBPS/8);
		lp->hsp.hsp_tx_pctutil = tmp;

		tmp  = HAL_SPEEDOMETER_RATE(lp->rx_bytes);
		tmp = (tmp*(100/lp->LinkSpeed)+MBPS/16)/(MBPS/8);
		lp->hsp.hsp_rx_pctutil = tmp;
	
	}

	/* update maximums */

	if (lp->hsp.hsp_tx_pps > lp->hsp.hsp_tx_pps_hwm)
		lp->hsp.hsp_tx_pps_hwm = lp->hsp.hsp_tx_pps;
 
	if (lp->hsp.hsp_rx_pps > lp->hsp.hsp_rx_pps_hwm)
		lp->hsp.hsp_rx_pps_hwm = lp->hsp.hsp_rx_pps;

	if (lp->hsp.hsp_tx_ips > lp->hsp.hsp_tx_ips_hwm)
		lp->hsp.hsp_tx_ips_hwm = lp->hsp.hsp_tx_ips;
 
	if (lp->hsp.hsp_rx_ips > lp->hsp.hsp_rx_ips_hwm)
		lp->hsp.hsp_rx_ips_hwm = lp->hsp.hsp_rx_ips;

	if (lp->hsp.hsp_tx_pctutil > 99)
		lp->hsp.hsp_tx_pctutil = 99;
	
	if (lp->hsp.hsp_rx_pctutil > 99)
		lp->hsp.hsp_rx_pctutil = 99;

	if (lp->hsp.hsp_tx_pctutil > lp->hsp.hsp_tx_pctutil_hwm)
		lp->hsp.hsp_tx_pctutil_hwm = lp->hsp.hsp_tx_pctutil;

	if (lp->hsp.hsp_rx_pctutil > lp->hsp.hsp_rx_pctutil_hwm)
		lp->hsp.hsp_rx_pctutil_hwm = lp->hsp.hsp_rx_pctutil;

#define PSDB_ALARM_NULL 0

/* check for greater than alarm conditions */
	{
	uint32_t alarm_condition=0;

	if (lp->set_params_gte_hsp.notification_type == PSDB_GREATER_THAN_OR_EQUAL ) {

		if ( ( lp->hsp.hsp_rx_pctutil >= lp->set_params_gte_hsp.hsp_rx_pctutil )
			&& ( lp->set_params_gte_hsp.hsp_rx_pctutil != PSDB_ALARM_NULL ) )
			alarm_condition=1;
	
		if ( ( lp->hsp.hsp_tx_pctutil >= lp->set_params_gte_hsp.hsp_tx_pctutil ) 
			&& ( lp->set_params_gte_hsp.hsp_tx_pctutil != PSDB_ALARM_NULL ) )
			alarm_condition=1;	       
	
		if ( (lp->hsp.hsp_rx_pps >= lp->set_params_gte_hsp.hsp_rx_pps )
			&& ( lp->set_params_gte_hsp.hsp_rx_pps != PSDB_ALARM_NULL ) )
			alarm_condition=1;
	
		if ( (lp->hsp.hsp_tx_pps >= lp->set_params_gte_hsp.hsp_tx_pps )
			&& ( lp->set_params_gte_hsp.hsp_tx_pps !=  PSDB_ALARM_NULL ) )
			alarm_condition=1;

		if ( (lp->hsp.hsp_rx_eps >= lp->set_params_gte_hsp.hsp_rx_eps )
			&& ( lp->set_params_gte_hsp.hsp_rx_eps != PSDB_ALARM_NULL ) )
			alarm_condition=1;
	
		if ( (lp->hsp.hsp_tx_eps >= lp->set_params_gte_hsp.hsp_tx_eps )
			&& ( lp->set_params_gte_hsp.hsp_tx_eps !=  PSDB_ALARM_NULL ) )
			alarm_condition=1;

		if ( (lp->hsp.hsp_rx_ips >= lp->set_params_gte_hsp.hsp_rx_ips )
			&& ( lp->set_params_gte_hsp.hsp_rx_ips != PSDB_ALARM_NULL ) )
			alarm_condition=1;
	
		if ( (lp->hsp.hsp_tx_ips >= lp->set_params_gte_hsp.hsp_tx_ips )
			&& ( lp->set_params_gte_hsp.hsp_tx_ips !=  PSDB_ALARM_NULL ) )
			alarm_condition=1;

		if (alarm_condition)
			hal_psdb_speedometer_notification(ppa,PSDB_GREATER_THAN_OR_EQUAL);
	
		}
	}

/* check for less than or equal to alarm conditions */

	{
	uint32_t alarm_condition=0;
	if (lp->set_params_gte_hsp.notification_type == PSDB_LESS_THAN) {

		if ( ( lp->hsp.hsp_rx_pctutil < lp->set_params_lt_hsp.hsp_rx_pctutil )
			&& ( lp->set_params_lt_hsp.hsp_rx_pctutil != PSDB_ALARM_NULL ) )
			alarm_condition=1;
	
		if ( ( lp->hsp.hsp_tx_pctutil < lp->set_params_lt_hsp.hsp_tx_pctutil )
			&& ( lp->set_params_lt_hsp.hsp_tx_pctutil != PSDB_ALARM_NULL ) )
			alarm_condition=1;
	
		if ( ( lp->hsp.hsp_rx_pps < lp->set_params_lt_hsp.hsp_rx_pps )
			&& ( lp->set_params_lt_hsp.hsp_rx_pps != PSDB_ALARM_NULL ) )
			alarm_condition=1;
	
		if ( ( lp->hsp.hsp_tx_pps < lp->set_params_lt_hsp.hsp_tx_pps )
			&& ( lp->set_params_lt_hsp.hsp_tx_pps != PSDB_ALARM_NULL ) )
			alarm_condition=1;

		if ( (lp->hsp.hsp_rx_eps < lp->set_params_lt_hsp.hsp_rx_eps )
			&& ( lp->set_params_lt_hsp.hsp_rx_eps != PSDB_ALARM_NULL ) )
			alarm_condition=1;
	
		if ( (lp->hsp.hsp_tx_eps < lp->set_params_lt_hsp.hsp_tx_eps )
			&& ( lp->set_params_lt_hsp.hsp_tx_eps !=  PSDB_ALARM_NULL ) )
			alarm_condition=1;

		if ( (lp->hsp.hsp_rx_ips < lp->set_params_lt_hsp.hsp_rx_ips )
			&& ( lp->set_params_lt_hsp.hsp_rx_ips != PSDB_ALARM_NULL ) )
			alarm_condition=1;
	
		if ( (lp->hsp.hsp_tx_ips < lp->set_params_lt_hsp.hsp_tx_ips )
			&& ( lp->set_params_lt_hsp.hsp_tx_ips !=  PSDB_ALARM_NULL ) )
			alarm_condition=1;
	 
		if (alarm_condition)
			hal_psdb_speedometer_notification(ppa,PSDB_LESS_THAN);
	
		}
	}

	/* now remember the current totals */

	lp->prev_Statistics.hs_tx_packets = lp->Statistics.hs_tx_packets;
	lp->prev_Statistics.hs_rx_packets = lp->Statistics.hs_rx_packets;
	lp->prev_Statistics.hs_tx_err_packets = lp->Statistics.hs_tx_err_packets;
	lp->prev_Statistics.hs_rx_err_packets = lp->Statistics.hs_rx_err_packets;
	lp->prev_tx_icount = lp->tx_icount;
	lp->prev_rx_icount = lp->rx_icount;
	if (lp->tx_bytes > tx_bytes) tx_bytes=lp->tx_bytes;
	if (lp->rx_bytes > rx_bytes) rx_bytes=lp->rx_bytes;
	lp->tx_bytes=0;
	lp->rx_bytes=0;

	lp->hal_speedometer_timeout_id = pdk_timeout(hal_speedometer_timeout, (void *)ppa, HAL_SPEEDOMETER_TICKS);
}


#endif

int led_test = 2;


uint32_t
pdk_hal_ctl(ppa, opcode, param)
int32_t ppa;
uint32_t opcode;
void *param;
{
	local_adapter_t	*lp;
	uint32_t		err;
	uint32_t		local_settings;


	lp = local_adapters[ppa];

	switch(opcode) {
		case PDK_PSDB:
		case PDK_PSDB_INLINE: {
			pdk_psdb_rhdr_t *prp;
			uint32_t record_type;

			prp = (pdk_psdb_rhdr_t *) param;

			record_type = psdb_pton(prp->ppr_header.Type);
			ppa = (int32_t) psdb_pton(prp->ppr_ppa);

			err = hal_psdb(prp);

			/* Inline calls return the error, they do not create
			 * a PSDB status upcall.
			 */
			if (opcode == PDK_PSDB_INLINE)
				return(err);

			hal_psdb_return(lp, record_type, err);

			if ((record_type == PSDB_TYPE_HAL_GET_STATE) &&
				(err == 0)) {

				hal_psdb_return_state(lp, (pdk_psdb_hal_state_t *) prp);
			}

#ifdef PDK_HAL_SPEEDO
			if ((record_type == PSDB_TYPE_HAL_SPEEDOMETER_GET) &&
				(err == 0)) {

				hal_psdb_return_speed(lp, (pdk_psdb_hal_speedometer_get_t *) prp);
			}
#endif
			break;
		}

		case PDK_HAL_EXT_FAULT:
			if (lp == 0)
				break;

			zxe_external_fault_led(ppa, (int32_t) param, 1);
			break;

		case PDK_HAL_MTU_SIZE:
			return(PDK_ETHERMTU);

		case PDK_HAL_RECEIVE_INIT:
			return(pdk_buf_receive_init(ppa));
			break;

		case PDK_HAL_RECEIVE_DEINIT:
			pdk_buf_receive_deinit(ppa);
			break;

		case PDK_HAL_GET_NUM_RECEIVE_BUFS:
			*((uint32_t *) param) = lp->rx_num_bufs;
			break;

		case PDK_HAL_ENABLE_INTERRUPTS:
			SET_CSR7(ppa,CSR7_CFG);
			break;

		case PDK_HAL_DISABLE_INTERRUPTS:
			SET_CSR7(ppa,0);
			break;

		case PDK_HAL_GET_INFLIGHT:
			break;

		case PDK_HAL_GETID:
			*((uint32_t *) param) = lp->AdapterId;
			break;

		case PDK_HAL_SETID:
			break;

		case PDK_HAL_RESTART:
			pdk_hal_restart(ppa);
			break;

		case PDK_HAL_INIT_AND_START:
			if (lp->rx_descriptor_lock) {
				pdk_lock_acquire(lp->rx_descriptor_lock);
			}

			if (lp->tx_descriptor_lock) {
				pdk_lock_acquire(lp->tx_descriptor_lock);
			}

			err = (uint32_t) pdk_hal_start(ppa,1);

			if (lp->tx_descriptor_lock) {
				pdk_lock_release(lp->tx_descriptor_lock);
			}

			if (lp->rx_descriptor_lock) {
				pdk_lock_release(lp->rx_descriptor_lock);
			}

			return(err);

		case PDK_HAL_STOP_AND_RESET:
			if (lp->rx_descriptor_lock) {
				pdk_lock_acquire(lp->rx_descriptor_lock);
			}

			if (lp->tx_descriptor_lock) {
				pdk_lock_acquire(lp->tx_descriptor_lock);
			}

			pdk_hal_stop(ppa);
			lp->state = LOCAL_ADAPTER_ALLOCATED;

			if (lp->tx_descriptor_lock) {
				pdk_lock_release(lp->tx_descriptor_lock);
			}

			if (lp->rx_descriptor_lock) {
				pdk_lock_release(lp->rx_descriptor_lock);
			}
			pdk_time_delay(1);

			/* Need to do buffer recycling ?????? */
			break;

		case PDK_HAL_DEBUG:
			pdk_hal_debug = (int32_t) param;
			break;

		case PDK_HAL_REVLEVEL:
			if (lp == 0)
				return(PDK_OUT_STATE);

			break;

		case PDK_HAL_GET_COUNTERS:
			return(hal_get_stats(ppa, (pdk_hal_stat_t *) param));

		case PDK_HAL_SET_COUNTERS:
			return(hal_set_stats(ppa, (pdk_hal_stat_t *) param));

		case PDK_HAL_GET_HDWRADDR:
			if (lp == 0)
				return(PDK_OUT_STATE);

			pdk_bcopy(lp->PermanentNetworkAddress,
					(uchar *) param, PDK_ETHERADDRL);
			break;

		case PDK_HAL_GET_PHYSADDR:
			if (lp == 0)
				return(PDK_OUT_STATE);

			pdk_bcopy(lp->CurrentNetworkAddress,
					(uchar *) param, PDK_ETHERADDRL);

			break;

		case PDK_HAL_SET_PHYSADDR:
			if (lp == 0)
				return(PDK_OUT_STATE);

			pdk_bcopy((uchar *) param,
					lp->CurrentNetworkAddress, PDK_ETHERADDRL);

			if (lp->tx_descriptor_lock) {
				pdk_lock_acquire(lp->tx_descriptor_lock);
			}

			err = 0;
			/* If not running address will be setup and loaded
			 * on PDK_HAL_INIT_AND_START.
			 */
			if (lp->state == LOCAL_ADAPTER_RUNNING) {
				init_setup_buf(lp->ppa);
				err = load_setup_buf(lp->ppa);
			}

			if (lp->tx_descriptor_lock) {
				pdk_lock_release(lp->tx_descriptor_lock);
			}

			return(err);

		case PDK_HAL_SET_PROMISCUOUS_PHYS:
			if (lp == 0)
				return(PDK_OUT_STATE);

			lp->PacketFilter |= H_PROMISCUOUS_PHYSICAL;
			lp->PacketFilter &= ~H_PROMISCUOUS_MULTICAST;

			if (lp->state == LOCAL_ADAPTER_RUNNING) 
				ZxeStartTxAndRx(ppa);
			break;


		case PDK_HAL_SET_PROMISCUOUS_MULTI:
			if (lp == 0)
				return(PDK_OUT_STATE);

			lp->PacketFilter |= H_PROMISCUOUS_MULTICAST;
			lp->PacketFilter &= ~H_PROMISCUOUS_PHYSICAL;

			if (lp->state == LOCAL_ADAPTER_RUNNING) 
				ZxeStartTxAndRx(ppa);
			break;

		case PDK_HAL_CLEAR_PROMISCUOUS:
			if (lp == 0)
				return(PDK_OUT_STATE);

			lp->PacketFilter &= ~H_PROMISCUOUS_MULTICAST;
			lp->PacketFilter &= ~H_PROMISCUOUS_PHYSICAL;

			if (lp->state == LOCAL_ADAPTER_RUNNING) 
				ZxeStartTxAndRx(ppa);

			break;

		case PDK_HAL_ADD_MULTICAST:
			if (lp == 0)
				return(PDK_OUT_STATE);

			if (lp->MCAddressCount+1 == ZXE_MAXIMUM_MULTICAST_ADDRESSES) {
				return(PDK_NO_ROOM);
			}

			return(add_multicast_address(lp->ppa, (uchar *) param));

		case PDK_HAL_DEL_MULTICAST:
			if (lp == 0)
				return(PDK_OUT_STATE);

			if (lp->MCAddressCount == 0) {
				return(PDK_NOT_FOUND);
			}

			return(del_multicast_address(lp->ppa, (uchar *) param));

		case PDK_HAL_CLEAR_MULTICAST:
			if (lp == 0)
				return(PDK_OUT_STATE);

			lp->MCAddressCount = 0;
			if (lp->tx_descriptor_lock) {
				pdk_lock_acquire(lp->tx_descriptor_lock);
			}

			if (lp->state == LOCAL_ADAPTER_RUNNING) {
				init_setup_buf(lp->ppa);
				err = load_setup_buf(lp->ppa);
			}

			if (lp->tx_descriptor_lock) {
				pdk_lock_release(lp->tx_descriptor_lock);
			}

			return(err);

		case PDK_HAL_DUMPCAM:

		case PDK_HAL_DUMPRING:

       		case PDK_HAL_LED:
            		break;

		case PDK_HAL_SET_MEDIATYPE:
			if (lp == 0)
				return(PDK_OUT_STATE);

			if (((int32_t) param) > ZXE_MEDIA_MAX)
				return(PDK_BAD_PARAM);

			lp->ConfiguredMediaType = (int32_t) param;
			break;

		case PDK_HAL_GET_MEDIATYPE:
			if (lp == 0)
				return(PDK_OUT_STATE);

			*((uint32_t *) param) = lp->CurrentMediaType;
			break;

		case PDK_HAL_SET_TX_PENDING_THRESHOLD:
			lp->tx_pending_threshold = (int32_t) param;
			lp->tx_pending_count = (int32_t) 0;
			break;

		case PDK_HAL_HOTSWAP:
			if (lp == 0) {
				return(PDK_OUT_STATE);
			} else if ((int32_t) param == 0) {	/* card gone */
				if (lp->state == LOCAL_ADAPTER_ALLOCATED) {
					lp->state = LOCAL_ADAPTER_RELEASED;
					lp->srom_image_valid = 0;
					lp->PermanentNetworkAddress[0] =
						lp->PermanentNetworkAddress[1] =
						lp->PermanentNetworkAddress[2] = 0;
				} else
					return(PDK_OUT_STATE);
			} else {	/* card back */
				if (lp->state == LOCAL_ADAPTER_RELEASED) 
					lp->state = LOCAL_ADAPTER_ALLOCATED;
				else
					return(PDK_OUT_STATE);
			}
			break;

#ifdef PDK_HAL_SPEEDO
		case PDK_HAL_SET_SPEEDOMETER_POLL_RATE_AND_START:
			/* input is in ms */
			DEBUG2("SPEEDOMETER start: PPA: %d\n",ppa);

			if (param == 0) {
				DEBUG1("pdk_hal_ctl: ZERO poll interval\n");
				return(PDK_BAD_PARAM);
			}

			lp->speedometer_poll_rate = (uint32_t)param;
			if (lp->hal_speedometer_timeout_id == 0) {
				lp->hal_speedometer_timeout_id = pdk_timeout(hal_speedometer_timeout, (void *)ppa, HAL_SPEEDOMETER_TICKS);
			}

			break;
		case PDK_HAL_STOP_SPEEDOMETER_POLL:
			if (lp->hal_speedometer_timeout_id != 0) {
				pdk_untimeout(lp->hal_speedometer_timeout_id);
				lp->hal_speedometer_timeout_id = 0;
			}
			lp->speedometer_poll_rate = 0;
			break;
#endif
		default:
			return(PDK_BAD_PARAM);
	}

	return(0);
}

void
hal_psdb_return(lp, type, return_val)
local_adapter_t *lp;
uint32_t type;
uint32_t return_val;
{
	int32_t ppa;
	pdk_psdb_rhdr_t *prp;
	pdk_psdb_success_t ps;
	pdk_psdb_error_t pe;
	
	ppa = lp->ppa;

	if (return_val == 0) {
		ps.ps_header.ppr_header.Size = psdb_ntop(sizeof(ps));
		ps.ps_header.ppr_header.Type = psdb_ntop(PSDB_TYPE_SUCCESS);
		ps.ps_header.ppr_ppa = psdb_ntop(ppa);
		ps.ps_success_type = psdb_ntop(type);

		prp = &ps.ps_header;
	}
	else {
		pe.pe_header.ppr_header.Size = psdb_ntop(sizeof(pe));
		pe.pe_header.ppr_header.Type = psdb_ntop(PSDB_TYPE_ERROR);
		pe.pe_header.ppr_dev_id = psdb_ntop(ppa);
		pe.pe_error_type = psdb_ntop(type);
		pe.pe_error = psdb_ntop(return_val);

		prp = &pe.pe_header;
	}

	(*lp->layer_routines.pdk_status)(ppa, PDK_PSDB, (void *) prp);
}

void
hal_psdb_return_state(lp, phsp)
local_adapter_t *lp;
pdk_psdb_hal_state_t *phsp;
{
	DEBUG2("hal_psdb_return_state(%d): Called\n", lp->ppa);

	(*lp->layer_routines.pdk_status)(lp->ppa, PDK_PSDB, (void *) phsp);
}

#ifdef PDK_HAL_SPEEDO
void
hal_psdb_return_speed(lp, phsp)
local_adapter_t *lp;
pdk_psdb_hal_speedometer_get_t *phsp;
{
	DEBUG2("hal_psdb_return_speed(%d): Called\n", lp->ppa);

	(*lp->layer_routines.pdk_status)(lp->ppa, PDK_PSDB, (void *) phsp);

}
#endif

uint32_t
hal_psdb(record)
pdk_psdb_rhdr_t *record;
{
	uint32_t record_type, record_size;
	uint32_t *lp;
	int32_t ppa;
	int32_t i;

	record_type = psdb_pton(record->ppr_header.Type);
	record_size = psdb_pton(record->ppr_header.Size);

	lp = (uint32_t *) record;
	for (i=0;i<(int32_t)(record_size/sizeof(uint32_t));i++)
		lp[i] = psdb_pton(lp[i]);

	ppa = (int32_t) record->ppr_ppa;

	/* Let hal_get_state check the ppa and fill in structure */
	if (record_type != PSDB_TYPE_HAL_GET_STATE) {
		if ((ppa < 0) || (ppa >= NUM_PORTS))
			return(PSDB_PPA_OUT_RANGE);
	}
		
	switch(record_type) {
		case PSDB_TYPE_HAL_GET_STATE: {
			pdk_psdb_hal_state_t *phsp;
			uint32_t err;

			phsp = (pdk_psdb_hal_state_t *) record;

			if (record_size != sizeof(pdk_psdb_hal_state_t))
				return(PSDB_INCORRECT_SIZE);

			err = hal_get_state(ppa, &phsp->phs_state);
			if (err)
				return(err);
			else {
				register uint32_t *lp, *record_lim;

				lp = (uint32_t *) record;
				record_lim = (uint32_t *)((char *)lp + record->ppr_header.Size);
				while (lp < record_lim)
					*lp++ = psdb_ntop(*lp);
			}
			return(0);
		}
		case PSDB_TYPE_HAL_FAULT_NOTIFICATION_SET:
		 	if (local_adapters[ppa] == 0)
				return(PSDB_PPA_OUT_RANGE);

			local_adapters[ppa]->fault_notification = 1;
			return(0);

		case PSDB_TYPE_HAL_FAULT_NOTIFICATION_CLEAR:
		 	if (local_adapters[ppa] == 0)
				return(PSDB_PPA_OUT_RANGE);

			local_adapters[ppa]->fault_notification = 0;
			return(0);

#ifdef PDK_HAL_SPEEDO
		case PSDB_TYPE_HAL_SPEEDOMETER_GET: {
			pdk_psdb_hal_speedometer_get_t *phsp;
			uint32_t err;

			phsp = (pdk_psdb_hal_speedometer_get_t *) record;

			if (record_size != sizeof(pdk_psdb_hal_speedometer_get_t))
				return(PSDB_INCORRECT_SIZE);

			err = hal_get_speed(ppa, &phsp->phsg_speed);
			if (err)
				return(err);
			else {
				register uint32_t *lp, *record_lim;

				lp = (uint32_t *) record;
				record_lim = (uint32_t *)((char *)lp + record->ppr_header.Size);
				while (lp < record_lim)
					*lp++ = psdb_ntop(*lp);
			}
			return(0);
		}
		case PSDB_TYPE_HAL_SPEEDOMETER_NOTIFICATION: {
			/* XXX get alam thresholds here */
			return(0);
		}
		case PSDB_TYPE_HAL_SPEEDOMETER_SET_PARAMS: {
			pdk_psdb_hal_speedometer_set_params_t *phsp;

			phsp = (pdk_psdb_hal_speedometer_set_params_t *) record;

			if (record_size != sizeof(pdk_psdb_hal_speedometer_set_params_t))
				return(PSDB_INCORRECT_SIZE);

			return(hal_set_speedometer_params(ppa, &phsp->phss_speed));
		}
#endif
		default:
			DEBUG3("hal_psdb:  Record type 0x%x unknown\n", record_type);
			break;
	}

	return(PSDB_PROTOCOL_ERROR);
}

void
ZxeStartTxAndRx(ppa)
int32_t ppa;
{
	local_adapter_t *lp;

	lp = local_adapters[ppa];

	/* Setup OperationMode and output to chip */

	lp->OperationMode = lp->OperationModeBase | CSR6_ST | CSR6_SR;

	if (lp->PacketFilter & H_PROMISCUOUS_MULTICAST)
		lp->OperationMode |= CSR6_PM;

	if (lp->PacketFilter & H_PROMISCUOUS_PHYSICAL)
		lp->OperationMode |= CSR6_PR;

	pdk_reg_put32(ppa, CSR6, lp->OperationMode);
}

#define HAL_DEVICE_STOP_DELAY 2

int32_t
ZxeStopTxAndRx(ppa)
int32_t ppa;
{
	local_adapter_t *lp;
	int32_t i;

	lp = local_adapters[ppa];

	/* Stop TX and RX */

	lp->OperationMode = lp->OperationModeBase;

	pdk_reg_put32(ppa, CSR6, lp->OperationMode);

	/* Wait for processes to enter stopped state */
	for (i = 0; i < HAL_DEVICE_STOP_DELAY;i++) {

		pdk_time_delay(1);

		if ((pdk_reg_get32(ppa, CSR5) & (CSR5_TS_MASK|CSR5_RS_MASK))
			== (CSR5_TS_STOPPED|CSR5_RS_STOPPED)) {

			return(0);	/* Device Stopped */
		}
	}

	return(1);	/* Failed to stop */
}

/* MII/SYM management code */

void
ZxeWriteMIIBit(ppa, Data)
int32_t ppa;
int16_t Data;
{
	uint32_t	Base = Data ? CSR9_MMD_WR1 : CSR9_MMD_WR0;

	pdk_reg_put32(ppa, CSR9, Base);
	pdk_reg_put32(ppa, CSR9, Base | CSR9_MDC);
	pdk_reg_put32(ppa, CSR9, Base | CSR9_MDC);
	pdk_reg_put32(ppa, CSR9, Base);

	return;
}


void
ZxeTristateMIIBit(ppa)
{
	pdk_reg_put32(ppa, CSR9, CSR9_MMD_RD);
	pdk_reg_put32(ppa, CSR9, CSR9_MMD_RD | CSR9_MDC);
	pdk_reg_put32(ppa, CSR9, CSR9_MMD_RD | CSR9_MDC);
	pdk_reg_put32(ppa, CSR9, CSR9_MMD_RD);

	return;
}


uint16_t
ZxeReadMIIBit(ppa, Data)
int32_t ppa;
uint16_t Data;
{
	pdk_reg_put32(ppa, CSR9, CSR9_MMD_RD);
	pdk_reg_put32(ppa, CSR9, CSR9_MMD_RD | CSR9_MDC);
	pdk_reg_put32(ppa, CSR9, CSR9_MMD_RD | CSR9_MDC);
	pdk_reg_put32(ppa, CSR9, CSR9_MMD_RD);

	return ((Data<<1) | (pdk_reg_get32(ppa, CSR9) & CSR9_MDI ? 1 : 0));
}

int32_t
ZxeReadMIIRegister(ppa, Phy, Addr, Ptr)
int32_t ppa;
int32_t Phy;
int32_t Addr;
uint16_t *Ptr;
{
#ifdef NOTDEF
	int_t Extra = 0;
	uint32_t bmsr = 0;

	if (ZxeReadMIIRegisterX(ppa, Phy, 1, 0, &bmsr))
		Extra = 1;
	else if(bmsr & 0x780)
		Extra = 1;

	return(ZxeReadMIIRegisterX(ppa, Phy, Addr, Extra, Ptr));
}

int32_t
ZxeReadMIIRegisterX(ppa, Phy, Addr, Extra, Ptr)
int32_t ppa;
int32_t Phy;
int32_t Addr;
int32_t	Extra;
uint16_t *Ptr;
{
#endif
	uint16_t	Data;
	int_t	i;

	/* Preamble and Start */

	PHYLOCK(ppa);

	pdk_reg_put32(ppa, CSR9, 0);

	for (i=0; i<32; i++)
		ZxeWriteMIIBit(ppa, 1);

	ZxeWriteMIIBit(ppa, 0);
	ZxeWriteMIIBit(ppa, 1);

	/* Operation type */

	ZxeWriteMIIBit(ppa, 1);
	ZxeWriteMIIBit(ppa, 0);

	/* PHY address */

	for (i=0; i<5; i++)
		ZxeWriteMIIBit(ppa, (Phy>>(4-i)) & 1);

	/* Register address */

	for (i=0; i<5; i++)
		ZxeWriteMIIBit(ppa, (Addr>>(4-i)) & 1);

	/* Turnaround */
	Data = ZxeReadMIIBit(ppa, 0);

#ifdef NOTDEF
	if (Extra)
#else
	if (Data)
#endif
		Data = ZxeReadMIIBit(ppa, 0);

#ifdef NOTDEF
	if (Data) {
		PHYUNLOCK(ppa);

		DEBUG2("ZxeReadMIIRegister(%d): Turnaround error\n", ppa);
		return(PDK_HW_ERR);
	}
#endif
	
	/* Read data */
	for (Data=0, i=0; i<16; i++) {
		Data = ZxeReadMIIBit(ppa, Data);
	}

	ZxeTristateMIIBit(ppa);

	PHYUNLOCK(ppa);

	if (Ptr) *Ptr = Data;

	DEBUG4("ZxeReadMIIRegister(%d,%d): Data 0x%x\n",ppa, Addr, Data);
	return(0);
}

int32_t
ZxeWriteMIIRegister(ppa, Phy, Addr, Word)
int32_t ppa;
int32_t Phy;
int32_t Addr;
uint16_t Word;
{
	int16_t Data;
	int_t i;

	PHYLOCK(ppa);

	/* Preamble and Start */
	pdk_reg_put32(ppa, CSR9, 0);

	for (i=0; i<32; i++)
		ZxeWriteMIIBit(ppa, 1);

	ZxeWriteMIIBit(ppa, 0);
	ZxeWriteMIIBit(ppa, 1);

	/* Operation type */

	ZxeWriteMIIBit(ppa, 0);
	ZxeWriteMIIBit(ppa, 1);

	/* PHY address */

	for (i=0; i<5; i++)
		ZxeWriteMIIBit(ppa, (Phy>>(4-i)) & 1);

	/* Register address */

	for (i=0; i<5; i++)
		ZxeWriteMIIBit(ppa, (Addr>>(4-i)) & 1);

	/* Turnaround */

	ZxeWriteMIIBit(ppa, 1);
	ZxeWriteMIIBit(ppa, 0);

	/* Write data */

	for (Data=0, i=0; i<16; i++) {
		ZxeWriteMIIBit(ppa, (Word >> (15-i)) & 1);
	}

	ZxeTristateMIIBit(ppa);

	PHYUNLOCK(ppa);

	return(0);
}


int32_t
ZxeDeterminePhyAddress(ppa)
int32_t ppa;
{
	local_adapter_t *lp;
	int32_t		p;
	uint16_t	Data;

	if ((lp = local_adapters[ppa])  == 0) 
		return(-1);

	if (lp->PhyType == DEVICE_BCM5208) {
		if (lp->Info.AIFlags & AI_SHARED_PHYLOCK) {

			p = pdk_ntbridge_unit_ppa(ppa, -1);
			p ^= 3;
			p &= 7;
	
			DEBUG2("ZxeDeterminePhyAddress(%d):  Broadcom 5208 phy address %d\n", ppa, p);
			return(p);
		} else
			return(1);	/* Always use address 1 for non shared PHY */
	}

	for (p=1; p<32; p++) {
		if ((ZxeReadMIIRegister(ppa, p, 2, &Data) != 0) || Data == 0xffff)
			continue;
		if ((ZxeReadMIIRegister(ppa, p, 3, &Data) != 0) || Data == 0xffff)
			continue;

		return(p);
	}

	return(-1);
}

int32_t
ZxeReadPHY(ppa, Addr, Ptr)
int32_t ppa;
int32_t Addr;
uint16_t *Ptr;
{
	int32_t PhyAddress = local_statics[ppa].PhyAddress;

	DEBUG4("ZxeReadPHY(%d): Addr 0x%x\n", ppa, Addr);

	if (PhyAddress < 0) {
		if ((PhyAddress = ZxeDeterminePhyAddress(ppa)) < 0)
			return(PDK_HW_ERR);
		else
			local_statics[ppa].PhyAddress = PhyAddress;
	}

	if (ZxeReadMIIRegister(ppa, PhyAddress, Addr, Ptr) != 0) {
		return(PDK_HW_ERR);
	}

	DEBUG4("ZxeReadPHY(%d): 0x%x\n", ppa, Ptr?*Ptr:0);
	return(0);
}


int32_t
ZxeWritePHY(ppa, Addr, Word)
int32_t ppa;
int32_t Addr;
uint16_t Word;
{
	int32_t PhyAddress = local_statics[ppa].PhyAddress;

	DEBUG4("ZxeWritePHY(%d): Addr 0x%x\n", ppa, Addr);

	if (PhyAddress < 0) {
		if ((PhyAddress = ZxeDeterminePhyAddress(ppa)) < 0)
			return(PDK_HW_ERR);
		else
			local_statics[ppa].PhyAddress = PhyAddress;
	}

	if (ZxeWriteMIIRegister(ppa, PhyAddress, Addr, Word) != 0) {
		return(PDK_HW_ERR);
	}

	DEBUG4("ZxeWritePHY(%d): 0x%x\n", ppa, Word);
	return(0);
}

#define CSR9_SROM (CSR9_SWO | CSR9_SSR | CSR9_SCS)

/* Write a bit to the microwire eeprom */

int32_t
ZxeWriteMicroWireBit(ppa, Data)
int32_t ppa;
int16_t Data;
{
	uint32_t Base;

	Base = CSR9_SROM | (Data?CSR9_SDI:0);

	pdk_reg_put32(ppa, CSR9, Base);
	pdk_reg_put32(ppa, CSR9, Base);

	pdk_reg_put32(ppa, CSR9, Base | CSR9_SCLK);
	pdk_reg_put32(ppa, CSR9, Base | CSR9_SCLK);
	pdk_reg_put32(ppa, CSR9, Base | CSR9_SCLK);
	pdk_reg_put32(ppa, CSR9, Base | CSR9_SCLK);

	pdk_reg_put32(ppa, CSR9, Base);
	pdk_reg_put32(ppa, CSR9, Base);

	return(0);
}



/* Read a bit from the microwire eeprom */

uint16_t
ZxeReadMicroWireBit(ppa, Data)
int32_t ppa;
uint16_t Data;
{

	pdk_reg_put32(ppa, CSR9, CSR9_SROM);
	pdk_reg_put32(ppa, CSR9, CSR9_SROM);

	pdk_reg_put32(ppa, CSR9, CSR9_SROM | CSR9_SCLK);
	pdk_reg_put32(ppa, CSR9, CSR9_SROM | CSR9_SCLK);
	pdk_reg_put32(ppa, CSR9, CSR9_SROM | CSR9_SCLK);
	pdk_reg_put32(ppa, CSR9, CSR9_SROM | CSR9_SCLK);

	pdk_reg_put32(ppa, CSR9, CSR9_SROM);
	pdk_reg_put32(ppa, CSR9, CSR9_SROM);

	return( Data<<1 | (pdk_reg_get32(ppa, CSR9) & CSR9_SDO ? 1 : 0));
}



/* Make sure the eeprom is not in the middle of an operation */

int32_t
ZxeMakeSureMicroWireIsIdle(ppa)
int32_t ppa;
{
	int_t	i;

	pdk_reg_put32(ppa, CSR9, CSR9_SWO|CSR9_SSR);

	for (i=0; i<3+6+16; i++) {
		if (ZxeWriteMicroWireBit(ppa, 0) != 0) {
			return(PDK_HW_ERR);
		}
	}

	pdk_reg_put32(ppa, CSR9, CSR9_SWO|CSR9_SSR);

	return(0);
}



/* Read a 16bit word from the microwire eeprom */

int16_t
ZxeReadMicroWireWord(ppa, Addr, Ptr)
int32_t ppa;
int32_t Addr;
uint16_t *Ptr;
{
	int16_t		Bit;
	int_t		i;
	int16_t		Data;
	uint32_t	Status;


	if (ZxeWriteMicroWireBit(ppa, 1) != 0)
		return(PDK_HW_ERR);
	if (ZxeWriteMicroWireBit(ppa, 1) != 0)
		return(PDK_HW_ERR);
	if (ZxeWriteMicroWireBit(ppa, 0) != 0)
		return(PDK_HW_ERR);

	for (i=0; i<6; i++)
		ZxeWriteMicroWireBit(ppa, (Addr>>(5-i)) & 1);

	Status = pdk_reg_get32(ppa, CSR9);
	if (Status & CSR9_SDO) {
		return(PDK_HW_ERR);
	}

	/* Read data */
	for (Data=0, i=0; i<16; i++) {
		Data = ZxeReadMicroWireBit(ppa, Data);
	}

	/* Drop CS */
	pdk_reg_put32(ppa, CSR9, CSR9_SWO|CSR9_SSR);

	if (Ptr) *Ptr = Data;

	return(0);
}


/* Read a number of bytes from the node address eeprom */

uint32_t
ZxeReadMicroWireBytes(ppa, Buffer, Count)
int32_t ppa;
uchar *Buffer;
int32_t Count;
{
	uint16_t	Word=0;
	uchar		*Ptr = Buffer;
	int32_t		wcount=0, waddr=0;
	int32_t		Error = 0;

	PHYLOCK(ppa);

	Error = ZxeMakeSureMicroWireIsIdle(ppa);
	if (Error == 0) {
		wcount = (Count+1)/2;
		for (waddr=0; wcount; wcount--, waddr++, Count-=2) {
			Error = ZxeReadMicroWireWord(ppa, waddr, &Word);
			if (Error != 0)
				break;

			*Ptr++ = (uchar)Word;
			if (Count>1) *Ptr++ = (uchar)(Word>>8);
		}
	}

	pdk_reg_put32(ppa, CSR9, CSR9_IDLE);
	pdk_time_delay(1);

	PHYUNLOCK(ppa);

	return(Error);
}


static uchar srom_image[SROM_SIZE];

uint32_t
read_srom_device(ppa, buffer)
int32_t ppa;
uchar **buffer;
{
	local_adapter_t *lp;
	int32_t Error;
	uint32_t csr0;
	uint32_t Id;
	uint32_t ppa_zero;

	*buffer = srom_image;

	if ((lp = local_adapters[ppa]) != 0) {
		*buffer = lp->srom_image;
		if (lp->srom_image_valid)
			return(0);
	}

	if ((Error = ZxeReadMicroWireBytes(ppa, *buffer, 128)) == 0)
		if (lp) lp->srom_image_valid = 1;

	return(Error);
}


uint32_t
ZxeComputeSromCRC(Buffer, Length)
uchar *Buffer;
int32_t Length;
{
	uint32_t	crc = 0xFFFFFFFF;
	uint32_t	flipped = 0;
	uchar		ch;
	int_t		bit;
	int32_t		msb;


	while (Length--) {

		ch = *Buffer++;
		for (bit=0; bit<8; bit++) {
			msb = (crc>>31) & 1;
			crc <<= 1;
			if (msb ^ (ch & 1)) {
				crc ^= 0x04C11DB6L;
				crc |= 1;
			}
			ch >>= 1;
		}
	}

	for (bit=0; bit<32; bit++) {
		flipped <<= 1;
		flipped |= (crc & 1);
		crc >>= 1;
	}

	flipped ^= 0xFFFFFFFF;
	return(flipped);
}


uint32_t
ZxeReadNodeAddressRaw(ppa, Buffer, AdapterId)
int32_t ppa;
uchar *Buffer;
uint32_t *AdapterId;
{
	uint32_t	Error, Crc, Crc94;
	uchar		*Prom = 0, *Info;
	uint16_t	Offset, Id = 0;
	int_t		i;
	uchar		AddrOffset;

	AddrOffset = 0;

	if ((Error = read_srom_device(ppa, &Prom)) != 0) {
		DEBUG3("ZxeReadNodeAddressRaw(%d): read_srom_device failed.  Assuming single SROM adapter\n", ppa);
		for (i = (ppa-1); (i >= 0) && (Error != 0); i--) {
			/* Take chip out of sleep/snooze */
			pdk_reg_config_put32(i, CFDA, 0);

			Error = read_srom_device(i, &Prom);
			AddrOffset++;
		}

		if (Error != 0) {
			DEBUG1("ZxeReadNodeAddressRaw(%d): read_srom_device failed\n", ppa);
			return(Error);
		}
	}

	if (Prom[18]==4 && (Prom[20]&3)==0) {
		Crc94 = ZxeComputeSromCRC(Prom, 94);
		Crc = ZxeComputeSromCRC(Prom, 126);
		if ((uint16_t)Crc == pdk_htobs(*((uint16_t *)(Prom+126)))) {
			pdk_bcopy(Prom+20, Buffer, 6);
			if ((Offset = pdk_htobs(*((uint16_t *)(Prom+124)))) < 124) {
				Info = Prom+Offset;
				if (Info[0]=='J' && Info[1]=='R' && Info[2]==1)
					Id = pdk_htobs(*((uint16_t *)(Info+4)));
			}
		}
		else if ((uint16_t)Crc94 == pdk_htobs(*((uint16_t *)(Prom+94)))) {
			pdk_bcopy(Prom+20, Buffer, 6);
			if ((Offset = pdk_htobs(*((uint16_t *)(Prom+92)))) < 92) {
				Info = Prom+Offset;
				if (Info[0]=='J' && Info[1]=='R' && Info[2]==1)
					Id = pdk_htobs(*((uint16_t *)(Info+4)));
			}
		}
		else {
			local_adapter_t *lp = local_adapters[ppa-AddrOffset];
			if (lp)
				lp->srom_image_valid = 0;
			DEBUG1("ZxeReadNodeAddressRaw(%d): Type 4 - CRC error Crc94 0x%x!=0x%x Crc126 0x%x!=0x%x\n", ppa-AddrOffset, (uint16_t)Crc94, pdk_htobs(*((uint16_t *)(Prom+94))), (uint16_t)Crc, pdk_htobs(*((uint16_t *)(Prom+126))));
			return(PDK_HW_ERR);
		}
	}
	else if ((Prom[18]==1 || Prom[18]==3) && (Prom[20]&3)==0) {
		/* Check for version 1 or version 3 format */
		Crc = ZxeComputeSromCRC(Prom, 126);
		if ((uint16_t)Crc == pdk_htobs(*((uint16_t *)(Prom+126)))) {
			pdk_bcopy(Prom+20, Buffer, 6);
			if ((Offset = pdk_htobs(*((uint16_t *)(Prom+124)))) < 124) {
				Info = Prom+Offset;
				if (Info[0]=='J' && Info[1]=='R' && Info[2]==1)
					Id = pdk_htobs(*((uint16_t *)(Info+4)));
			}
		}
		else {
			local_adapter_t *lp = local_adapters[ppa-AddrOffset];
			if (lp)
				lp->srom_image_valid = 0;
			DEBUG1("ZxeReadNodeAddressRaw(%d): Type 3 - CRC error at 126: Computed 0x%x, Actual 0x%x\n", ppa, (uint16_t)Crc, pdk_htobs(*((uint16_t *)(Prom+126))));
			return(PDK_HW_ERR);
		}
	}
	else {
		DEBUG2("ZxeReadNodeAddressRaw(%d): Legacy format\n", ppa);
		/* Must be legacy format otherwise */
		if (Prom[0] == 0xFF && Prom[1] == 0xFF && Prom[2] == 0xFF)
			return(PDK_HW_ERR);
		if (Prom[0] == 0 && Prom[1] == 0 && Prom[2] == 0)
			return(PDK_HW_ERR);
		pdk_bcopy(Prom, Buffer, 6);
	}
	if (AddrOffset > 0) {
		zxe_adapterinfo_t	*InfoBase;
		for (InfoBase=ZxePhyInformation; InfoBase->AdapterId; InfoBase++) {
			if (InfoBase->AdapterId == (Id & ADAPTER_ZX_MASK)) 
				break;
		}
		if (InfoBase->AdapterId == 0 || !(InfoBase->AIFlags & AI_SHARED_SROM)
			|| (AddrOffset >= InfoBase->NumChannels)) {
			/* Error case */
			DEBUG1("ZxeReadNodeAddressRAW(%d): No Valid SROM\n",ppa);
			return(PDK_HW_ERR);
		}
	}

	Buffer[5] += AddrOffset;

	if (AdapterId) *AdapterId = Id;
	return(0);
}


uint32_t
get_network_address(ppa)
int32_t ppa;
{
	local_adapter_t *lp;
	uint32_t	 Error;

	/* No local adapter structure */
	if ((lp = local_adapters[ppa]) == 0) {
		return(PDK_OUT_STATE);
	}

	if ((Error = ZxeReadNodeAddressRaw(ppa, lp->PermanentNetworkAddress, &lp->AdapterId)) != 0) {
		return(Error);
	}

	if (!lp->CurrentNetworkAddress[0]
		&& !lp->CurrentNetworkAddress[1]
		&& !lp->CurrentNetworkAddress[2])	/* Check for OID == 0 */
		pdk_bcopy(lp->PermanentNetworkAddress, lp->CurrentNetworkAddress, PDK_ETHERADDRL);
	if (lp->AdapterId == 0) {
		return(PDK_HW_ERR);
	}

	return(0);
}


uint32_t
ZxeSetupAdapterInfo(ppa)
int32_t ppa;
{
	local_adapter_t		*lp;
	zxe_adapterinfo_t	*InfoBase, *Info;

	lp = local_adapters[ppa];

	DEBUG3("ZxeSetupAdapterInfo(%d): Called\n", ppa);

	Info = &lp->Info;
	Info->AdapterId = 0;

	/* Look for exact match */
	for (InfoBase=ZxePhyInformation; InfoBase->AdapterId; InfoBase++) {
		if (InfoBase->AdapterId == lp->AdapterId) 
			break;
	}

	if (InfoBase->AdapterId == 0) {
		/* Look for good match */
		for (InfoBase=ZxePhyInformation; InfoBase->AdapterId; InfoBase++) {
			if (InfoBase->AdapterId == (lp->AdapterId & ADAPTER_ZX_MASK)) 
				break;
		}
		if (InfoBase->AdapterId == 0) {
			/* Error case */
			DEBUG1("ZxeSetupAdapterInfo(%d): Unknown adapter=0x%X\n",
				ppa, lp->AdapterId);
			return(PDK_HW_ERR);
		}
	}
	pdk_bcopy(InfoBase, Info, sizeof(zxe_adapterinfo_t));
	lp->PhyType = lp->Info.PhyId;
	switch (lp->PhyType) {
		case DEVICE_QS6611:
			break;

		case DEVICE_ICS1890:
		case DEVICE_DP83840:
		case DEVICE_BCM5208:
		default:
			local_statics[ppa].PhyAddress = ZxeDeterminePhyAddress(ppa);
	}

	DEBUG3("ZNB(%d): AdapterId 0x%x, Phytype %d\n", ppa, Info->AdapterId, lp->PhyType);
	return(0);
}



int32_t
ZxePreConfigure21140Media(ppa)
int32_t ppa;
{
	local_adapter_t *lp;
	uint16_t	PhyControl = (uint16_t)-1;

	lp = local_adapters[ppa];

	if ((lp->PhyType == DEVICE_DP83840 ) ||
	   (lp->PhyType == DEVICE_DP83843)) {

		switch(lp->ConfiguredMediaType) {
		case ZXE_MEDIA_AUTO:
			PhyControl = DP83840_BMCR_ANENA;
			ZxeWritePHY(ppa, DP83840_BMCR_REG, (uint16_t)0);
			break;
		case ZXE_MEDIA_TP:
			PhyControl = 0;
			break;
		case ZXE_MEDIA_TPFD:
			PhyControl = DP83840_BMCR_FD;
			break;
		case ZXE_MEDIA_TX:
			PhyControl = DP83840_BMCR_100;
			break;
		case ZXE_MEDIA_TXFD:
			PhyControl = DP83840_BMCR_100 | DP83840_BMCR_FD;
			break;
		}

		ZxeWritePHY(ppa, DP83840_BMCR_REG, PhyControl);
		lp->PhyControl = PhyControl;
	}
	else if (lp->PhyType == DEVICE_ICS1890) {

		ZxeWritePHY(ppa, ICS1890_MIICR_REG, (uint16_t)ICS1890_MIICR_RESET);
		pdk_time_delay(500000);

		switch(lp->ConfiguredMediaType) {
		case ZXE_MEDIA_AUTO:
			PhyControl = ICS1890_MIICR_ANENA;
			break;
		case ZXE_MEDIA_TP:
			PhyControl = 0;
			break;
		case ZXE_MEDIA_TPFD:
			PhyControl = ICS1890_MIICR_FD;
			break;
		case ZXE_MEDIA_TX:
			PhyControl = ICS1890_MIICR_100;
			break;
		case ZXE_MEDIA_TXFD:
			PhyControl = ICS1890_MIICR_100 | ICS1890_MIICR_FD;
			break;
		}

		ZxeWritePHY(ppa, ICS1890_MIICR_REG, PhyControl);
		lp->PhyControl = PhyControl;
	}

	return(1);
}


int32_t
ZxePreConfigure21143Media(ppa)
int32_t ppa;
{
	local_adapter_t *lp;
	uint16_t	PhyControl = (uint16_t)-1;
	uint16_t	phyid;

	lp = local_adapters[ppa];

	if (lp->PhyType == DEVICE_QS6611) {
		pdk_reg_put32(ppa, CSR15, CSR15_142_CWE|0xF0000|CSR15_142_LGS1 | CSR15_142_ABM);
		pdk_time_delay(200);

		pdk_reg_put32(ppa, CSR15, 0x40000);  /* Hold off reset to phy */

		pdk_time_delay(10000);
	}
	else if (lp->PhyType == DEVICE_BCM5208) {
		pdk_reg_put32(ppa, CSR15, CSR15_142_CWE|CSR15_142_GEI0|0xA0000|CSR15_142_LGS1 | CSR15_142_ABM);

		if (lp->AdapterId == ADAPTER_ZX222)
			lp->ConfiguredMediaType = ZXE_MEDIA_TXFD;

		switch(lp->ConfiguredMediaType) {
		case ZXE_MEDIA_AUTO:
			PhyControl = BCM5208_MIICR_ANENA;
			break;
		case ZXE_MEDIA_TP:
			PhyControl = 0;
			break;
		case ZXE_MEDIA_TPFD:
			PhyControl = BCM5208_MIICR_FD;
			break;
		case ZXE_MEDIA_TX:
			PhyControl = BCM5208_MIICR_100;
			break;
		case ZXE_MEDIA_TXFD:
			PhyControl = BCM5208_MIICR_100 | BCM5208_MIICR_FD;
			break;
		}
		lp->PhyControl = PhyControl;

		/* Reset PHY - clears Super Isolation Mode */
		ZxeWritePHY(ppa, BCM5208_MIICR_REG, (uint16_t)BCM5208_MIICR_RESET);
		pdk_time_delay(2);
		ZxeWritePHY(ppa, BCM5208_MIICR_REG,(uint16_t)lp->PhyControl);
		if (lp->AdapterId == ADAPTER_ZX222)
			ZxeWritePHY(ppa, BCM5208_TXACR_REG,
				(uint16_t)(BCM5208_TXACR_FEF | BCM5208_TXACR_CIMDIS));
	}

	return(1);
}

int32_t
ZxePreConfigureMedia(ppa)
int32_t ppa;
{
	local_adapter_t *lp;
	int32_t		MediaPreConfigured;

	lp = local_adapters[ppa];

	lp->LinkSpeed = 10;

	if (lp->Info.NicType == DEVICE_DEC21140) {
		MediaPreConfigured = ZxePreConfigure21140Media(ppa);
	}
	else if (lp->Info.NicType == DEVICE_DEC21143) {
		MediaPreConfigured = ZxePreConfigure21143Media(ppa);
	}
	else {
		MediaPreConfigured = 0;
	}

	lp->CurrentMediaType = lp->ConfiguredMediaType;

	return(MediaPreConfigured);
}


int32_t
ZxeConfigure21140Media(ppa)
int32_t ppa;
{
	local_adapter_t *lp;
#ifdef DYNAMIC_THRESHOLD
	uint32_t	TxFifoThreshold;
#endif

	lp = local_adapters[ppa];

	lp->OperationModeBase = CSR6_BIT25|CSR6_SC|CSR6_CA|CSR6_HBD|CSR6_PS;

	switch(lp->CurrentMediaType) {
		default:
			lp->CurrentMediaType = ZXE_MEDIA_AUTO;
			/* Fall Through */

		case ZXE_MEDIA_AUTO:
			lp->LinkSpeed = 10;
			break;

		case ZXE_MEDIA_TX:
			lp->LinkSpeed = 100;
			break;

		case ZXE_MEDIA_TXFD:
			lp->OperationModeBase |= CSR6_FD;
			lp->LinkSpeed = 100;
			break;

		case ZXE_MEDIA_TP:
			lp->LinkSpeed = 10;
			break;

		case ZXE_MEDIA_TPFD:
			lp->OperationModeBase |= CSR6_FD;
			lp->LinkSpeed = 10;
			break;
	}

	if (lp->LinkSpeed > 10) {
#ifdef DYNAMIC_THRESHOLD
		TxFifoThreshold = (lp->TxFifoThreshold+99)/100;
		if (TxFifoThreshold < NUM_TXFIFOCONFIG)
			lp->OperationModeBase |= TxFifoConfigurations[TxFifoThreshold];
		else
			lp->OperationModeBase |= CSR6_TFF_1024 | CSR6_FP;
#else
		lp->OperationModeBase |= CSR6_TR_160 | CSR6_FP;
#endif
	}
	else {
		lp->OperationModeBase |= CSR6_TR_160 | CSR6_MDR;
	}

	pdk_reg_put32(ppa, CSR12, CSR12_140_GPC | 0);	/* All GEP ports inputs */
	pdk_reg_put32(ppa, CSR6, lp->OperationModeBase);
	pdk_time_delay(5);

	return(1);
}

/*
 * Current timeout is 1/2 second.
 * If changing, change all three values at same time (and the comment)
 */

#define H10_100_MSEC_DELAY 2442		/* Actually 500 msecs */
#define H100_100_MSEC_DELAY 2442	/* Actually 200 msecs */
#define H10_HALF_SECOND_DELAY 610
#define H100_HALF_SECOND_DELAY 6104

#define ZXE_WAKEUP_INTERVAL 500

/* Timeouts in Milliseconds */
#define ZXE_TX_TIMEOUT_LIMIT 10000
#define ZXE_RX_TIMEOUT_LIMIT 10000
/*
#define ZXE_NWAY_TIMEOUT 3000
#define ZXE_LINK_SETTLE 3000
*/
#define ZXE_NWAY_TIMEOUT 1000
#define ZXE_LINK_SETTLE 1000
#define ZXE_NWAY_MSECOND 5	/* Timer iterations in 10BASE-T mode for 1024 usecs*/

int32_t
ZxeConfigure21143Media(ppa)
int32_t ppa;
{
	local_adapter_t *lp;
	uint32_t	csr13, csr14, csr15, csr15gpd;
#ifdef DYNAMIC_THRESHOLD
	uint32_t	TxFifoThreshold;
#endif

	lp = local_adapters[ppa];

	lp->OperationModeBase = CSR6_SC|CSR6_BIT25|CSR6_CA;
	if(lp->PhyType == DEVICE_BCM5208) {
		lp->OperationModeBase |= CSR6_PS|CSR6_HBD;
		csr15gpd = CSR15_142_ABM;
	} else {
		lp->OperationModeBase |= CSR6_CS|CSR6_SCR;
		csr15gpd = 0x40000 | CSR15_142_ABM;	/* Symbol PHY uses GEP2 low as reset */
	}

	lp->LinkSpeed = 10;

	switch(lp->CurrentMediaType) {

		default:
			lp->CurrentMediaType = ZXE_MEDIA_AUTO;

		case ZXE_MEDIA_AUTO:
			csr13 = 0x0001;
			csr14 = 0x37FFF;
			lp->OperationModeBase |= CSR6_FD;
			break;

		case ZXE_MEDIA_TP:
			csr13 = 0x0001;
			csr14 = 0x7F3F;
			break;

		case ZXE_MEDIA_TPFD:
			csr13 = 0x0001;
			csr14 = 0x7F3D;
			lp->OperationModeBase |= CSR6_FD;
			break;

		case ZXE_MEDIA_TXFD:
			lp->OperationModeBase |= CSR6_FD;
			/* Fall Through */

		case ZXE_MEDIA_TX:
			csr13 = 0x0000;
			csr14 = 0x0000;
			lp->OperationModeBase |= CSR6_PS|CSR6_HBD;
			lp->LinkSpeed = 100;
			break;

#ifdef LOOPBACK_21143
		case ZXE_LOOPBACK_TPINT:
			csr13 = 0x0009;
			csr14 = 0x0000;
			csr15gpd |= 0x19;		/* AUI */
			lp->OperationModeBase |= CSR6_OM_INTLB;	/* 10BASE-T MAC Loopback */
			break;

		case ZXE_LOOPBACK_TXINT:
			csr13 = 0x0000;
			csr14 = 0x0000;
			csr15gpd |= CSR15_142_ABM;		/* AUI */
			lp->OperationModeBase |= CSR6_PS|CSR6_OM_INTLB;	/*100TX MAC loopback*/
			lp->LinkSpeed = 100;
			break;

		case ZXE_LOOPBACK_TXEXT:
			csr13 = 0x0000;
			csr14 = 0x0000;
			csr15gpd |= 0x10000|CSR15_142_ABM;		/* Set PHY LOOPBACK */
			lp->OperationModeBase |= CSR6_PS|CSR6_OM_EXTLB;	/*100TX PHY loopback*/
			lp->LinkSpeed = 100;
			break;
#endif

	}

	if (lp->LinkSpeed > 10) {
#ifdef DYNAMIC_THRESHOLD
		TxFifoThreshold = (lp->TxFifoThreshold+99)/100;
		if (TxFifoThreshold < NUM_TXFIFOCONFIG)
			lp->OperationModeBase |=
				TxFifoConfigurations[TxFifoThreshold];
		else
			lp->OperationModeBase |= CSR6_TFF_1024 | CSR6_FP;
#else
		lp->OperationModeBase |= CSR6_FP;
#endif
	}
	else {
		lp->OperationModeBase |= CSR6_TR_160 | CSR6_MDR;
	}

	pdk_reg_put32(ppa, CSR13, 0);

	pdk_time_delay(1);
	if (lp->MediaConnectState == MEDIA_CONNECTED)
		csr15gpd |= 0x80000; /* Link LED */

	pdk_reg_put32(ppa, CSR15, csr15gpd);
	pdk_reg_put32(ppa, CSR14, csr14);
	pdk_reg_put32(ppa, CSR13, csr13);
	pdk_reg_put32(ppa, CSR6, lp->OperationModeBase);
	pdk_time_delay(5);

	return(1);
}

int32_t
ZxeConfigureMedia(ppa)
int32_t ppa;
{
	local_adapter_t *lp;
	int32_t  MediaConfigured;

	lp = local_adapters[ppa];

	if (lp->Info.NicType == DEVICE_DEC21140) {
		MediaConfigured = ZxeConfigure21140Media(ppa);
		if (MediaConfigured) {
			if ((lp->LinkSpeed == 100) || (lp->CurrentMediaType == ZXE_MEDIA_AUTO))
				lp->PollTimeout = H100_HALF_SECOND_DELAY;
			else {
				lp->PollTimeout = H10_HALF_SECOND_DELAY;
				pdk_reg_put32(ppa, CSR11, lp->PollTimeout);
			}
		}
	}
	else if (lp->Info.NicType == DEVICE_DEC21143) {
		MediaConfigured = ZxeConfigure21143Media(ppa);
	}
	else {
		MediaConfigured = 0;
	}

	return(MediaConfigured);
}

char	*MediaNames[] = {
	"Link Down", "10Mbps Half-Duplex", "10Mbps BNC", "10Mbps AUI",
	"10Mbps Full-Duplex", "100Mbps Half-Duplex", "100Mbps Full-Duplex",
	"100Mbps T4", "Link Auto-Negotiating" };

int32_t
ZxeCheckNWayStatus(ppa,isr_value)
int32_t ppa;
int32_t isr_value;
{
	local_adapter_t	*lp;
	int32_t 		MediaChanged = 0;
	uint32_t		Csr15;
	uint16_t		Status, PhyAnlpar;
	int32_t			Index;
	uint32_t		NewMediaType;
	int32_t			i = 0;

	lp = local_adapters[ppa];

	NewMediaType = lp->CurrentMediaType;

	if ((lp->PhyType == DEVICE_DP83840) || (lp->PhyType == DEVICE_DP83843)) {
		uint16_t PhyPar;
		/* Check ANLPAR and PAR registers */
		ZxeReadPHY(ppa, DP83840_ANLPAR_REG, &PhyAnlpar);
		ZxeReadPHY(ppa, DP83840_PAR_REG, &PhyPar);
		ZxeReadPHY(ppa, DP83840_BMSR_REG, &Status);

		if (Status & DP83840_BMSR_RFLT) {
			NewMediaType = ZXE_MEDIA_AUTO;
		}
		else if (((Status & DP83840_BMSR_LNKST) == 0) && (NewMediaType != ZXE_MEDIA_AUTO))
			NewMediaType = ZXE_MEDIA_AUTO;
		else if (Status & DP83840_BMSR_ANCPL) {
			if (PhyAnlpar & DP83840_ANLPAR_TXFD)
				NewMediaType = ZXE_MEDIA_TXFD;
			else if (PhyAnlpar & DP83840_ANLPAR_TX)
				NewMediaType = ZXE_MEDIA_TX;
			else if (PhyAnlpar & DP83840_ANLPAR_TPFD)
				NewMediaType = ZXE_MEDIA_TPFD;
			else if (PhyAnlpar & DP83840_ANLPAR_TP)
				NewMediaType = ZXE_MEDIA_TP;
			else if (PhyPar & DP83840_PAR_SPEED10) {
				NewMediaType = ZXE_MEDIA_TP;
				DEBUG1("SPEED10\n");
			}
			else
				NewMediaType = ZXE_MEDIA_AUTO;
		}

		DEBUG4("ZNB(%d): NewMediaType %s ANLPAR %x Status %x - Poll %d\n", ppa, MediaNames[NewMediaType], PhyAnlpar, Status, lp->PollTimeout);
		if (NewMediaType == ZXE_MEDIA_AUTO) {
			DEBUG4("ZNB(%d): MEDIA_AUTO - ANLPAR %x Status %x\n", ppa, PhyAnlpar, Status);
			if (lp->MediaConnectState == MEDIA_CONNECTED) {
				lp->MediaConnectState = MEDIA_DISCONNECTED;
				MediaChanged = 1;
			}
		}
		else {
			lp->MediaConnectState = MEDIA_CONNECTED;

			if (NewMediaType != lp->CurrentMediaType) {
				DEBUG2("ZNB(%d): MediaChanged - %s to %s ANLPAR %x Status %x\n", ppa, MediaNames[lp->CurrentMediaType], MediaNames[NewMediaType], PhyAnlpar, Status);
				MediaChanged = 1;
			}
		}

	}
	else if (lp->PhyType == DEVICE_ICS1890) {
		/* Check ANLPAR and PAR registers */
		ZxeReadPHY(ppa, ICS1890_ANLPAR_REG, &PhyAnlpar);
		ZxeReadPHY(ppa, ICS1890_MIISR_REG, &Status);

		/* Determine media type */
		if (Status & ICS1890_MIISR_RFLT) {
			DEBUG3("ZxeCheckNWayStatus(%d): ANLPAR=0x%04X, STATUS=0x%04X\n", ppa, PhyAnlpar, Status);
			NewMediaType = ZXE_MEDIA_AUTO;
		}
		else if (Status & ICS1890_MIISR_ANCPL) {
			if ((Status & ICS1890_MIISR_LNKST) == 0) {
				/* Clr latched status */
				ZxeReadPHY(ppa, ICS1890_MIISR_REG, &Status);
			}
			if ((Status & ICS1890_MIISR_LNKST) == 0) {
				DEBUG2("ZxeCheckNWayStatus(%d): ANLPAR=0x%04X, STATUS=0x%04X\n", ppa, PhyAnlpar, Status);
				NewMediaType = ZXE_MEDIA_AUTO;
				lp->PollTimeout = 1250;  /* About 100 ms */
			}
			else if (PhyAnlpar & ICS1890_ANLPAR_TXFD)
				NewMediaType = ZXE_MEDIA_TXFD;
			else if (PhyAnlpar & ICS1890_ANLPAR_TX)
				NewMediaType = ZXE_MEDIA_TX;
			else if (PhyAnlpar & ICS1890_ANLPAR_TPFD)
				NewMediaType = ZXE_MEDIA_TPFD;
			else if (PhyAnlpar & ICS1890_ANLPAR_TP)
				NewMediaType = ZXE_MEDIA_TP;
			else
				NewMediaType = ZXE_MEDIA_AUTO;
		}
		else
			NewMediaType = ZXE_MEDIA_AUTO;

		DEBUG3("ZxeCheckNWayStatus(%d): ANLPAR=0x%04X, STATUS=0x%04X - Media %d --> %d link %d\n", ppa, PhyAnlpar, Status, lp->CurrentMediaType, NewMediaType, lp->LinkSpeed);
		if (NewMediaType != lp->CurrentMediaType) {
			if (NewMediaType == ZXE_MEDIA_AUTO)
				lp->MediaConnectState = MEDIA_DISCONNECTED;
			else
				lp->MediaConnectState = MEDIA_CONNECTED;

			MediaChanged = 1;
		}
	}
	else if (lp->PhyType == DEVICE_BCM5208) {
		lp->PollTimeout = 0;

		/* Check ANLPAR and PAR registers */
		ZxeReadPHY(ppa, BCM5208_ANLPAR_REG, &PhyAnlpar);
		ZxeReadPHY(ppa, BCM5208_MIISR_REG, &Status);

		/* Determine media type */
		if (Status & BCM5208_MIISR_RFLT) {
			DEBUG3("ZxeCheckNWayStatus(%d): ANLPAR=0x%04X, STATUS=0x%04X\n", ppa, PhyAnlpar, Status);
			NewMediaType = ZXE_MEDIA_AUTO;
		}
		else if (Status & BCM5208_MIISR_ANCPL) {
			if ((Status & BCM5208_MIISR_LNKST) == 0) {
				/* Clr latched status */
				ZxeReadPHY(ppa, BCM5208_MIISR_REG, &Status);
			}
			if ((Status & BCM5208_MIISR_LNKST) == 0) {
				DEBUG2("ZxeCheckNWayStatus(%d): ANLPAR=0x%04X, STATUS=0x%04X\n", ppa, PhyAnlpar, Status);
				NewMediaType = ZXE_MEDIA_AUTO;
				lp->PollTimeout = 1250;  /* About 100 ms */
			}
			else if (PhyAnlpar & BCM5208_ANLPAR_TXFD)
				NewMediaType = ZXE_MEDIA_TXFD;
			else if (PhyAnlpar & BCM5208_ANLPAR_TX)
				NewMediaType = ZXE_MEDIA_TX;
			else if (PhyAnlpar & BCM5208_ANLPAR_TPFD)
				NewMediaType = ZXE_MEDIA_TPFD;
			else if (PhyAnlpar & BCM5208_ANLPAR_TP)
				NewMediaType = ZXE_MEDIA_TP;
			else if (Status & BCM5208_MIISR_LNKST)	
				NewMediaType = ZXE_MEDIA_TX;
			else
				NewMediaType = ZXE_MEDIA_AUTO;

		}
		else
			NewMediaType = ZXE_MEDIA_AUTO;

		DEBUG3("ZxeCheckNWayStatus(%d): ANLPAR=0x%04X, STATUS=0x%04X - Media %d --> %d link %d\n", ppa, PhyAnlpar, Status, lp->CurrentMediaType, NewMediaType, lp->LinkSpeed);
		if (NewMediaType != lp->CurrentMediaType) {
			if (NewMediaType == ZXE_MEDIA_AUTO)
				lp->MediaConnectState = MEDIA_DISCONNECTED;
			else
				lp->MediaConnectState = MEDIA_CONNECTED;

			MediaChanged = 1;
		}
	}
	else if (lp->PhyType == DEVICE_QS6611) {
		uint32_t bittest = CSR12_143_LS100;
		uint32_t Csr12 = pdk_reg_get32(ppa, CSR12);

		DEBUG4("ZxeCheckNWayStatus(%d):  CurrentMediaType 0x%x, csr12 0x%x\n",
			ppa, lp->CurrentMediaType,Csr12);

		switch(lp->CurrentMediaType) {
			case ZXE_MEDIA_AUTO: {
				if ((Csr12 & CSR12_142_ANMSK) != CSR12_142_ANCPL) {
					DEBUG3("ZxeCheckNWayStatus(%d): ANCPL not reached: csr12 0x%0x\n", ppa,Csr12);
					break;
				}
				/* Autonegotiation */
				DEBUG3("ZxeCheckNWayStatus(%d): ANCPL: csr12 0x%x\n", ppa,Csr12);
				if (Csr12 & CSR12_142_LPN) {	/* Link Partner AutoNegotiates */
					if (Csr12 & CSR12_142_TXFD)
						lp->AutoNegotiatedMediaType = ZXE_MEDIA_TXFD;
					else if (Csr12 & CSR12_142_TXHD)
						lp->AutoNegotiatedMediaType = ZXE_MEDIA_TX;
					else {
						bittest = CSR12_142_LS10;
						if (Csr12 & CSR12_142_TPFD)
							lp->AutoNegotiatedMediaType = ZXE_MEDIA_TPFD;
						else if (Csr12 & CSR12_142_TPHD)
							lp->AutoNegotiatedMediaType = ZXE_MEDIA_TP;
						else
							break;
					}

					Csr12 = pdk_reg_get32(ppa, CSR12);
					if (Csr12 & bittest) {
						/* Link status(HCD) not good, check for up to 1 sec. */
						lp->PollTimeout = ZXE_NWAY_MSECOND<<3;  /* Initial interval */
						lp->CurrentMediaType=NewMediaType = ZXE_MEDIA_AUTO_COMPLETE;
						if (bittest == CSR12_143_LS100) {
							lp->OperationModeBase |= CSR6_PS|CSR6_HBD;
							lp->OperationMode |= CSR6_PS|CSR6_HBD;
							pdk_reg_put32(ppa, CSR6, lp->OperationMode);
						}
					} else {
						lp->MediaConnectState = MEDIA_CONNECTED;
						NewMediaType = lp->AutoNegotiatedMediaType;
						lp->PollTimeout = (bittest == CSR12_143_LS100)?
							H100_100_MSEC_DELAY:H10_100_MSEC_DELAY;
					}
					break;
				}

				/* Autosensing */

				DEBUG3("Autosensing - csr12 0x%x\n", Csr12);

				lp->MediaConnectState = MEDIA_CONNECTED;
				lp->PollTimeout = (bittest == CSR12_143_LS100)?
					H100_100_MSEC_DELAY:H10_100_MSEC_DELAY;

				if (!(Csr12 & CSR12_143_LS100)) {
					/* 100 Mbps Link Status OK */
					NewMediaType = ZXE_MEDIA_TX;
				} else if (!(Csr12 & CSR12_142_LS10)) {
					/* 10 Mbps Link Status OK */
					NewMediaType = ZXE_MEDIA_TP;
				} else {
					lp->MediaConnectState = MEDIA_DISCONNECTED;
					lp->PollTimeout = 0;
				}
				break;
			}

			case ZXE_MEDIA_AUTO_COMPLETE: {

				switch(lp->AutoNegotiatedMediaType) {
					case ZXE_MEDIA_TPFD:
					case ZXE_MEDIA_TP:
						bittest = CSR12_142_LS10;
					case ZXE_MEDIA_TXFD:
					case ZXE_MEDIA_TX:
				    	if ((Csr12 & bittest) == 0) {
							pdk_reg_put32(ppa, CSR11, 0);	/* Stop Timer */
							DEBUG2("ZxeCheckNWayStatus(%d): Link up < %d msecs\n",
								ppa,lp->PollTimeout*2/ZXE_NWAY_MSECOND-1);
							NewMediaType = lp->AutoNegotiatedMediaType;
							lp->MediaConnectState = MEDIA_CONNECTED;
							lp->PollTimeout = (bittest == CSR12_143_LS100)?
								H100_100_MSEC_DELAY:H10_100_MSEC_DELAY;
						} else if (isr_value & CSR5_GTE) {
							if (lp->PollTimeout > ZXE_NWAY_MSECOND<<8) {
								DEBUG2("ZxeCheckNWayStatus(%d): LINK FAILED, csr12=0x%x\n",ppa,Csr12);
							if (bittest == CSR12_143_LS100) {
								lp->OperationModeBase &= ~(CSR6_PS|CSR6_HBD);
								lp->OperationMode &= ~(CSR6_PS|CSR6_HBD);
								pdk_reg_put32(ppa, CSR6, lp->OperationMode);
						}
								lp->PollTimeout = 0;
								lp->CurrentMediaType = ZXE_MEDIA_AUTO;
								NewMediaType = ZXE_MEDIA_AUTO;
								pdk_reg_put32(ppa, CSR12, 0x1000);	/* Restart AN */
							} else {
								lp->PollTimeout <<= 1;	/* Double timeout */
							}
						}
				}
				break;
			}
			case ZXE_MEDIA_TP:
			case ZXE_MEDIA_TPFD:
				bittest = CSR12_142_LS10;

			case ZXE_MEDIA_TX:
			case ZXE_MEDIA_TXFD: {

				/* Check for settle timer set and not timer expired */
				if (lp->PollTimeout && !(isr_value & CSR5_GTE)) {
					DEBUG3("ZxeCheckNWayStatus(%d): Media %s let link settle\n",
						ppa, MediaNames[lp->CurrentMediaType]);
					break;		/* Ignore Link Change during settle time */
				} else if (Csr12 & bittest) {
					/* Link Status down */
					DEBUG3("ZxeCheckNWayStatus(%d): Media %s - link down\n",
						ppa, MediaNames[lp->CurrentMediaType]);
					NewMediaType = ZXE_MEDIA_AUTO;
					lp->MediaConnectState = MEDIA_DISCONNECTED;
				}
				lp->PollTimeout = 0;
				break;
			}
		} /* End switch(lp->CurrentMediaType) */

		if (NewMediaType != lp->CurrentMediaType) {
			MediaChanged = 1;
		}
	}

	if (MediaChanged) {
		DEBUG3("ZxeCheckNWayStatus(%d): NewMediaType 0x%x\n", ppa, NewMediaType);
		lp->CurrentMediaType = NewMediaType;
	}

	return(MediaChanged);
}

void
ZxeCheckLinkStatus(ppa,isr_value)
int32_t ppa;
int32_t isr_value;
{
	local_adapter_t *lp;
	uint32_t csr12;
	uint16_t Status;

	lp = local_adapters[ppa];

	switch(lp->PhyType) {
		case DEVICE_QS6611:
			csr12 = pdk_reg_get32(ppa, CSR12);			
			DEBUG2("ZxeCheckLinkStatus(%d): csr12 = 0x%x\n", ppa, csr12);

			switch(lp->ConfiguredMediaType) {
				case ZXE_MEDIA_TXFD:
				case ZXE_MEDIA_TX:
					lp->MediaConnectState = (!(csr12 & CSR12_143_LS100))?
						MEDIA_CONNECTED : MEDIA_DISCONNECTED;
					break;

				case ZXE_MEDIA_TPFD:
				case ZXE_MEDIA_TP:
					lp->MediaConnectState = (!(csr12 & CSR12_142_LS10))?
						MEDIA_CONNECTED : MEDIA_DISCONNECTED;
					break;

				default:
					return;
			}

			break;

		case DEVICE_BCM5208:
			ZxeReadPHY(ppa, BCM5208_MIISR_REG, &Status); /* Clear latched */
			ZxeReadPHY(ppa, BCM5208_MIISR_REG, &Status);	/* Get current */
			DEBUG4("ZxeCheckLinkStatus(%d): PhyStatus = 0x%x\n", ppa, Status);

			if (Status & BCM5208_MIISR_RFLT) {
				DEBUG2("ZxeCheckLinkStatus(%d): STATUS=0x%04X - FEF \n", ppa, Status);
			}

			lp->MediaConnectState = (Status & BCM5208_MIISR_LNKST) ?
					MEDIA_CONNECTED : MEDIA_DISCONNECTED;

			break;


		case DEVICE_ICS1890:
		case DEVICE_DP83840:
			ZxeReadPHY(ppa, DP83840_BMSR_REG, &Status);
			if (Status & DP83840_BMSR_LNKST)
				lp->MediaConnectState = MEDIA_CONNECTED;
			else
				lp->MediaConnectState = MEDIA_DISCONNECTED;

			break;

		default:
			DEBUG1("ZxeCheckLinkStatus(%d): Unknown phy type 0x%x\n", ppa, lp->PhyType);
			break;
	}

	return;
}


int32_t
ZxeReconfigureMediaType(ppa)
int32_t ppa;
{
	local_adapter_t	*lp;
	uint32_t	Csr7;

	/* Stop TX and RX */

	lp = local_adapters[ppa];
	Csr7 = pdk_reg_get32(ppa, CSR7) & ~0xF3FE0000;
	SET_CSR7(ppa,0);

	if (!ZxeStopTxAndRx(ppa)) {

		/* Setup new media type */
		ZxeConfigureMedia(ppa);
		ZxeStartTxAndRx(ppa);
		SET_CSR7(ppa,Csr7);
		return(0);

	} else	/* Could not stop device, need reset */
		return(1);
}


/* This function is called twice a second or when a link status interrupt
   is received to configure the media */

void
ZxeMediaPoll(ppa,isr_value)
int32_t ppa;
uint32_t isr_value;
{
	local_adapter_t	*lp;
	int_t 	DoReset = 0;
	int_t 	TransmitTimeout = 0;
	uint32_t		MediaConnectState;


	lp = local_adapters[ppa];

	MediaConnectState = lp->MediaConnectState;
	DEBUG4("ZxeMediaPoll(%d):  MediaConnectState 0x%x\n", ppa, MediaConnectState);

	if (lp->free_tx_descriptor_count != num_transmit_desc) {
		pdk_lock_acquire(lp->tx_descriptor_lock);
		lp->TxTimeoutCounter += ZXE_WAKEUP_INTERVAL;
		if (lp->TxTimeoutCounter > ZXE_TX_TIMEOUT_LIMIT) {
			TransmitTimeout = 1;
			DEBUG1("ZxeMediaPoll(%d): Transmit Counter Reset\n", ppa);
			DoReset = 1;
		}
		pdk_lock_release(lp->tx_descriptor_lock);
	}

	if (!(lp->rx_dequeue_here->RDES0 & RDES0_OWN)) {
		lp->RxTimeoutCounter += ZXE_WAKEUP_INTERVAL;
		if (lp->RxTimeoutCounter > ZXE_RX_TIMEOUT_LIMIT) {
			DEBUG1("ZxeMediaPoll(%d):  Receive Counter Reset\n", ppa);
			DoReset = 1;
		}
	}

	if (lp->ConfiguredMediaType == ZXE_MEDIA_AUTO) {
		if (ZxeCheckNWayStatus(ppa,isr_value)) {
			if (ZxeReconfigureMediaType(ppa)) {
				DEBUG1("ZxeMediaPoll(%d):  Reconfigure Reset\n", ppa);
				DoReset = 1;
			}
		}
	}
	else
		ZxeCheckLinkStatus(ppa,isr_value);

	if (DoReset) {
		/* Use old ConnectState.  pdk_hal_restart via pdk_hal_stop
		 * will set to DISCONNECTED and appropriatly notify the upper
		 * layer
		 */
		lp->MediaConnectState = MediaConnectState;
		pdk_hal_restart(ppa);
	}
	else if (MediaConnectState != lp->MediaConnectState) {
		uint32_t csr15gpd;
		if (lp->Info.NicType == DEVICE_DEC21143)
			csr15gpd = pdk_reg_get32(ppa, CSR15) & ~0xFFFA0000;		
		if (lp->MediaConnectState == MEDIA_CONNECTED) {
			DEBUG1("ZNB(%d): %s\n", ppa, MediaNames[lp->CurrentMediaType]);
			zxe_external_fault_led(ppa, 0, 0);
			ZxeStartTxAndRx(ppa);
			(*lp->layer_routines.pdk_status)(ppa, PDK_STATUS_LINK_UP, 0);
			csr15gpd |= 0x80000;
		}
		else {
			DEBUG1("ZNB(%d): %s\n", ppa, MediaNames[0]);
			(*lp->layer_routines.pdk_status)(ppa, PDK_STATUS_LINK_DOWN, 0);
			if (!zxe_test_fault(ppa))	/* Check for internal fault */
				zxe_external_fault_led(ppa, 1, 0);
		}
		if (lp->Info.NicType == DEVICE_DEC21143)
			pdk_reg_put32(ppa, CSR15, csr15gpd);
	}

	/* load timer */
	if (lp->PollTimeout)
		pdk_reg_put32(ppa, CSR11, lp->PollTimeout);
}

void
init_setup_buf(ppa)
int32_t ppa;
{
	local_adapter_t	*lp;
	uchar		*cp;
	int32_t		i, j;

	if ((lp = local_adapters[ppa]) == 0) {
		return;
	}

	/* Preset entire setup buffer to our network address */
	cp = lp->setup_buf;
	for (i=0; i<(MAX_SETUP_BUF_ADDRS); i++) {
		for (j=0; j<3; j++) {
			*cp++ = lp->CurrentNetworkAddress[j*2];
			*cp++ = lp->CurrentNetworkAddress[j*2+1];
			*cp++ = 0;
			*cp++ = 0;
		}
	}

	/* Put in multicast addresses */
	cp = lp->setup_buf;
	for (i=0; i<lp->MCAddressCount; i++) {
		for (j=0; j<3; j++) {
			*cp++ = lp->MCAddresses[i][j*2];
			*cp++ = lp->MCAddresses[i][j*2+1];
			*cp++ = 0;
			*cp++ = 0;
		}
	}
	/* Add the broadcast address */
	for (j=0; j<3; j++) {
		*cp++ = 0xff;
		*cp++ = 0xff;
		*cp++ = 0;
		*cp++ = 0;
	}
}

int32_t
add_multicast_address(ppa, addr)
int32_t ppa;
uchar *addr;
{
	int32_t i;
	local_adapter_t *lp;
	uint32_t err;


	if ((lp = local_adapters[ppa]) == 0) {
		return(PDK_OUT_STATE);
	}

	if (lp->tx_descriptor_lock) {
		pdk_lock_acquire(lp->tx_descriptor_lock);
	}

	for (i=0; i<lp->MCAddressCount; i++) {
		if (pdk_bcmp(lp->MCAddresses[i], addr, PDK_ETHERADDRL) == 0) {
			if (lp->tx_descriptor_lock) {
				pdk_lock_release(lp->tx_descriptor_lock);
			}

			return(0);
		}
	}

	pdk_bcopy(addr, lp->MCAddresses[lp->MCAddressCount], PDK_ETHERADDRL);
	lp->MCAddressCount++;

	err = 0;
	if (lp->state == LOCAL_ADAPTER_RUNNING) {
		init_setup_buf(ppa);
		err = load_setup_buf(ppa);
	}

	if (lp->tx_descriptor_lock) {
		pdk_lock_release(lp->tx_descriptor_lock);
	}

	return(err);
}

int32_t
del_multicast_address(ppa, addr)
int32_t ppa;
uchar *addr;
{
	int32_t i;
	int32_t		err;
	local_adapter_t *lp;

	if ((lp = local_adapters[ppa]) == 0) {
		return(PDK_OUT_STATE);
	}

	if (lp->MCAddressCount == 0)
		return(PDK_NOT_FOUND);

	if (lp->tx_descriptor_lock) {
		pdk_lock_acquire(lp->tx_descriptor_lock);
	}

	for (i=0; i<lp->MCAddressCount; i++) {
		if (pdk_bcmp(lp->MCAddresses[i], addr, PDK_ETHERADDRL) == 0) {
			lp->MCAddressCount--;
			/* Copy the last over the entry being removed */
			pdk_bcopy(lp->MCAddresses[lp->MCAddressCount],
				lp->MCAddresses[i], PDK_ETHERADDRL);
			if (lp->state == LOCAL_ADAPTER_RUNNING) {
				init_setup_buf(ppa);
				err = load_setup_buf(ppa);
			}
			if (lp->tx_descriptor_lock) {
				pdk_lock_release(lp->tx_descriptor_lock);
			}
			return(err);
		}
	}

	if (lp->tx_descriptor_lock) {
		pdk_lock_release(lp->tx_descriptor_lock);
	}

	return(PDK_NOT_FOUND);
}

/* Must hold transmit lock when calling */
uint32_t
load_setup_buf(ppa)
int32_t ppa;
{
	local_adapter_t	*lp;
	pdk_buf_t	*bp;
	pdk_map_t	*mp;
	desc_t		*dd;
	desc_t		*zero_dd;
	uint32_t	length;
	uint32_t	paddr;
	uchar QNX_FAR	*rptr;
	int32_t		free_txds;
	int32_t		err;


	if ((lp = local_adapters[ppa]) == 0) {
		DEBUG1("load_setup_buf(%d):  Out of state\n", ppa);
		return(PDK_OUT_STATE);
	}

	process_xmit_complete(lp);	/* Already hold lock */

	if ((bp = pdk_buf_alloc(ppa, SETUP_FRAME_SIZE, 0)) == 0) {
		DEBUG1("load_setup_buf(%d):  pdk_buf_alloc failed\n", ppa);
		return(PDK_SYS_ERR);
	}

	pdk_buf_set_len(bp, SETUP_FRAME_SIZE);

	/* Always leave one */
	if ((free_txds = lp->free_tx_descriptor_count) <= 2) {
		DEBUG1("load_setup_buf(%d):  Transmit ring full\n", ppa);
		pdk_buf_free(bp);
		return(PDK_NO_ROOM);
	}

	free_txds -= 2;

	rptr = pdk_buf_rptr(bp);

	/* copy in setup buf */
#ifdef QNX4
	_fmemcpy(rptr, MK_FP(my_ds(), lp->setup_buf), SETUP_FRAME_SIZE);
#else
	pdk_bcopy(lp->setup_buf, rptr, SETUP_FRAME_SIZE);
#endif

	/* Map this pdk transmit buffer */
	if ((mp = pdk_buf_map(bp, PDK_BUF_XMIT, &err)) == 0) {
		DEBUG1("load_setup_buf(%d):  pdk_buf_map failed\n", ppa);
		pdk_buf_free(bp);
		return(err);
	}

	mp = pdk_buf_get_pmap(mp, &paddr, &length, &err);
	if (err) {
		DEBUG1("load_setup_buf(%d):  pdk_buf_get_pmap failed\n", ppa);
		pdk_buf_unmap(bp);
		pdk_buf_free(bp);
		return(err);
	}

	/* create zero length descriptor */
	zero_dd = lp->tx_enqueue_here;
	zero_dd->TDES1 = TDES1_IC | TDES1_SET;
	if (zero_dd->NEXT == lp->tx_queue)
		zero_dd->TDES1 |= TDES1_TER;

	zero_dd->BP = 0;

	/* fill in descriptor */
	dd = zero_dd->NEXT;

	dd->TDES2 = pdk_htobl(paddr);
	dd->TDES1 = pdk_htobl(SETUP_FRAME_SIZE);
	dd->TDES1 |= (TDES1_IC | TDES1_SET);
	if (dd->NEXT == lp->tx_queue)
		dd->TDES1 |= TDES1_TER;
	dd->TDES0 = TDES0_OWN;

	dd->BP = bp;

	/* Update Tx descriptor list... */
	lp->tx_enqueue_here = dd->NEXT;
	lp->tx_enqueue_here->TDES0 = 0;

	if ((lp->free_tx_descriptor_count = free_txds) < lp->tx_lwm)
		lp->tx_lwm = free_txds;

	/* guarantee in order... */
	SYS_INSTR_SYNC();

	zero_dd->TDES0 = TDES0_OWN;

	descriptor_flush(lp, H_WRITE);

	return(err);
}


/* In microseconds */
#define LOOPBACK_DELAY_INTERVAL 50
#define LOOPBACK_DELAY 10000

pdk_buf_t *
zxe_loopback_ringget(handlep, ddp, setup_buf)
pdk_desc_handle_t *handlep;
desc_t **ddp;
int32_t setup_buf;
{
	pdk_buf_t *bp;
	int32_t i;

	for (i=0;i<LOOPBACK_DELAY;i+=LOOPBACK_DELAY_INTERVAL) {
		SYS_DESC_SYNC(handlep, *ddp, *ddp+1, PDK_DESC_SYNC_FORCPU);

		if (((*ddp)->TDES0 & TDES0_OWN) == 0) {
			bp = (*ddp)->BP;

			if (setup_buf) {
				if (bp != 0) 
					return(0);

				*ddp = (*ddp)->NEXT;

				return(zxe_loopback_ringget(handlep, ddp, 0));
			}

			(*ddp)->BP = 0;
			*ddp = (*ddp)->NEXT;

			return(bp);
		}

		pdk_time_delay(LOOPBACK_DELAY_INTERVAL);
	}

	return(0);
}

desc_t *
zxe_loopback_ringput(lp, ddp, bp, flag, qd)
local_adapter_t *lp;
desc_t **ddp;
pdk_buf_t *bp;
int32_t flag;
desc_t *qd;
{
	pdk_map_t	*mp;
	uint32_t	length;
	uint32_t	paddr;
	int32_t		err;
	int_t		i;

	if ((mp = pdk_buf_map(bp, flag, &err)) == 0) {
		return(0);
	}

	mp = pdk_buf_get_pmap(mp, &paddr, &length, &err);
	if (err) {
		pdk_buf_unmap(bp);
		return(0);
	}

	(*ddp)->BP = bp;

	(*ddp)->RDES2 = pdk_htobl(paddr);
	(*ddp)->RDES1 = pdk_htobl(length);

	if ((*ddp)->NEXT == qd)
		(*ddp)->RDES1 |= RDES1_RER;

	if (flag == PDK_BUF_XMIT)
		(*ddp)->TDES1 |= TDES1_FS | TDES1_LS;

	/* guarantee in order... */
	SYS_INSTR_SYNC();

	(*ddp)->RDES0 = RDES0_OWN;

	(*ddp) = (*ddp)->NEXT;

	descriptor_flush(lp, H_WRITE);
	descriptor_flush(lp, H_READ);

	return(*ddp);
}

void
zxe_loopback_data(uchar *buff,int32_t len)
{
	uint32_t i,j;
	uchar val;
	while (len > 0) {
		for (j=0;(j<8) && (len>0);j++) {
			val = 1<<j;
			for (i=0;(i<8) && (len>0);i++,len--) {
				*buff++ = val;
				if (val & 0x80)
					val = 1;
				else
					val = val<<1;
			}
			val = ~(1<<j);
			for (i=0;(i<8) && (len>0);i++,len--) {
				*buff++ = val;
				if (val & 0x80)
					val = (val<<1)|1;
				else
					val = 0xfe;
			}
		}
	}
}

int32_t
zxe_loopback_test(ppa)
int32_t ppa;
{
	local_adapter_t *lp;
	pdk_buf_t	 *tx_bp, *allocated_rx_bp, *rx_bp;
	pdk_ether_header_t  *ethp;
	uint32_t	len;
	uint32_t	orig_state;
	uint32_t	orig_csr7;
	uint32_t	orig_media;
	uchar		*rptr;
	int32_t		ret_val;

	ret_val = 0;
	tx_bp = 0;
	allocated_rx_bp = 0;
	rx_bp = 0;

	lp = local_adapters[ppa];
	if (lp == 0) {
		DEBUG1("zxe_loopback_test(%d):  Out of state\n", ppa);
		return(0);
	}

#ifdef LOOPBACK_21143
	if (lp->PhyType != DEVICE_BCM5208 && lp->Info.NicType != DEVICE_DEC21143)
#else
	if (lp->PhyType != DEVICE_BCM5208)
#endif
	{
		DEBUG3("zxe_loopback_test(%d): loopback not supported\n", ppa);
		return(0);
	}

	orig_state = lp->state;
	orig_csr7 = pdk_reg_get32(ppa, CSR7) & ~0xF3FE0000;
	orig_media = lp->CurrentMediaType;

	if (orig_state == LOCAL_ADAPTER_RUNNING) {
		SET_CSR7(ppa,0);
		lp->state = LOCAL_ADAPTER_STOPPED;
		if (lp->MediaConnectState == MEDIA_CONNECTED) {
			(*lp->layer_routines.pdk_status)(ppa, PDK_STATUS_LINK_DOWN, 0);
		}
	}

	lp->MediaConnectState = MEDIA_DISCONNECTED;

	/* WAIT FOR SETUP BUFFER TO COMPLETE if one exists */
	if (lp->tx_dequeue_here == lp->tx_enqueue_here) {
		DEBUG3("zxe_loopback_test(%d): Allocating transmit buffer\n", ppa);
		tx_bp = pdk_buf_alloc(ppa, SETUP_FRAME_SIZE, 0);
		if (tx_bp == 0) {
			DEBUG1("zxe_loopback_test(%d):  Could not allocate transmit resources\n", ppa);
			ret_val = PDK_NO_MEM;
			goto loopback_cleanup1;
		}

		pdk_buf_set_len(tx_bp, SETUP_FRAME_SIZE);
	}
	else {
		tx_bp = zxe_loopback_ringget(lp->tx_descriptor_handlep, &lp->tx_dequeue_here, 1);

		if (tx_bp == 0) {
			DEBUG1("zxe_loopback_test(%d):  Transmit ring contains other than a setup buffer\n", ppa);
			ret_val = PDK_SYS_ERR;
			goto loopback_cleanup1;
		}

		lp->free_tx_descriptor_count += 2;
		pdk_buf_unmap(tx_bp);
	}

	DEBUG3("zxe_loopback_test(%d):  SETUP BUFFER COMPLETE\n", ppa);

	/* PREPARE TRANSMIT PACKET */
	rptr = pdk_buf_rptr(tx_bp);
	pdk_buf_contiguous(tx_bp, PDK_BUF_WRITE, rptr, &len);
	if (len < SETUP_FRAME_SIZE) {
		DEBUG1("zxe_loopback_test(%d):  Transmit buffer too small\n", ppa);
		ret_val = PDK_SYS_ERR;
		goto loopback_cleanup1;
	}
	zxe_loopback_data(rptr, SETUP_FRAME_SIZE);

	ethp = (pdk_ether_header_t *) rptr;
	pdk_bcopy(lp->CurrentNetworkAddress, &ethp->ether_dhost, PDK_ETHERADDRL);
	pdk_bcopy(lp->CurrentNetworkAddress, &ethp->ether_shost, PDK_ETHERADDRL);
	ethp->ether_type = 0x800;

	/* PUT PHY INTO LOOPBACK MODE */
#ifdef LOOPBACK_21143
	if (lp->PhyType != DEVICE_BCM5208) {

		if (lp->Info.AIFlags & AI_GEP0_LPBK)	/* Use loopback through PHY */
			lp->CurrentMediaType = ZXE_LOOPBACK_TXEXT;
		else
			lp->CurrentMediaType = ZXE_LOOPBACK_TPINT;

		if (ZxeReconfigureMediaType(ppa) != 0) {
			goto loopback_cleanup1;
		}
		if (lp->CurrentMediaType == ZXE_LOOPBACK_TXEXT) {
			pdk_time_delay(zxe_loopback_delay);
		}
	} else
#endif
	{
		ZxeWritePHY(ppa, BCM5208_MIICR_REG, (uint16_t)BCM5208_MIICR_RESET);
		pdk_time_delay(1);
		ZxeWritePHY(ppa, BCM5208_TXACR_REG, (uint16_t)(BCM5208_TXACR_TXDIS | BCM5208_TXACR_CIMDIS));
		pdk_time_delay(1);
		ZxeWritePHY(ppa, BCM5208_MIICR_REG, (uint16_t)(BCM5208_MIICR_LOOPBACK|BCM5208_MIICR_100));
		pdk_time_delay(1);
	
		ZxeWritePHY(ppa, BCM5208_ACSR_REG, (uint16_t)BCM5208_ACSR_FORCELINK);
		pdk_reg_put32(ppa, CSR6, lp->OperationModeBase | CSR6_OM_EXTLB | CSR6_ST | CSR6_SR);
	}

	/* ALLOCATE AND MAP RECEIVE PACKET(if needed) */
	if (lp->rx_enqueue_here == lp->rx_dequeue_here) {
		DEBUG3("zxe_loopback_test(%d): Allocating receive buffer\n", ppa);
		if ((allocated_rx_bp = pdk_buf_alloc(ppa, SETUP_FRAME_SIZE+8, 0)) == 0) {
			DEBUG1("zxe_loopback_test(%d):  Could not allocate receive resources\n", ppa);
			ret_val = PDK_NO_MEM;
			goto loopback_cleanup2;
		}

		pdk_buf_set_len(allocated_rx_bp, SETUP_FRAME_SIZE+8);

		/* PLACE RECEIVE PACKET ON THE RING */
		if (zxe_loopback_ringput(lp, &lp->rx_enqueue_here, allocated_rx_bp, PDK_BUF_RECV, lp->rx_queue) == 0) {
			DEBUG1("zxe_loopback_test(%d):  Ringput failed on loading receive packet\n", ppa);
			ret_val = PDK_SYS_ERR;
			goto loopback_cleanup2;
		}
	}
	else {
		pdk_time_delay(LOOPBACK_DELAY_INTERVAL);
		if ((lp->rx_dequeue_here->RDES0 & RDES0_OWN) == 0) {
			DEBUG1("zxe_loopback_test(%d):  Receive ring has completion  - aborting loopback test\n", ppa);


			pdk_buf_unmap(tx_bp);
			pdk_buf_free(tx_bp);

			/* RESTORE THE PHY */
#ifdef LOOPBACK_21143
			if (lp->PhyType != DEVICE_BCM5208) {
				lp->CurrentMediaType = orig_media;
				ZxeReconfigureMediaType(ppa);
			} else
#endif
			{
				ZxeWritePHY(ppa, BCM5208_MIICR_REG, (uint16_t)BCM5208_MIICR_RESET);
				pdk_time_delay(2);
				ZxeWritePHY(ppa, BCM5208_MIICR_REG, (uint16_t)lp->PhyControl);
				if (lp->AdapterId == ADAPTER_ZX222)
					ZxeWritePHY(ppa, BCM5208_TXACR_REG,
						(uint16_t)(BCM5208_TXACR_FEF | BCM5208_TXACR_CIMDIS));
				pdk_reg_put32(ppa, CSR6, lp->OperationModeBase);
			}

			lp->state = orig_state;

			if (orig_state == LOCAL_ADAPTER_RUNNING)
				SET_CSR7(ppa,orig_csr7);

			return(0);
		}
	}

	DEBUG3("zxe_loopback_test(%d):  RECEIVE BUFFER LOADED\n", ppa);

	/* SEND THE TRANSMIT PACKET */
	if (zxe_loopback_ringput(lp, &lp->tx_enqueue_here, tx_bp, PDK_BUF_XMIT, lp->tx_queue) == 0) {
		DEBUG1("zxe_loopback_test(%d):  Ringput failed on loading transmit packet\n", ppa);
		ret_val = PDK_SYS_ERR;
		goto loopback_cleanup2;
	}
	DEBUG3("zxe_loopback_test(%d):  TRANSMIT BUFFER LOADED\n", ppa);

	/* GET THE TRANMIT COMPLETE */
	if (zxe_loopback_ringget(lp->tx_descriptor_handlep, &lp->tx_dequeue_here, 0) == 0) {
		DEBUG1("zxe_loopback_test(%d):  No Transmit complete\n", ppa);
		ret_val = PDK_HW_ERR;
		goto loopback_cleanup2;
	}
	DEBUG3("zxe_loopback_test(%d):  TRANSMIT COMPLETE\n", ppa);

	/* LOOK FOR RECEIVE COMPLETE */
	if ((rx_bp = zxe_loopback_ringget(lp->rx_descriptor_handlep, &lp->rx_dequeue_here, 0)) == 0) {
		DEBUG1("zxe_loopback_test(%d):  No receive complete\n", ppa);
		ret_val = PDK_HW_ERR;
		goto loopback_cleanup2;
	}
	DEBUG3("zxe_loopback_test(%d):  RECEIVE COMPLETE\n", ppa);
	/* COMPARE DATA */
	pdk_buf_set_len(rx_bp, SETUP_FRAME_SIZE);

	PDK_BUF_SYNC(rx_bp, SETUP_FRAME_SIZE, PDK_BUF_READ);

	if (pdk_bcmp(pdk_buf_rptr(tx_bp),pdk_buf_rptr(rx_bp),SETUP_FRAME_SIZE) != 0) {
		DEBUG1("zxe_loopback_test(%d):  Data compare error\n", ppa);
		ret_val = PDK_HW_ERR;
		goto loopback_cleanup2;
	}

	DEBUG3("zxe_loopback_test(%d): loopback successful\n", ppa);
	ret_val = 0;

loopback_cleanup2:
	/* RESTORE THE PHY */
#ifdef LOOPBACK_21143
	if (lp->PhyType != DEVICE_BCM5208) {
		lp->CurrentMediaType = orig_media;
		ZxeStopTxAndRx(ppa);
		ZxeConfigureMedia(ppa);
	} else
#endif
	{
		ZxeWritePHY(ppa, BCM5208_MIICR_REG, (uint16_t)BCM5208_MIICR_RESET);
		pdk_time_delay(2);
		ZxeWritePHY(ppa, BCM5208_MIICR_REG, (uint16_t)lp->PhyControl);
		if (lp->AdapterId == ADAPTER_ZX222)
			ZxeWritePHY(ppa, BCM5208_TXACR_REG,
				(uint16_t)(BCM5208_TXACR_FEF | BCM5208_TXACR_CIMDIS));
		pdk_reg_put32(ppa, CSR6, lp->OperationModeBase);
	}

loopback_cleanup1:

	lp->state = orig_state;

	/* FREE RESOURCES */
	if (allocated_rx_bp) {
		lp->rx_enqueue_here = lp->rx_dequeue_here;
		lp->rx_enqueue_here->RDES0 = 0;
		pdk_buf_unmap(allocated_rx_bp);
		pdk_buf_free(allocated_rx_bp);
	}
	else if (rx_bp) {
		pdk_buf_unmap(rx_bp);
		lp->rx_num_bufs--;
		zxe_set_recv_buf(ppa, rx_bp, lp);
	}

	if (tx_bp) {
		lp->tx_enqueue_here = lp->tx_dequeue_here;
		lp->free_tx_descriptor_count = num_transmit_desc;
		lp->tx_enqueue_here->TDES0 = 0;
		pdk_buf_unmap(tx_bp);
		pdk_buf_free(tx_bp);
	}

	if (orig_state == LOCAL_ADAPTER_RUNNING) {
		ZxeStartTxAndRx(ppa);
		SET_CSR7(ppa,orig_csr7);
	}

	return(ret_val);
}

/*  zxe_test_fault - use loopback test to determine if fault internal and
	set internal fault led.  Must not hold rx or tx descriptor lock when called.
*/
int32_t
zxe_test_fault(int32_t ppa)
{
	local_adapter_t *lp;
	int32_t		err;
	uint16_t	phy_isr_value;

	lp = local_adapters[ppa];

	if (lp->tx_enqueue_here != lp->tx_dequeue_here) {
		pdk_lock_acquire(lp->tx_descriptor_lock);
		process_xmit_complete(lp);
		pdk_lock_release(lp->tx_descriptor_lock);
	}

	if (lp->tx_enqueue_here != lp->tx_dequeue_here) {
		DEBUG2("zxe_test_fault(%d): Cannot run loopback test - Ring not empty\n", ppa);
		return(0);
	} else {
		pdk_lock_acquire(lp->rx_descriptor_lock);
		pdk_lock_acquire(lp->tx_descriptor_lock);

		err = zxe_loopback_test(ppa);

		pdk_lock_release(lp->tx_descriptor_lock);
		pdk_lock_release(lp->rx_descriptor_lock);

		if (err == PDK_HW_ERR) {
			DEBUG1("zxe_test_fault(%d): Loopback test failed\n", ppa);

			zxe_operational_led(ppa, 0);
			zxe_internal_fault_led(ppa, 1);
			return(PDK_HW_ERR);
		}

		/* Re-Enable PHY interrupts */
		if (lp->PhyType == DEVICE_BCM5208) {
			ZxeWritePHY(ppa, BCM5208_IR_REG, (uint16_t)BCM5208_IR_INTEN);
			ZxeReadPHY(ppa, BCM5208_IR_REG, &phy_isr_value);
		}
	
		return(0);
	}
}

void
zxe_internal_fault_led(ppa, on)
int32_t ppa;
int32_t on;
{
	pdk_psdb_hal_fault_t phf;
	local_adapter_t *lp;
	int32_t		ppa_zero;
	int32_t		orig_ppa;
	int32_t		unit;
	int32_t		csr15;
	int32_t		num_ports;

	orig_ppa = ppa;
	lp = local_adapters[ppa];

	DEBUG3("zxe_internal_fault_led(%d):  Called to turn %s the LED\n", orig_ppa, on ? "on" : "off");

	if (lp == 0)
		return;

	(*lp->layer_routines.pdk_status)(ppa, PDK_STATUS_INTERNAL_FAULT, (void *)on);
	if (lp->fault_notification) {
		phf.phf_header.ppr_header.Size = psdb_ntop(sizeof(phf));
		phf.phf_header.ppr_header.Type = psdb_ntop(PSDB_TYPE_HAL_FAULT_NOTIFICATION);
		phf.phf_header.ppr_ppa = psdb_ntop(ppa);
		phf.phf_internal_fault = 1;  /* 0 indicates external fault */
		phf.phf_on = psdb_ntop(on);

		(*lp->layer_routines.pdk_status)(ppa, PDK_PSDB, (void *) &phf);
	}

	if (on)
		lp->led_local_settings |= INT_FAULT_LED;
	else
		lp->led_local_settings &= ~INT_FAULT_LED;

	if (!(lp->Info.AIFlags & AI_LEDS))
		return;

	if (lp->Info.AIFlags & AI_SHARED_LEDS) {
		ppa_zero = pdk_ntbridge_unit_ppa(ppa, 0);

		if (ppa_zero == -1) {
			DEBUG1("zxe_internal_fault_led(%d): Cannot find first ppa\n", orig_ppa);	
			return;
		}
	} else
		ppa_zero = ppa;

	if (on) {

		local_statics[ppa_zero].led_shadow |= INT_FAULT_LED;

		PHYLOCK(ppa_zero);

		pdk_reg_put32(ppa_zero, CSR10, 0);
		pdk_time_delay(1);
		pdk_reg_put32(ppa_zero, CSR9, CSR9_SWO | CSR9_SREG |
			local_statics[ppa_zero].led_shadow);
		pdk_time_delay(1);
		pdk_reg_put32(ppa_zero, CSR9, CSR9_IDLE);
		pdk_time_delay(1);

		PHYUNLOCK(ppa_zero);

		DEBUG3("zxe_internal_fault_led(%d):  LED ON\n", orig_ppa);
		return;
	} else {
		if (lp->Info.AIFlags & AI_SHARED_LEDS) {
			/* Turn off the fault LED only if no port is in fault state */
			num_ports = lp->Info.NumChannels;
			for (unit = 0; unit < num_ports; unit++) {
				ppa = pdk_ntbridge_unit_ppa(orig_ppa, unit);
				if (ppa == -1) {
					DEBUG1("zxe_internal_fault_led(%d): Cannot find unit %d\n",
						orig_ppa, unit);	
					return;
				}
	
				lp = local_adapters[ppa];
	
				if (lp == 0)
					continue;
	
				if (lp->led_local_settings & INT_FAULT_LED) {
					DEBUG3("zxe_internal_fault_led(%d):  ppa %d in Fault state\n",
						orig_ppa, lp->ppa);
					return;
				}
			}
		}

		local_statics[ppa_zero].led_shadow &= ~INT_FAULT_LED;

		PHYLOCK(ppa_zero);
		
		pdk_reg_put32(ppa_zero, CSR10, 0);
		pdk_time_delay(1);
		pdk_reg_put32(ppa_zero, CSR9, CSR9_SWO|CSR9_SREG|
			local_statics[ppa_zero].led_shadow);
		pdk_time_delay(1);
		pdk_reg_put32(ppa_zero, CSR9, CSR9_IDLE);

		PHYUNLOCK(ppa_zero);
	}

	DEBUG3("zxe_internal_fault_led(%d):  LED OFF\n", orig_ppa);
}

void
zxe_external_fault_led(ppa, on, failover)
int32_t ppa;
int32_t on;
int32_t failover;
{
	pdk_psdb_hal_fault_t phf;
	local_adapter_t *lp;
	int32_t		ppa_zero;
	int32_t		orig_ppa;
	int32_t		unit;
	int32_t		csr15;
	int32_t		num_ports;

	orig_ppa = ppa;
	lp = local_adapters[ppa];

	DEBUG3("zxe_external_fault_led(%d):  Called to turn %s the LED for %s\n", orig_ppa, on ? "on" : "off", failover ? "FAILOVER" : "LINK");

	if (lp == 0)
		return;

	if (on) {
		if (failover)
			lp->led_local_settings |= FAILOVER_FAULT_LED;
		else
			lp->led_local_settings |= EXT_FAULT_LED;
	} else {
		if (failover)
			lp->led_local_settings &= ~FAILOVER_FAULT_LED;
		else
			lp->led_local_settings &= ~EXT_FAULT_LED;
	}

	if (failover) {
		if ((lp->led_local_settings & EXT_FAULT_LED) == 0)
			(*lp->layer_routines.pdk_status)(ppa, PDK_STATUS_EXTERNAL_FAULT, (void *) on);
	}
	else {
		if ((lp->led_local_settings & FAILOVER_FAULT_LED) == 0)
			(*lp->layer_routines.pdk_status)(ppa, PDK_STATUS_EXTERNAL_FAULT, (void *) on);
	}

	if (lp->fault_notification & (failover == 0)) {
		phf.phf_header.ppr_header.Size = psdb_ntop(sizeof(phf));
		phf.phf_header.ppr_header.Type = psdb_ntop(PSDB_TYPE_HAL_FAULT_NOTIFICATION);
		phf.phf_header.ppr_ppa = psdb_ntop(ppa);
		phf.phf_internal_fault = 0;  /* 0 indicates external fault */
		phf.phf_on = psdb_ntop(on);

		(*lp->layer_routines.pdk_status)(ppa, PDK_PSDB, (void *) &phf);
	}

	if (!(lp->Info.AIFlags & AI_LEDS))
		return;

	if (lp->Info.AIFlags & AI_SHARED_LEDS) {
		ppa_zero = pdk_ntbridge_unit_ppa(ppa, 0);

		if (ppa_zero == -1) {
			DEBUG1("zxe_external_fault_led(%d):  Cannot find first ppa\n", orig_ppa);	
			return;
		}
	} else
		ppa_zero = ppa;

	if (on) {
		local_statics[ppa_zero].led_shadow |= EXT_FAULT_LED;

		PHYLOCK(ppa_zero);

		pdk_reg_put32(ppa_zero, CSR10, 0);
		pdk_time_delay(1);
		pdk_reg_put32(ppa_zero, CSR9, CSR9_SWO|CSR9_SREG|
			local_statics[ppa_zero].led_shadow);
		pdk_time_delay(1);
		pdk_reg_put32(ppa_zero, CSR9, CSR9_IDLE);

		PHYUNLOCK(ppa_zero);

		DEBUG3("zxe_external_fault_led(%d):  LED ON for %s\n", orig_ppa, failover ? "FAILOVER" : "LINK");

	} else {
		if (lp->Info.AIFlags & AI_SHARED_LEDS) {
			/* Turn off the fault LED only if no port is in fault state */
			num_ports = lp->Info.NumChannels;
			for (unit = 0; unit < num_ports; unit++) {
				ppa = pdk_ntbridge_unit_ppa(orig_ppa, unit);
				if (ppa == -1) {
					DEBUG1("zxe_external_fault_led(%d): Cannot find unit %d\n", orig_ppa, unit);	
					return;
				}
				lp = local_adapters[ppa];
				if (lp == 0)
					continue;
				if (lp->state != LOCAL_ADAPTER_RUNNING)
					continue;
		
				if (lp->led_local_settings & (EXT_FAULT_LED|FAILOVER_FAULT_LED)) {
					DEBUG3("zxe_external_fault_led(%d):  ppa %d in Fault state\n", orig_ppa, lp->ppa);
					return;
				}
			}
		}
		
		local_statics[ppa_zero].led_shadow &= ~EXT_FAULT_LED;
	
		PHYLOCK(ppa_zero);
	
		pdk_reg_put32(ppa_zero, CSR10, 0);
		pdk_time_delay(1);
		pdk_reg_put32(ppa_zero, CSR9, CSR9_SWO|CSR9_SREG|
			local_statics[ppa_zero].led_shadow);
		pdk_time_delay(1);
		pdk_reg_put32(ppa_zero, CSR9, CSR9_IDLE);
	
		PHYUNLOCK(ppa_zero);

		DEBUG3("zxe_external_fault_led(%d):  LED OFF for %s\n", orig_ppa, failover ? "FAILOVER" : "LINK");
	}
}

void
zxe_operational_led(ppa, on)
int32_t ppa;
int32_t on;
{
	local_adapter_t *lp;
	int32_t		ppa_zero;
	int32_t		orig_ppa;
	int32_t		unit;
	int32_t		csr15;
	int32_t		num_ports;

	orig_ppa = ppa;
	lp = local_adapters[ppa];

	DEBUG3("zxe_operational_led(%d):  Called to turn %s the LED\n", ppa, on ? "on" : "off");

	if (lp == 0)
		return;

	if (on)
		lp->led_local_settings |= OPERATIONAL_LED;
	else
		lp->led_local_settings &= ~OPERATIONAL_LED;

	if (!(lp->Info.AIFlags & AI_LEDS))
		return;

	if (lp->Info.AIFlags & AI_SHARED_LEDS) {
		ppa_zero = pdk_ntbridge_unit_ppa(ppa, 0);

		if (ppa_zero == -1) {
			DEBUG1("zxe_operational_led(%d): Cannot find zeroeth ppa\n", ppa);	
			return;
		}
	} else
		ppa_zero = ppa;

	/* If any port turns on the operational LED, its on. */
	if (on) {

		local_statics[ppa_zero].led_shadow |= OPERATIONAL_LED;

		PHYLOCK(ppa_zero);

		pdk_reg_put32(ppa_zero, CSR10, 0);
		pdk_time_delay(1);
		pdk_reg_put32(ppa_zero, CSR9, CSR9_SWO|CSR9_SREG|
			local_statics[ppa_zero].led_shadow);
		pdk_time_delay(1);
		pdk_reg_put32(ppa_zero, CSR9, CSR9_IDLE);

		PHYUNLOCK(ppa_zero);

		DEBUG3("zxe_operational_led(%d):  LED ON\n", orig_ppa);
	} else {
	
		num_ports = lp->Info.NumChannels;

		if (lp->Info.AIFlags & AI_SHARED_LEDS) {
			/* Only if all ports turn it off is it off */
			for (unit = 0; unit < num_ports; unit++) {
				ppa = pdk_ntbridge_unit_ppa(orig_ppa, unit);
				if (ppa == -1) {
					DEBUG1("zxe_operational_led(%d): Cannot find unit %d\n", orig_ppa, unit);	
					return;
				}
		
				lp = local_adapters[ppa];
				if (lp == 0)
					continue;
				if (lp->state != LOCAL_ADAPTER_RUNNING)
					continue;
		
				if (lp->led_local_settings & OPERATIONAL_LED) {
					DEBUG3("zxe_operational_led(%d):  ppa %d in Operational state\n", orig_ppa, ppa);
					return;
				}
			}
		}
	
		local_statics[ppa_zero].led_shadow &= ~OPERATIONAL_LED;
	
		PHYLOCK(ppa_zero);
	
		pdk_reg_put32(ppa_zero, CSR10, 0);
		pdk_time_delay(1);
		pdk_reg_put32(ppa_zero, CSR9, CSR9_SWO|CSR9_SREG|
			local_statics[ppa_zero].led_shadow);
		pdk_time_delay(1);
		pdk_reg_put32(ppa_zero, CSR9, CSR9_IDLE);
	
		PHYUNLOCK(ppa_zero);
	
		DEBUG3("zxe_operational_led(%d):  LED OFF\n", orig_ppa);

		if ((lp->Info.AIFlags & AI_SHARED_PHYLOCK)
			&& (lp->PhyType == DEVICE_BCM5208)) {
	
			pdk_time_delay(100);		/* RESET_WAIT time for PHY RESET */
	
			/* Isolate all nodes initially, or if board not operational */
			for (unit = 0; unit < num_ports; unit++) {
				ppa = pdk_ntbridge_unit_ppa(orig_ppa, unit);
	
				/* Set PHY Super-Isolate */
				ZxeWritePHY(ppa, BCM5208_AMPR_REG, (uint16_t)BCM5208_AMPR_SUPI);
			}
		}
	}	
	return;
}

#ifdef DEBUG_CHECK_RING
void
check_ring(lp)
local_adapter_t *lp;
{
	desc_t *dd;
	int_t num_desc, num_own;

	num_desc = 0;
	num_own = 0;

	dd = lp->rx_dequeue_here;
	if (dd != lp->rx_queue->NEXT) {
		DEBUG1("check_ring(%d): Entries don't start at beginning of queue\n", lp->ppa);
		return;
	}

	while (dd != lp->rx_enqueue_here) {

		num_desc++;
		if (dd->RDES0 & RDES0_OWN) {
			num_own++;
		}
		else if (num_own == 0) {
			DEBUG1("check_ring(%d): Entry(%d) 0x%x unowned descriptors in middle of list\n", lp->ppa, num_desc, dd);
		}
		dd = dd->NEXT;
		if (num_desc > 100) {
			DEBUG1("check_ring(%d): num_desc overflow\n", lp->ppa);
			break;
		}
	}

	DEBUG2("check_ring(%d):  Ring good %d descriptors %d owned\n", lp->ppa, num_desc, num_own);
}
#endif /* DEBUG_CHECK_RING */


